# Common.pri - This script is included from every .pro file
#
# This file provides the following macros for c files, they are also
# added to CONFIG for use in qmake files.
# * PLATFORM_OMAP
# * PLATFORM_ANDROID
# * PLATFORM_WORKSTATION
# * PRODUCT_ATOM
# * PRODUCT_AXIOM
# * PRODUCT_HORNET
#
# The PRODUCT_* macros should be used for features in different products.
# i.e. the code has been partitioned because a product manager said so.
#
# The PLATFORM_* macros should be used for different implementations due to
# underlying platforms.
# i.e. WiFi on android cannot be implemented the same way we did on OMAP.

## Work Out What We're Building.
# This table shows our supported combinations of products and platforms.
# | Platforms\Products| Axiom     | Atom      | Hornet    |
# |===================|===========|===========|===========|
# | Android           | Customer  | N/A       | Customer  |
# | Busybox on OMAP   | N/A       | Customer  | N/A       |
# | Desktop           | Developer | Developer | Developer |

## Derive Platform.
# The platform is entirely dependent on which toolchain we are using.
# We cannot build android code with the omap toolchain.
#
# Interesting notes:
# * The unix scope is valid for all of our platforms.
#   and as such is not interesting to us.
# * The plain linux scope is valid across both OMAP & Desktop
#   so is also less useful.
# * The ANDROID and __ANDROID__ definition are set by different
#   mechanisms, ANDROID is set by ndk-build. However the Android Standalone Toolchain
#   defines __ANDORID__ (source: https://stackoverflow.com/a/15335544)
#
# All these are why we will define our own, and it can be a single central source of truth
# If we later change toolchains, we simply need to update this small section.

linux-g++ {
    # We are running with a standard desktop toolchain
    DEFINES += PLATFORM_WORKSTATION
    CONFIG += PLATFORM_WORKSTATION
}
linux-buildroot-g++ {
    # This is how the OMAP toolchain identifys
    DEFINES += PLATFORM_OMAP
    CONFIG += PLATFORM_OMAP
}
android {
    # We are building with the android version of Qt.
    DEFINES += PLATFORM_ANDROID
    CONFIG += PLATFORM_ANDROID
}

# Defining the Product. This is simple passed by qmake
contains(PRODUCT, ATOM|AXIOM|HORNET) {
    DEFINES += PRODUCT_$$PRODUCT
    CONFIG += PRODUCT_$$PRODUCT
} else {
    error("PRODUCT has been given the invalid value: $$PRODUCT")
}

# A Shim for all existing qmake code that currently queries TARGETNAME
PLATFORM_OMAP:PRODUCT_ATOM:TARGETNAME=ATOM_OMAP4
PLATFORM_WORKSTATION:TARGETNAME=WORKSTATION

# Build System Versioning
defineReplace(stripLeadingZeros) {
    varname = $$1
    tmp = $$find($$varname, [^0][0-9]*)
    isEmpty(tmp): tmp=0
    return($$tmp)
}

defineReplace(setOrDefault) {
    tmp = $$1
    isEmpty(tmp): tmp=$$2
    return($$tmp)
}

BUILD_ID = $$setOrDefault($$(BuildNum), "99999")
BUILD_ID = $$stripLeadingZeros(BUILD_ID)

BUILD_VERSION = $$setOrDefault($$(BuildVer), "99.99")
BUILD_VER_LIST = $$split(BUILD_VERSION, '.')

BUILD_VERSION_MAJOR=$$member(BUILD_VER_LIST, 0)
BUILD_VERSION_MAJOR = $$stripLeadingZeros(BUILD_VERSION_MAJOR)

BUILD_VERSION_MINOR=$$member(BUILD_VER_LIST, 1)
BUILD_VERSION_MINOR = $$stripLeadingZeros(BUILD_VERSION_MINOR)

DEFINES *= _AUTO_BUILD_ID_=$$BUILD_ID
DEFINES *= _AUTO_BUILD_MAJOR_=$$BUILD_VERSION_MAJOR
DEFINES *= _AUTO_BUILD_MINOR_=$$BUILD_VERSION_MINOR
# END Build System Versioning

DEFINES *= __STDC_FORMAT_MACROS
DEFINES *= GL_GLEXT_PROTOTYPES


DEFINES *= EXCLUDE_CLOUD

## Target specific defines for ATOM_OMAP4 ------------
contains(TARGETNAME, ATOM_OMAP4) {
    DEFINES *= HARDWARE_SUPPORTS_FLOB_MTD

    DEFINES *= HARDWARE_CONFIG_FLOB_MTD_PARTITION=/dev/mtd0
    DEFINES *= HARDWARE_CONFIG_LAST_GUID_MTD_PARTITION=/dev/mtd1
#    DEFINES *= HARDWARE_SUPPORTS_OMAP4_VIDEO_IN
#    DEFINES *= HARDWARE_SUPPORTS_MULTITOUCH
}

#    C++ warning settings
#COMMON_WARNINGS = -Wall -Wextra -Winit-self -Wno-psabi -Werror
COMMON_WARNINGS = -Wall -Wextra -Winit-self -Werror -Wno-strict-overflow -Wno-unknown-pragmas
# COMMON_WARNINGS *= -Wshadow
QMAKE_CFLAGS_WARN_ON *= $$COMMON_WARNINGS
QMAKE_CXXFLAGS_WARN_ON *= $$COMMON_WARNINGS
QMAKE_CXXFLAGS_WARN_OFF *= -w -Wno-c++11-narrowing
QMAKE_CFLAGS_WARN_OFF *= -w

SUPPRESS_ERRORS += -Wno-error=constant-conversion
SUPPRESS_ERRORS += -Wno-error=enum-compare-switch
SUPPRESS_ERRORS += -Wno-error=inconsistent-missing-override
SUPPRESS_ERRORS += -Wno-error=missing-braces
SUPPRESS_ERRORS += -Wno-error=non-literal-null-conversion
SUPPRESS_ERRORS += -Wno-error=overloaded-virtual
SUPPRESS_ERRORS += -Wno-error=pointer-bool-conversion
SUPPRESS_ERRORS += -Wno-error=potentially-evaluated-expression
SUPPRESS_ERRORS += -Wno-error=shift-count-overflow
SUPPRESS_ERRORS += -Wno-error=sign-compare
SUPPRESS_ERRORS += -Wno-error=undefined-internal
SUPPRESS_ERRORS += -Wno-error=unknown-warning-option
SUPPRESS_ERRORS += -Wno-error=unused-const-variable
SUPPRESS_ERRORS += -Wno-error=unused-parameter
SUPPRESS_ERRORS += -Wno-error=unused-private-field
SUPPRESS_ERRORS += -Wno-error=unused-variable

PLATFORM_ANDROID {
    QMAKE_CFLAGS_WARN_ON += $$SUPPRESS_ERRORS
    QMAKE_CXXFLAGS_WARN_ON += $$SUPPRESS_ERRORS
    #DEFINES *= EXCLUDE_CMAP
    DEFINES *= ENABLE_REALBATHY
}

PLATFORM_WORKSTATION {
    CONFIG += exclude_external_libs
    DEFINES *= EXCLUDE_CZONE
    DEFINES *= EXCLUDE_CMAP
#    DEFINES *= EXCLUDE_MAPBOX
    DEFINES *= EXCLUDE_NAVIONICS
    #DEFINES *= ENABLE_REALBATHY
}

CONFIG *= no_keywords # Avoid Boost signals conflict with Qt signals

android {
    equals(ANDROID_TARGET_ARCH, armeabi-v7a) {
        ANDROID_LIBS_ROOT = $$PWD/AndroidLibs/android_armv7
        INCLUDEPATH *= $$ANDROID_LIBS_ROOT/include
        QMAKE_CFLAGS += -isystem $$ANDROID_LIBS_ROOT/include
        QMAKE_CXXFLAGS += -isystem $$ANDROID_LIBS_ROOT/include
        QMAKE_LFLAGS *= -L$$ANDROID_LIBS_ROOT/lib/
    }
    else {
        error(Android architecture $$ANDROID_TARGET_ARCH not supported)
    }

    DEFINES *= TARGET_ANDROID
    DEFINES *= HARDWARE_SUPPORTS_KERNEL_BUZZER_SOUND
}

!android {
    QMAKE_LFLAGS *= -Wl,--no-undefined -Wl,-z,noexecstack
    DEFINES *= TARGET_LINUX
}



PLATFORM_WORKSTATION {
    DEFINES *= TARGET_DESKTOP
    DEFINES *= HARDWARE_SUPPORTS_KVASER_CAN
    DEFINES *= HARDWARE_SUPPORTS_GLX
    DEFINES *= HARDWARE_SUPPORTS_MOUSE
    # FIXME: ASAN is currently broken on Workstation Builds
    # !DISABLE_SANITIZERS {
    #     QMAKE_CXXFLAGS+=-fsanitize=address -fno-omit-frame-pointer
    #     QMAKE_CFLAGS+=-fsanitize=address -fno-omit-frame-pointer
    #     QMAKE_LFLAGS+=-fsanitize=address
    #     QMAKE_CXXFLAGS+=-fsanitize=undefined -fno-omit-frame-pointer
    #     QMAKE_CFLAGS+=-fsanitize=undefined -fno-omit-frame-pointer
    #     QMAKE_LFLAGS+=-fsanitize=undefined
    # }
}

!contains(DEFINES, TARGET_DESKTOP) {
    DEFINES *= HARDWARE_SUPPORTS_SOCKET_CAN
    DEFINES *= HARDWARE_SUPPORTS_OPENGL_ES2
    DEFINES *= HARDWARE_SUPPORTS_EGL
    DEFINES *= HARDWARE_SUPPORTS_POWEROFF
    DEFINES *= HARDWARE_SUPPORTS_REBOOT
}

# To isolate MFDApplication running on a Workstation from other
# MFDApplication running on other workstations, uncomment the
# following line and replace YourName with your name.
#DEFINES *= UNIQUE_SERVICE_DISCOVERY_NAME='\\"YourName_MFD\\"'

#Used to be in project.h (project_ATOM_linux_rev_0.h)
DEFINES *= HARDWARE_SUPPORTS_LNET
DEFINES *= FEATURE_SUPPORTS_RADAR
DEFINES *= RAYCHAR_IS_WIDE
DEFINES *= TARGET_LH3

TARGET_INSTALL_PATH = /raymarine/Application
android {
    ASSETS_INSTALL_PATH = /assets
    CACHED_ASSETS_INSTALL_PATH = $$ASSETS_INSTALL_PATH/AssetsToCache
}
else {
    ASSETS_INSTALL_PATH = $$TARGET_INSTALL_PATH
    CACHED_ASSETS_INSTALL_PATH = $$TARGET_INSTALL_PATH
}

# For ATOM build...
contains(TARGETNAME, ATOM_OMAP4) {
    DEFINES *= TARGET_ATOM
    #DEFINES *= OSS_INCLUDE_THREAD_ANALYSIS

    QMAKE_LFLAGS *= -Wl,-rpath,.
    QMAKE_LFLAGS *= -Wl,-rpath,/raymarine/Application
    QMAKE_LFLAGS *= -L$$[QT_SYSROOT]/lib
}

contains(DEFINES, TARGET_ANDROID) {
    TARGETTYPE = Apollo $$(ANDROID_NDK_PLATFORM) for $$ANDROID_TARGET_ARCH
}
contains(DEFINES, TARGET_ATOM) {
    DEFINES *= HARDWARE_SUPPORTS_KERNEL_BUZZER_SOUND

    TARGETTYPE = Atom
}
contains(DEFINES, TARGET_DESKTOP) {
    TARGETTYPE = Desktop
}

contains(DEFINES, TARGET_DESKTOP) {
    # Exclude CZone library from workstation builds
    DEFINES *= EXCLUDE_CZONE
}

PLATFORM_ANDROID:PRODUCT_AXIOM {
    DEFINES += SUPPORTS_SR200
    CONFIG += SUPPORTS_SR200
}

# Override standard Qt optimisation settings from mkspecs to give consistent settings between platforms
# The recommended optimisation level for debug is -Og, however, this gives false positives for -Wmaybe-unitialized in boost::lexical_cast
# If you require -Og, then you must also disable the maybe unitialized warnings with -Wno-error=maybe-unitialized
# This is done for you when you pass `CONFIG+=debug CONFIG+=full_debug` to qmake (see below)
QMAKE_CFLAGS ~= s/-O[0123sg]/
QMAKE_CXXFLAGS ~= s/-O[0123sg]/
QMAKE_CFLAGS_DEBUG ~= s/-O[0123sg]/
QMAKE_CFLAGS_DEBUG *= -O2
QMAKE_CXXFLAGS_DEBUG ~= s/-O[0123sg]/
QMAKE_CXXFLAGS_DEBUG *= -O2
QMAKE_CFLAGS_RELEASE ~= s/-O[0123sg]/
QMAKE_CFLAGS_RELEASE *= -O2
QMAKE_CXXFLAGS_RELEASE ~= s/-O[0123sg]/
QMAKE_CXXFLAGS_RELEASE *= -O2

# Build all modes with debug symbols
QMAKE_CFLAGS_RELEASE *= -g
QMAKE_CXXFLAGS_RELEASE *= -g

# Ensure we always have unwind tables for backtracing
QMAKE_CFLAGS *= -funwind-tables
QMAKE_CXXFLAGS *= -funwind-tables
QMAKE_CFLAGS *= -fasynchronous-unwind-tables
QMAKE_CXXFLAGS *= -fasynchronous-unwind-tables

CONFIG(release, debug|release) {
    DEFINES *= NDEBUG
    DEFINES *= RELEASE
    DEBUGMODE = release
    !PRODUCT_HORNET {
        CONFIG+=qtquickcompiler
	}
}

CONFIG(debug, debug|release) {
    DEFINES *= ENABLE_MEMORY_EVENT_LOGGER
    DEBUGMODE *= debug
    android {
        #QT += GammaRayProbe
    }
}

CONFIG(full_debug) {
    QMAKE_CXXFLAGS_WARN_ON += "-Wno-error=maybe-uninitialized"
    QMAKE_CXXFLAGS_DEBUG ~= s/-O[0123sg]/
    QMAKE_CXXFLAGS_DEBUG *= -Og
    QMAKE_CFLAGS_DEBUG ~= s/-O[0123sg]/
    QMAKE_CFLAGS_DEBUG *= -Og
}

message("Building $$TARGET v$$BUILD_VERSION-$$BUILD_ID for $$TARGETTYPE - Qt $$QT_VERSION ($$DEBUGMODE)")

!DONT_MODIFY_COMMANDS {
    # Modify tools commands to print a nice message and not the command line
    # NOTE: Do this last as other things may modify these commands!!!
    !contains(QMAKE_CC, ccache) {
        !system(readlink $(which $$QMAKE_CC) | grep -q ccache):QMAKE_CC=ccache $$QMAKE_CC
    }
    !contains(QMAKE_CXX, ccache) {
        !system(readlink $(which $$QMAKE_CXX) | grep -q ccache):QMAKE_CXX=ccache $$QMAKE_CXX
    }
    !contains(QMAKE_LINK, ccache) {
        !system(readlink $(which $$QMAKE_LINK) | grep -q ccache):QMAKE_LINK=ccache $$QMAKE_LINK
    }
}

contains(DEFINES, TARGET_DESKTOP) {
    exists($$OUT_PWD/Makefile) {
         QMAKE_POST_LINK += $$PWD/Scripts/CodeAnalysis/cppcheck.sh $$OUT_PWD/Makefile --changed;
    }
}

android:equals(TEMPLATE, "app") {
    # QML_IMPORT_PATH cannot be set to use $(INSTALL_ROOT). Without setting it
    # we must put plugins into the $$[QT_QML_DIR], which is part of the system
    # installation. Not what we want. We could specify based on build directory,
    # but I couldn't get that to work. Instead, as an install step, we edit the
    # json file to use the value of $(INSTALL_ROOT).
    #
    # This scope is to limit it to projects that will actually have
    # deployment-settings.json files

    # Insert a dummy value
    IMP_UNIQUE_VALUE = TO_BE_UPDATED_BY_SED
    QML_IMPORT_PATH += $$IMP_UNIQUE_VALUE

    # What the json file is called
    deployment_json = $$OUT_PWD/android-lib$${TARGET}.$${QMAKE_EXTENSION_SHLIB}-deployment-settings.json
    # It needs a path to be added to installs
    json_fix.path = /raymarine/Application
    # Replace dummy value with the actual value. This allows androiddeployqt to
    # generate the apk correctly.
    json_fix.extra = sed -i \'s!$$IMP_UNIQUE_VALUE!$(INSTALL_ROOT)/raymarine/Application!\' $$deployment_json || true

    INSTALLS += json_fix
}

#SSO code excluded by default
#DEFINES *= SSO_ENABLED
