/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "BacktraceDirectory.h"

#include "RayAssert.h"
#include "EventLogger.h"

#include <boost/regex.hpp>
#include <boost/filesystem.hpp>
#include <boost/system/error_code.hpp>
#include <boost/date_time.hpp>
#include <boost/date_time/posix_time/posix_time_io.hpp>

#include <sstream>

#include <sys/stat.h>
#include <stdlib.h>

#include <cstdlib>
#include <cstring>

namespace bfs = boost::filesystem;
namespace bpt = boost::posix_time;

static const char* c_LogFileExtension = "log";

static const char* c_LastSeenLogFilename = "LastSeenLog.txt";

static const char* c_SysLogName = "SYSLOG";

static const char* c_ArchiveExtension = "tgz";

std::string CBacktraceDirectory::m_BacktraceDirectory;

//--------------------------------------------------------------------------------
CBacktraceDirectory::CBacktraceDirectory()
{
    OpenOrCreateBacktraceDirectory();
}

//--------------------------------------------------------------------------------
CBacktraceDirectory::~CBacktraceDirectory()
{
}

//--------------------------------------------------------------------------------
void CBacktraceDirectory::Init(const char* backtraceLogFolder)
{
    m_BacktraceDirectory = backtraceLogFolder;
}

//--------------------------------------------------------------------------------
bool CBacktraceDirectory::OpenOrCreateBacktraceDirectory()
{
    // Must call Init from main first
    RAYASSERT(!m_BacktraceDirectory.empty());

    bool result = m_directory.Open(m_BacktraceDirectory.c_str());
    if (!result)
    {
        if (CreateDirectoryRecursive(m_BacktraceDirectory))
        {
            result = m_directory.Open(m_BacktraceDirectory.c_str());
        }
    }
    return result;
}

//--------------------------------------------------------------------------------
int CBacktraceDirectory::CopySysLogToCard(const std::string& DestinationFolder,
                                          IStatusUpdate* /*statusUpdateInterface*/)
{
    int NumberCopied = 0;

    // Get the local time in a specific format. Not too worried about accuracy...
    bpt::ptime Time(bpt::second_clock::local_time());
    std::ostringstream DateTime;
    DateTime.imbue(std::locale(DateTime.getloc(), new bpt::time_facet("%Y-%m-%d-%H-%M-%S")));
    DateTime << Time;

    std::string ArchiveDestination = DestinationFolder + "/" + c_SysLogName;
    std::string ArchiveFile = ArchiveDestination + "_ALL-" + DateTime.str() + "." + c_ArchiveExtension;

    // Delete the file if it exists at the destination. This can only happen if
    // we do not have a real date/time and we have old log files left on the card.
    std::string ArchiveCommand = "rm -f " + ArchiveFile + "; tar -c -z -f " + ArchiveFile +
                                 " -C " + m_BacktraceDirectory + " syslog";

    RAYDETAIL("Archiving '%s' to '%s'.", (m_BacktraceDirectory + "/syslog").c_str(), ArchiveFile.c_str());

    int Status = std::system(ArchiveCommand.c_str()); // <- This does the actual tar/copy.
    if (_CheckSystemStatus(Status, "archive syslog"))
    {
        ++NumberCopied;
    }

    bfs::path Src(m_BacktraceDirectory + "/syslog/current");
    bfs::path Dst(ArchiveDestination + "_CURRENT-" + DateTime.str() + "." + c_LogFileExtension);

    RAYDETAIL("Copying '%s' to '%s'.", Src.string().c_str(), Dst.string().c_str());

    // Copy the current file, so we have something in plain text to look at in a CCB.
    boost::system::error_code ErrorCode;
    bfs::copy_file(Src, Dst, bfs::copy_option::overwrite_if_exists, ErrorCode);

    if (ErrorCode != boost::system::errc::success)
    {
        RAYERROR("Failed to copy current syslog to card. Actual error: %s.",
                 boost::system::error_code(ErrorCode).message().c_str());
    }
    else
    {
        ++NumberCopied;
    }

    sync();

    return NumberCopied;
}

//--------------------------------------------------------------------------------
int CBacktraceDirectory::MoveLogsFolderContentToCard(const std::string& DestinationFolder, IStatusUpdate* statusUpdateInterface)
{
    bool result = true;

    // Get the most up-to-date list to copy from.
    m_directory.Reload();

    m_directory.ResetGetNextFilename();
    int NumberCopied = 0;

    std::string filename;
    while (result && m_directory.GetNextFilename(filename))
    {
        //
        // Don't move the last seen log filename, as this will
        // insure that the log numbers for this unit continually
        // go upwards, even after they have been moved off the device.
        //
        if (filename.compare(c_LastSeenLogFilename) != 0)
        {
            RAYDETAIL("Moving file '%s' to '%s'", filename.c_str(), DestinationFolder.c_str());

            result = MoveFile(filename, DestinationFolder, statusUpdateInterface);

            if (result)
            {
                ++NumberCopied;
            }
        }
    }

    sync();

    return NumberCopied;
}

//--------------------------------------------------------------------------------
void
CBacktraceDirectory::ClearLogsFolder()
{
    DeleteLastSeenLogFile();
    
    m_directory.ResetGetNextFilename();
        
    std::string filename;
    while (m_directory.GetNextFilename(filename))
    {
        m_directory.DeleteFile(filename);
    }
    
}

//--------------------------------------------------------------------------------
bool CBacktraceDirectory::MoveFile(const std::string& SrcFilename,
                                   const std::string& DestDir,
                                   CBacktraceDirectory::IStatusUpdate* statusUpdateInterface)
{
    bool result = CopyFile(SrcFilename, DestDir, statusUpdateInterface);
    
    if (result)
    {
        result = m_directory.DeleteFile(SrcFilename);

        if (!result)
        {
            RAYWARNING("Failed to delete log: '%s'", SrcFilename.c_str());
        }
    }
    
    return result;
}

//--------------------------------------------------------------------------------
bool CBacktraceDirectory::CopyFile(const std::string& SrcFilename,
                                   const std::string& DestDir,
                                   CBacktraceDirectory::IStatusUpdate* statusUpdateInterface)
{
    bfs::path Src(m_directory.BuildFullname(SrcFilename));
    bfs::path Dst(DestDir + "/" + SrcFilename);

    boost::system::error_code ErrorCode;
    bfs::copy_file(Src, Dst, bfs::copy_option::overwrite_if_exists, ErrorCode);
    
    if (ErrorCode != boost::system::errc::success)
    {
        if (statusUpdateInterface != 0)
        {
            statusUpdateInterface->Update(std::string("Copying file ") + SrcFilename + " Failed");
        }

        RAYERROR("Failed to copy file from '%s' to '%s'. Actual error: %s.",
                 Src.string().c_str(), Dst.string().c_str(),
                 boost::system::error_code(ErrorCode).message().c_str());
    }
    
    return ErrorCode == boost::system::errc::success;
}

//--------------------------------------------------------------------------------
void CBacktraceDirectory::ShowFilenameInStatus(const std::string& filename,
                                               CBacktraceDirectory::IStatusUpdate* statusUpdateInterface)
{
    std::string StatusString("Copying: ");
    StatusString.append(filename);
    if (statusUpdateInterface != 0)
    {
        statusUpdateInterface->Update(StatusString);
    }
}

//--------------------------------------------------------------------------------
void
CBacktraceDirectory::DeleteLastSeenLogFile()
{
    std::string lastSeenLogFilename(c_LastSeenLogFilename);
    m_directory.DeleteFile(lastSeenLogFilename);
}

//--------------------------------------------------------------------------------
void
CBacktraceDirectory::Reload()
{
    m_directory.Reload();
}

//--------------------------------------------------------------------------------
unsigned CBacktraceDirectory::GetNumberLogFiles()
{
    return GetNumberFilesWithExtension(c_LogFileExtension);
}

//--------------------------------------------------------------------------------
unsigned  CBacktraceDirectory::GetNumberFilesWithExtension(const char* extension)
{
    std::string  filename;
    unsigned     logFileCount = 0;
    
    // Create a regular expression to find all files ending in ".extension".
    std::string FileSpec = ".*\\.";
    FileSpec += extension;
    const boost::regex Regex(FileSpec);

    m_directory.ResetGetNextFilename();
    
    while (m_directory.GetNextFilename(filename))
    {
        if (regex_match(filename.c_str(), Regex))
        {
            ++logFileCount;
        }
    }

    return logFileCount;
}

bool CBacktraceDirectory::_CheckSystemStatus(int /*Status*/, const char* /*Command*/)
{
    bool Success = false;

//    if (!WIFEXITED(Status))
//    {
//        if (WIFSIGNALED(Status))
//        {
//            RAYERROR("Failed to %s, killed by signal %s", Command, strsignal(WTERMSIG(Status)));
//        }
//        else if (WIFSTOPPED(Status))
//        {
//            RAYERROR("Failed to %s, stopped by signal %s", Command, strsignal(WSTOPSIG(Status)));
//        }
//        else
//        {
//            RAYERROR("Failed to %s, unknown issue - raw status %d", Command, Status);
//        }
//    }
//    else if (WEXITSTATUS(Status) != EXIT_SUCCESS)
//    {
//        RAYERROR("Failed to %s, return code %d", Command, WEXITSTATUS(Status));
//    }
//    else
//    {
//        Success = true;
//    }

    return Success;
}

bool CBacktraceDirectory::CreateDirectoryRecursive(const std::string& directory)
{
    bool result = false;
    struct stat st;
    if (!(stat(directory.c_str(), &st) == 0 && S_ISDIR(st.st_mode)))
    {
        size_t pos = directory.find_first_of('/', 0);
        std::string path;
        do
        {
            path = directory.substr(0, pos != std::string::npos ? pos + 1 : std::string::npos);
            struct stat st;
            if (stat(path.c_str(), &st) != 0)
            {
                if (mkdir(path.c_str(), S_IRWXU) != 0)
                {
                    RAYERROR("Failed to mkdir %s", path.c_str());
                    break;
                }
            }

            if (pos < directory.length() - 1)
            {
                pos = directory.find_first_of('/', pos + 1);
            }
        }
        while (directory != path);

        result = directory == path;
    }
    return result;
}
