/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef BACKTRACEDIRECTORY_H
#define BACKTRACEDIRECTORY_H

#include "FrameworkCommon/POSIXDirectory.h"
#include <string>

//--------------------------------------------------------------------------------
// BacktraceDirectory
//
//  This is responsible for managing the backtrace directory. This provides
//  methods for:-
 //
 //     +   ensuring that there are sufficient entries available to
//          add a new crash dump.
//
//      +   listing the logs currently in the backtrace folder.
//
//      +   detecting if a new log has been created since the last run.
//
//  Note this is implemented in the basebrick singleton style as it the
//  crash support is setup before RAYCOM has been initialised.
//
class CBacktraceDirectory
{
public:

    CBacktraceDirectory();
    virtual ~CBacktraceDirectory();

    //
    // The backtrace directory has to be initialised before it can be used.
    //
    static void Init(const char* backtraceLogFolder);

    //--------------------------------------------------------------------------
    // Reload backtrace folder, this should not be necessary, but just in case
    // ha.
    //
    void Reload();

    unsigned GetNumberLogFiles();

    //
    // The status update interface provides a method of giving textual information
    // on the progress of compact flash operations. The interface was created to
    // avoid coupling this class to the diagnostic's module.
    //
    class IStatusUpdate;

    int CopySysLogToCard(const std::string& DestinationFolder, IStatusUpdate* statusUpdateInterface = 0);

    int MoveLogsFolderContentToCard(const std::string& DestinationFolder, IStatusUpdate* statusUpdateInterface = 0);

    void ClearLogsFolder();

    class IStatusUpdate
    {
        public:
            virtual ~IStatusUpdate(){}
            virtual void Update(const std::string& information) = 0;
    };

private:

    static std::string  m_BacktraceDirectory;

    CPOSIXDirectory m_directory;
    std::string     m_logFileBaseName;

    bool            OpenOrCreateBacktraceDirectory();

    unsigned        GetNumberFilesWithExtension(const char* extension);

    bool _CheckSystemStatus(int, const char*);

    //--------------------------------------------------------------------------
    // SD card support.
    //
    bool MoveFile(const std::string& SrcFilename, const std::string& DestDir, IStatusUpdate*  statusUpdateInterface);
    bool CopyFile(const std::string& SrcFilename, const std::string& DestDir, IStatusUpdate*  statusUpdateInterface);

    std::string     BuildSDFilename(std::string& filename, int);

    void            DeleteLastSeenLogFile();

    void ShowFilenameInStatus(const std::string& filename,
                              CBacktraceDirectory::IStatusUpdate* statusUpdateInterface);

    static bool CreateDirectoryRecursive(const std::string& directory);

    //--------------------------------------------------------------------------
    // Disallow object copying by providing a private copy constuctor and
    // assignment operator.
    //
    CBacktraceDirectory(const CBacktraceDirectory& rOther);

    CBacktraceDirectory&
    operator = (const CBacktraceDirectory& rOther);
};

#endif
