//-------------------------------------------------------------------------//
// Copyright (c) Raymarine UK Limited 2013
//
// Reproduction or transmission in whole or in part (whether by photocopying or
// storing in any medium by electronic means or otherwise) without the written
// permission of Raymarine UK Limited is prohibited.
//
// Confidential
//-------------------------------------------------------------------------//

#include "CrashLogConfiguration.h"

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <assert.h>

#include <boost/filesystem.hpp>
#include <boost/regex.hpp>

#include "OMAPCrash.h"
#ifdef _PROCESS_USES_OSS
    #include "Framework/OSS/OSS.h"
#endif


// Create a convenient namespace to use boost::filesystem.
namespace bfs = boost::filesystem;

//
// Special file used to indicate whether a core dump should be generated or not.
// Note: the connects of the file are ignored.
//
static const char DO_CORE_DUMP_INDICATOR_FILE[] = "/raymarine/USBDisk/DoDump.txt";

static const char c_LogFileExtension[]          = "log";

static const char c_ScreenDumpFileExtension[]   = "bmp";

// The DSM300 build does not include this file, so for clarity the coupling between this file
// and Backtrace.cpp is conditionally compiled
#ifdef __i386
// Global flag to indicate whether or not the backtrace should be obtained via glibc or Raymarine backtrace
bool g_UseRayX86Backtrace = false;
#endif

#ifdef _PROCESS_USES_OSS

static void TaskEntryCallback(OSS_Task_t* task)
{
    eCrash_RegisterThread(task->pName, 0);
}

static void TaskExitCallback(OSS_Task_t* /* task */)
{
    // eCash unregisters the current thread/task.
    eCrash_UnregisterThread();
}

#endif // _PROCESS_USES_OSS

CCrashLogConfiguration::CCrashLogConfiguration()
{
}

CCrashLogConfiguration::~CCrashLogConfiguration()
{
}

void CCrashLogConfiguration::Init(
    const char*     backtraceLogFolder,
    const char*     logFileBaseName,
    const char*     applicationName,
    const char*     softwareVersionString,
    unsigned int    MaxNumBacktraceLogs,
    unsigned int    MaxNumScreenDumps,
    CrashLoggerOutputFunctionPtr_t    pLNetDiagnosticsOutput,
    CrashLoggerOutputFunctionPtr_t    pFunctionTraceOutput
)
{
    static eCrashParameters crashParams;
    bzero(&crashParams, sizeof(crashParams));
    
#ifdef __i386
    //
    // Determine whther to use glibc backtrace or not.
    // NOTE this is done in Init rather than at point of crash
    // so as to reduce dynamic memory usage during crash scenario.
    //
    if (getenv("USE_RAYMARINE_X86_BACKTRACE") != 0)
    {
        g_UseRayX86Backtrace = true;
    }
#endif

#ifdef _PROCESS_USES_OSS

    //
    // Register eCrash with OSS for thread entry and exit callbacks.
    //
    OSS_TaskRegisterEntryCallback(TaskEntryCallback);
    OSS_TaskRegisterExitCallback(TaskExitCallback);

#endif // _PROCESS_USES_OSS

    // Get the next log file
    std::string NextLogFilename;
    std::string NextScreenDumpFilename;
    GetNextFilename(backtraceLogFolder, logFileBaseName, c_LogFileExtension, MaxNumBacktraceLogs, NextLogFilename);

    if (NextLogFilename.empty())
    {
        crashParams.filename = 0;
        crashParams.filep = stderr;
    }
    else
    {
        crashParams.filename = NextLogFilename.c_str();
        crashParams.filep = 0;
    }
    
    // Test if a screen dump are required.
    // Screen dumps are probably only relevant to the mfdapp process.
    if (MaxNumScreenDumps > 0)
    {
        GetNextFilename(backtraceLogFolder, logFileBaseName, c_ScreenDumpFileExtension, MaxNumScreenDumps, NextScreenDumpFilename);

        if (NextScreenDumpFilename.empty())
        {
            crashParams.windowDumpFile = 0;
        }
        else
        {
            crashParams.windowDumpFile = NextScreenDumpFilename.c_str();
        }
    }
    else
    {
        crashParams.windowDumpFile = 0;
    }

    crashParams.doCoreDumpIndicatorFile = DO_CORE_DUMP_INDICATOR_FILE;
    crashParams.fd                      = -1; // open(backtraceFilename.c_str(), O_WRONLY|O_CREAT, S_IREAD|S_IWRITE|S_IRGRP|S_IROTH); // 0644

    crashParams.debugLevel              = ECRASH_DEBUG_DEFAULT;
    crashParams.dumpAllThreads          = true;
    crashParams.maxStackDepth           = 50;
    crashParams.defaultBacktraceSignal  = 0; // use default.
    crashParams.threadWaitTime          = 4000; // 4 seconds

    // Should we delay before the final abort?
    if (getenv("DELAY_BEFORE_ABORT") != 0)
    {
        crashParams.delayBeforeFinalAbort = true;
    }
    else
    {
        crashParams.delayBeforeFinalAbort = false;
    }

    crashParams.applicationName         = applicationName;
    crashParams.softwareVersionString   = softwareVersionString;
    crashParams.symbolTable             = 0;
    crashParams.signals[0]              = SIGSEGV;
    crashParams.signals[1]              = SIGFPE;
    crashParams.signals[2]              = SIGILL;
    crashParams.signals[3]              = SIGBUS;
    crashParams.signals[4]              = SIGSYS;
    crashParams.signals[5]              = SIGABRT;

    crashParams.pLNetDiagnosticsOutputFunction = pLNetDiagnosticsOutput;

    crashParams.pFunctionTraceOutputFunction = pFunctionTraceOutput;

    int ecrashResult = eCrash_Init(&crashParams);
    assert(ecrashResult == 0);
#ifdef NDEBUG
    (void)ecrashResult;
#endif
}

void CCrashLogConfiguration::DeInit()
{

#ifdef _PROCESS_USES_OSS

    //
    // Unregister eCrash with OSS for thread entry and exit callbacks.
    //
    OSS_TaskUnRegisterEntryCallback(TaskEntryCallback);
    OSS_TaskUnRegisterExitCallback(TaskExitCallback);

#endif // _PROCESS_USES_OSS

}

void CCrashLogConfiguration::GetNextFilename(const char* pLogDir,
                                             const char* pBase,
                                             const char* pExtension,
                                             unsigned int MaxNumFiles,
                                             std::string& NextFilename)
{
    // Create a regular expression to match the required files.
    //
    // Example filename: MFDAPP_0013.log
    //
    // where 'MFDAPP' is the specified base string and 'log' is the specified extension.
    //
    // The number field always has 4 digits giving a range from '0000' to '9999'.
    //
    const std::string BaseStr(pBase);
    const std::string ExtensionStr(pExtension);
    const std::string FileSpec = BaseStr + "_\\d{4}\\." + ExtensionStr;
    const boost::regex Regex(FileSpec);

    // Iterate over all files in the log directory.
    bfs::directory_iterator Iter(pLogDir);
    bfs::directory_iterator EndIter;

    // Create a list of file names with matching base and extension strings.
    std::list<std::string> FileNameList;

    while (Iter != EndIter)
    {
        // Test if this entry is a file.
        if (!bfs::is_directory(*Iter))
        {
#if BOOST_FILESYSTEM_VERSION == 3
            if (boost::regex_match((*Iter).path().filename().string(), Regex))
            {
                // Add this filename to the list.
                FileNameList.push_back((*Iter).path().filename().string());
            }
#else
            if (boost::regex_match((*Iter).filename(), Regex))
            {
                // Add this filename to the list.
                FileNameList.push_back((*Iter).filename());
            }
#endif
        }

        ++Iter;
    }

    // Sort the filename list.
    // As the only difference between filenames is the number this should sort by number.
    FileNameList.sort();

    // Remove files until there is one less than the maximum number of files.
    // In theory there should never be more than the maximum but we handle this
    // situation anyway.
    while (FileNameList.size() >= MaxNumFiles)
    {
        // Find the file in the log directory.
        bfs::directory_iterator DelIter(pLogDir);

        bool FileFound = false;

        while (!FileFound && (DelIter != EndIter))
        {
            // Test if this is a file.
            if (bfs::is_regular_file(*DelIter))
            {
                // Compare the filename with the lowest number in the list.
#if BOOST_FILESYSTEM_VERSION == 3
                if ((*DelIter).path().filename().string() == FileNameList.front())
#else
                if ((*DelIter).filename() == FileNameList.front())
#endif
                {
                    // Delete the file from the log directory.
                    bfs::remove(bfs::path(*DelIter));

                    // Delete the filename from the list.
                    FileNameList.erase(FileNameList.begin());

                    FileFound = true;
                }
            }

            ++DelIter;
        }
    }

    // Default the number string to "0000" in case there are no matching files.
    std::string NumberStr("0000");

    if (!FileNameList.empty())
    {
        // Get the filename with the largest number.
        std::string Filename(FileNameList.back());

        // Extract the filename number.
        size_t DotPosition = Filename.find('.');
        size_t NumberStartPosition = DotPosition - 4;
        std::string Number = Filename.substr(NumberStartPosition, 4);
        int FileNumber = -1;
        sscanf(Number.c_str(), "%10d", &FileNumber);

        // Get the next number as a string.
        char CharBuffer[5];
        sprintf(CharBuffer, "%04d", FileNumber + 1);
        NumberStr = CharBuffer;
    }

    // Build and return the next filename (including path).
    NextFilename = pLogDir;
    NextFilename += "/" + BaseStr + "_" + NumberStr + "." + ExtensionStr;
}

