/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

//-------------------------------------------------------------------------//
// INCLUDED FILES
//-------------------------------------------------------------------------//

#include "EventLogger.h"
#include "GetThreadId.h"

#ifdef ENABLE_MEMORY_EVENT_LOGGER
#include "MemoryEventLogger.h"
#endif

#include <pthread.h>
#include <time.h>
#include <stddef.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdbool.h>
#ifdef __ANDROID__
#include <QCoreApplication>
#include <android/log.h>
#endif

#ifdef ENABLE_NETWORK_LOGGING
#include <netinet/in.h>
#include <sys/socket.h>
#include <netinet/udp.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#endif

//-------------------------------------------------------------------------//
// LOCAL FUNCTIONS
//-------------------------------------------------------------------------//

// Used to strip the full path of the __FILE__ macro down to the file name.
static const char* ELAFileName(const char* pFilePath);

// Used to turn the debug level integer into an appropriate string.
static const char* ELADebugLevelString(int Level);

// Used to check the given level and names against the debug filter.
// The final two parameters are outputs specifying if the filter applies
// to syslog or stderr outputs (or both).
static bool ELADebugFilterPassed(unsigned int Level, const char* pProcessName,
                                 const char* pFileName, const char* pPrettyFunction,
                                 bool* pSyslog, bool* pStderr);

// Used to update the debug filter when necessary.
static void ELAUpdateDebugFilter();

// Used to parse the actual debug filter file.
static void ELAParseDebugFilterFile(FILE* pFile);

// Used to check each individual field of the filter against a value.
static bool ELATestFilter(const char* pFilterField, const char* pActualValue);

static void ELAGetProcessName(char* pProcessName, size_t BufLen);

static bool ELAUpdateStatistics(unsigned int Level);

#ifdef ENABLE_NETWORK_LOGGING
static bool LogToNetworkInitialised = false;
static bool LogToNetworkEnabled = false;
static void ELALogToNetwork(unsigned int Level, const char* pFileName,
                int LineNumber, const char* pFunctionName, const char* pFormatString,
                va_list arg );
#ifdef TARGET_ANDROID
static const char* c_NetLoggingConfigFile = "/data/data/com.raymarine.axiom.mfd/files/Data/NetLogging.conf";
static const char* c_CardNetLoggingConfigFile = "/mnt/internal_slot1/NetLogging.conf";
static const char* c_Card2NetLoggingConfigFile = "/mnt/external0_slot1/NetLogging.conf";
#else
static const char* c_NetLoggingConfigFile = "/raymarine/Data/NetLogging.conf";
static const char* c_CardNetLoggingConfigFile = "/tmp/mnt/internal_slot1/NetLogging.conf";
static const char* c_Card2NetLoggingConfigFile = "/tmp/mnt/external0_slot1/NetLogging.conf";
#endif
static const size_t c_MaxAddressSize = 32;
static char NetLoggingAddress[c_MaxAddressSize];
static short int NetLoggingPort = 0;
static unsigned int NetLoggingMaxLevel = 0u;
#endif

//-------------------------------------------------------------------------//
// LOCAL TYPES
//-------------------------------------------------------------------------//

// Used for debug filters on a file and method level.
typedef struct
{
    char Process[30];           // The process to apply the debug level to.
    char File[30];              // The file to apply the debug level to.
    char Method[30];            // The method to apply the debug level to.
    unsigned int MaximumLevel;  // The debug level to apply in the filter.
    bool Syslog;                // True if the filter applies to syslog.
    bool Stderr;                // True if the filter applies to stderr.
} DebugFilter_t;

//-------------------------------------------------------------------------//
// LOCAL CONSTANTS
//-------------------------------------------------------------------------//

// Specifies the name of the debug filter configuration file.
#ifndef __ANDROID__
static const char* c_DebugFilterConfigFile = "/raymarine/Data/DebugLevel.conf";
#else
static const char* c_DebugFilterConfigFile = "/data/data/com.raymarine.axiom.mfd/files/Data/DebugLevel.conf";
#endif

// Specifies how often the ELAUpdateDebugFilter function will read the file.
static const int c_DebugFilterUpdateInterval_Seconds = 10;

// Specifies the maximum number of debug filters for this process.
static const int c_MaxDebugFilters = 10;

// Specifies the maximum size of the array to use for sprintf debug output.
#ifdef PRODUCT_ATOM
static const int c_MaxDebugStringLength = 256;
#else
static const int c_MaxDebugStringLength = 1024;
#endif

// Specifies the maximum length of the process name
static const int c_MaxProcessName = 32;

// Specifies the default maximum log level for stderr with no config file.
static const unsigned int c_DefaultStderrLogLevel = RAYLOG_LEVEL_DEBUG;

// Specifies the default maximum log level for syslog with no config file.
static const unsigned int c_DefaultSyslogLogLevel = RAYLOG_LEVEL_INFO;

//-------------------------------------------------------------------------//
// LOCAL DATA
//-------------------------------------------------------------------------//

static unsigned int LogLevelCounters[c_MaxDebugFilters];

// The debug filters, periodically parsed from c_DebugFilterConfigFile.
static DebugFilter_t DebugFilters[c_MaxDebugFilters];

// Used to update the debug filters periodically.
static int LastFilterUpdateTime = 0;

// Used to filter at the top level for performance.
static unsigned int GlobalMaximumLevel = 0;

// We need a mutex to protect ELADebugEx, as it is called from every thread.
static pthread_mutex_t ELAMutex = PTHREAD_MUTEX_INITIALIZER;

#ifndef __ANDROID__
static unsigned int StartSeconds = 0u;
#endif
//-------------------------------------------------------------------------//
// PUBLIC FUNCTIONS
//-------------------------------------------------------------------------//

void ELADebugBasic(const char* pFullFileName, int LineNumber, const char* pFunctionName, const char* pFormatString, ...)
{
    if (pthread_mutex_lock(&ELAMutex) == 0)
    {
#ifdef ENABLE_MEMORY_EVENT_LOGGER
        {
            va_list ArgumentList;
            va_start(ArgumentList, pFormatString);
            MemoryEventLogger::LogToMemoryBuffer(pFullFileName, LineNumber, pFunctionName, pFormatString, ArgumentList);
            va_end(ArgumentList);
        }
#endif

        static char ProcessName[c_MaxProcessName] = {0};
        static bool GotProcessName = false;

        const char* pFileName = ELAFileName(pFullFileName);

        if (!GotProcessName)
        {
            GotProcessName = true;
            ELAGetProcessName(ProcessName, sizeof(ProcessName));
        }

        char DebugString[c_MaxDebugStringLength];

        // Decode variable argument list to produce the output debugging string.
        {
            va_list ArgumentList;
            va_start(ArgumentList, pFormatString);
            vsnprintf(DebugString, c_MaxDebugStringLength, pFormatString, ArgumentList);
            va_end(ArgumentList);
        }

        // Null Terminate Buffer in case of max buffer size write
        DebugString[c_MaxDebugStringLength - 1] = 0;

        fprintf(stderr, "%s [%d:%d] (%s:%d - %s): %s\n", ProcessName,
                getpid(),  GetThreadId(),
                pFileName, LineNumber, pFunctionName, DebugString);
        fflush(stderr);

        pthread_mutex_unlock(&ELAMutex);
    }
}

bool ELADebugEx(unsigned int Level, const char* pFullFileName, int LineNumber,
                const char* pFunctionName, const char* pFormatString, ...)
{
    bool Result = true;
#ifdef ENABLE_NETWORK_LOGGING
    {
        if (!LogToNetworkInitialised)
        {
            LogToNetworkInitialised = true;
            FILE* pFile = fopen(c_CardNetLoggingConfigFile, "r");
            if (pFile == NULL)
            {
                pFile = fopen(c_Card2NetLoggingConfigFile, "r");
            }

            if (pFile == NULL)
            {
                pFile = fopen(c_NetLoggingConfigFile, "r");
            }

            if (pFile != NULL)
            {
                char Address[c_MaxAddressSize];
                int Port = 0;
                int DebugLevel = 0;
                if (fscanf(pFile, "%30s %d %d\n", Address, &Port, &DebugLevel) == 3)
                {
                    strcpy(NetLoggingAddress, Address);
                    NetLoggingPort = (short int)Port;
                    NetLoggingMaxLevel = DebugLevel;
                    LogToNetworkEnabled = true;
                }
                fclose(pFile);
            }
        }

        if (LogToNetworkEnabled && Level <= NetLoggingMaxLevel)
        {
            va_list ArgumentList;
            va_start(ArgumentList, pFormatString);
            ELALogToNetwork(Level, pFullFileName, LineNumber, pFunctionName, pFormatString, ArgumentList);
            va_end(ArgumentList);
        }
    }
#endif

#ifdef ENABLE_MEMORY_EVENT_LOGGER
    {
        va_list ArgumentList;
        va_start(ArgumentList, pFormatString);
        MemoryEventLogger::LogToMemoryBuffer(pFullFileName, LineNumber, pFunctionName, pFormatString, ArgumentList);
        va_end(ArgumentList);
    }
#endif

    // For performance, RAYDEBUG or lower we can skip, based on the global.
    if (Level <= c_DefaultStderrLogLevel || Level <= GlobalMaximumLevel)
    {
        static char ProcessName[c_MaxProcessName] = {0};
        static bool GotProcessName = false;

        if (pthread_mutex_lock(&ELAMutex) == 0)
        {
            ELAUpdateDebugFilter();

            ELAUpdateStatistics(Level);

            // Second level of skipping, now that the filter update has happened.
            if (Level <= GlobalMaximumLevel)
            {
                const char* pFileName = ELAFileName(pFullFileName);

                if (!GotProcessName)
                {
                    ELAGetProcessName(ProcessName, sizeof(ProcessName));
                    GotProcessName = true;
                }
                bool Syslog = false;
                bool Stderr = false;

                if (ELADebugFilterPassed(Level, ProcessName, pFileName, pFunctionName, &Syslog, &Stderr))
                {
                    char DebugString[c_MaxDebugStringLength];

                    // Decode variable argument list to produce the output debugging string.
                    va_list ArgumentList;
                    va_start(ArgumentList, pFormatString);
                    vsnprintf(DebugString, c_MaxDebugStringLength, pFormatString, ArgumentList);
                    va_end(ArgumentList);

                    // Null Terminate Buffer in case of max buffer size write
                    DebugString[c_MaxDebugStringLength - 1] = 0;

#ifdef __ANDROID__
                    if (Syslog)
                    {
                        android_LogPriority androidPriority = ANDROID_LOG_UNKNOWN;
                        switch (Level)
                        {
                        case RAYLOG_LEVEL_FATAL:
                            androidPriority = ANDROID_LOG_FATAL;
                            break;

                        case RAYLOG_LEVEL_ERROR:
                            androidPriority = ANDROID_LOG_ERROR;
                            break;

                        case RAYLOG_LEVEL_WARNING:
                            androidPriority = ANDROID_LOG_WARN;
                            break;

                        case RAYLOG_LEVEL_INFO:
                            androidPriority = ANDROID_LOG_INFO;
                            break;

                        case RAYLOG_LEVEL_DEBUG:
                            androidPriority = ANDROID_LOG_DEBUG;
                            break;

                        case RAYLOG_LEVEL_DETAIL:
                        case RAYLOG_LEVEL_TRACE:
                            androidPriority = ANDROID_LOG_VERBOSE;
                            break;

                        default:
                            androidPriority = ANDROID_LOG_SILENT;
                            break;

                        }

                        if (androidPriority != ANDROID_LOG_SILENT && androidPriority != ANDROID_LOG_UNKNOWN)
                        {
                            __android_log_print(androidPriority, ProcessName, "%s:%d (%s): %s", pFileName, LineNumber, pFunctionName, DebugString);
                        }
                    }
#else

                    const char* pDebugLevelstring = ELADebugLevelString(Level);
                    if ((Level == RAYLOG_LEVEL_DETAIL) || (Level == RAYLOG_LEVEL_TRACE))
                    {
                        Level = LOG_DEBUG;  // Detail and trace don't exist in the syslog.
                    }

                    if (Syslog)
                    {
                        openlog(ProcessName, LOG_PID, LOG_USER);
                        syslog(Level, "[%s] %s:%d (%s): %s", pDebugLevelstring, pFileName,
                               LineNumber, pFunctionName, DebugString);
                        closelog();
                    }

                    if (Stderr)
                    {
                        struct timespec ts;
                        unsigned int seconds = 0u;
                        unsigned int microsecs = 0u;
                        if (clock_gettime(CLOCK_MONOTONIC, &ts) == 0)
                        {
                            if (StartSeconds == 0u)
                            {
                                StartSeconds = static_cast<unsigned int>(ts.tv_sec);
                            }
                            seconds = static_cast<unsigned int>(ts.tv_sec) - StartSeconds;
                            microsecs = static_cast<unsigned int>(ts.tv_nsec) / 1000u;
                        }
                        if (seconds != 0u && microsecs != 0u)
                        {
                            fprintf(stderr, "%u.%06u: %s (%d:%d): (%s) %s:%d (%s): %s\n", seconds, microsecs,
                                    ProcessName, getpid(), GetThreadId(),
                                    pDebugLevelstring, pFileName, LineNumber,
                                    pFunctionName, DebugString);
                        }
                        else
                        {
                            fprintf(stderr, "%s (%d:%d): (%s) %s:%d (%s): %s\n", ProcessName,
                                    getpid(),  GetThreadId(),
                                    pDebugLevelstring, pFileName, LineNumber,
                                    pFunctionName, DebugString);
                       }
                        fflush(stderr);
                    }
#endif
                }
            }

            if (pthread_mutex_unlock(&ELAMutex) != 0)
            {
                openlog(ProcessName, LOG_PID, LOG_USER);
                syslog(Level, "[%s] %s:%d (%s): Failed to unlock mutex in %s",
                       ELADebugLevelString(Level), ELAFileName(pFullFileName), LineNumber,
                       pFunctionName, __FUNCTION__);
                closelog();

                fprintf(stderr, "%s [%d]: [%s] %s:%d (%s): Failed to unlock mutex in %s\n",
                        ProcessName, getpid(), ELADebugLevelString(Level), ELAFileName(pFullFileName),
                        LineNumber, pFunctionName, __FUNCTION__);
                fflush(stderr);

                RAYASSERT(false);  // This should never happen, as we only use the mutex here.

                Result = false;
            }
        }
        else
        {
            if (!GotProcessName)
            {
                ELAGetProcessName(ProcessName, sizeof(ProcessName));
                GotProcessName = true;
            }

            openlog(ProcessName, LOG_PID, LOG_USER);
            syslog(Level, "[%s] %s:%d (%s): Failed to lock mutex in %s",
                   ELADebugLevelString(Level), ELAFileName(pFullFileName),
                   LineNumber, pFunctionName, __FUNCTION__);
            closelog();

            fprintf(stderr, "%s [%d]: [%s] %s:%d (%s): Failed to lock mutex in %s\n",
                    ProcessName, getpid(), ELADebugLevelString(Level), ELAFileName(pFullFileName),
                    LineNumber, pFunctionName, __FUNCTION__);
            fflush(stderr);

            RAYASSERT(false);  // This should never happen, as we only use the mutex here.

            Result = false;
        }
    }

    return Result;
}

//-------------------------------------------------------------------------//
// LOCAL FUNCTIONS
//-------------------------------------------------------------------------//

static const char* ELAFileName(const char* pFilePath)
{
    // strip off any directories before the file name
    const char* pFileName = strrchr(pFilePath, '/');
    if (pFileName == 0)
    {
        return pFilePath;
    }
    else
    {
        return pFileName + 1;
    }
}

static const char* ELADebugLevelString(int Level)
{
    const char* pDebugLevelString = "????";

    switch (Level)
    {
    case RAYLOG_LEVEL_FATAL:   pDebugLevelString = "FATAL";   break;
    case RAYLOG_LEVEL_ERROR:   pDebugLevelString = "ERROR";   break;
    case RAYLOG_LEVEL_WARNING: pDebugLevelString = "WARNING"; break;
    case RAYLOG_LEVEL_NOTICE:  pDebugLevelString = "NOTICE";  break;
    case RAYLOG_LEVEL_INFO:    pDebugLevelString = "INFO";    break;
    case RAYLOG_LEVEL_DEBUG:   pDebugLevelString = "DEBUG";   break;
    case RAYLOG_LEVEL_DETAIL:  pDebugLevelString = "DETAIL";  break;
    case RAYLOG_LEVEL_TRACE:   pDebugLevelString = "TRACE";   break;
    default: break;
    }

    return pDebugLevelString;
}

static bool ELADebugFilterPassed(unsigned int Level, const char* pProcessName,
                                 const char* pFileName, const char* pPrettyFunction,
                                 bool* pSyslog, bool* pStderr)
{
    *pStderr = true;
    *pSyslog = true;

    int CurrentFilter;

    // Check the user defined filters.
    for (CurrentFilter = 0; CurrentFilter < c_MaxDebugFilters; ++CurrentFilter)
    {
        if (ELATestFilter(DebugFilters[CurrentFilter].Process, pProcessName) &&
            ELATestFilter(DebugFilters[CurrentFilter].File, pFileName) &&
            ELATestFilter(DebugFilters[CurrentFilter].Method, pPrettyFunction))
        {
            if (DebugFilters[CurrentFilter].Syslog)
            {
                *pSyslog = (Level <= DebugFilters[CurrentFilter].MaximumLevel);
            }
            if (DebugFilters[CurrentFilter].Stderr)
            {
                *pStderr = (Level <= DebugFilters[CurrentFilter].MaximumLevel);
            }
        }
    }

    return *pSyslog || *pStderr;
}

static void ELAUpdateDebugFilter()
{
    struct timespec tp;
    clock_gettime(CLOCK_MONOTONIC, &tp);

    int ElapsedTime = abs(tp.tv_sec - LastFilterUpdateTime);

    if (ElapsedTime > c_DebugFilterUpdateInterval_Seconds)
    {
        LastFilterUpdateTime = tp.tv_sec;

        FILE* pFile = fopen(c_DebugFilterConfigFile, "r");
        if (pFile != 0)
        {
            ELAParseDebugFilterFile(pFile);
            fclose(pFile);
        }
        else
        {
            // No config file, so just use the defaults.
            memset(DebugFilters, 0, sizeof(DebugFilters));

            DebugFilters[0].Process[0] = '*';
            DebugFilters[0].File[0] = '*';
            DebugFilters[0].Method[0] = '*';
            DebugFilters[0].Stderr = true;
            DebugFilters[0].Syslog = false;
            DebugFilters[0].MaximumLevel = c_DefaultStderrLogLevel;

            DebugFilters[1].Process[0] = '*';
            DebugFilters[1].File[0] = '*';
            DebugFilters[1].Method[0] = '*';
            DebugFilters[1].Stderr = false;
            DebugFilters[1].Syslog = true;
            DebugFilters[1].MaximumLevel = c_DefaultSyslogLogLevel;
        }
    }

    if (GlobalMaximumLevel == 0)
    {
        for (int i = 0; i < c_MaxDebugFilters; i++)
        {
#ifndef RELEASE
            bool Suitable = DebugFilters[i].Stderr || DebugFilters[i].Syslog;
#else
            // No point in going to stderr if it won't be seen by anyone...
            bool Suitable = DebugFilters[i].Syslog;
#endif

            if (Suitable && DebugFilters[i].MaximumLevel > GlobalMaximumLevel)
            {
                GlobalMaximumLevel = DebugFilters[i].MaximumLevel;
            }
        }
    }
}

static void ELAParseDebugFilterFile(FILE* pFile)
{
    // Throw out all the filters. Good job we have a mutex locked in ELADebugEx.
    memset(DebugFilters, 0, sizeof(DebugFilters));

    unsigned int NewGlobal = 0;
    GlobalMaximumLevel = 0;
    if (fscanf(pFile, "Global: %1u\n", &NewGlobal) == 1)
    {
        GlobalMaximumLevel = NewGlobal;
    }

    int CurrentFilter = 0;
    char SyslogOrStderr[7];

    do
    {
        memset(SyslogOrStderr, 0, sizeof(SyslogOrStderr));
        int Found = fscanf(pFile, "%29[^.].%29[^.].%29s %1u %6s\n",
                           DebugFilters[CurrentFilter].Process,
                           DebugFilters[CurrentFilter].File,
                           DebugFilters[CurrentFilter].Method,
                           &DebugFilters[CurrentFilter].MaximumLevel,
                           SyslogOrStderr);
        if (Found != 5)
        {
            break;
        }

        // Anything other than "syslog" or "stderr" defaults to both.
        DebugFilters[CurrentFilter].Syslog = true;
        DebugFilters[CurrentFilter].Stderr = true;
        if (strncmp(SyslogOrStderr, "syslog", 6) == 0)
        {
            DebugFilters[CurrentFilter].Stderr = false;
        }
        else if (strncmp(SyslogOrStderr, "stderr", 6) == 0)
        {
            DebugFilters[CurrentFilter].Syslog = false;
        }
    } while (++CurrentFilter < c_MaxDebugFilters);
}

static bool ELATestFilter(const char* pFilterField, const char* pActualValue)
{
    bool Match = false;

    if ((pFilterField != 0) && (pFilterField[0] != '\0') &&
        (pActualValue != 0) && (pActualValue[0] != '\0'))
    {
        if ((pFilterField[0] == '*') || (strstr(pActualValue, pFilterField) != 0))
        {
            Match = true;
        }
    }

    return Match;
}

static void ELAGetProcessName(char* pProcessName, size_t BufLen)
{
#ifdef __ANDROID__
    strncpy(pProcessName, QCoreApplication::applicationName().toLatin1().data(), BufLen -1);
    pProcessName[BufLen - 1] = 0;
#else
    char Self[1024];
    int Length = readlink("/proc/self/exe", Self, sizeof(Self) - 1);
    if (Length >= 0)
    {
        Self[Length] = 0;
        const char* pBackslash = strrchr(Self, '/');
        if (pBackslash != 0)
        {
            strncpy(pProcessName, pBackslash + 1, BufLen - 1);
        }
        else
        {
            strncpy(pProcessName, Self, BufLen - 1);
        }
        pProcessName[BufLen - 1] = 0;
    }
#endif
}

#ifdef ENABLE_NETWORK_LOGGING

static void ELALogToNetwork(unsigned int Level, const char* pFileName,
                int LineNumber, const char* pFunctionName, const char* pFormatString,
                va_list arg)
{
    static int NetworkLoggingSocket = -1;
    static struct sockaddr_in NetworkLoggingDestAddr;

    static char ProcessName[c_MaxProcessName] = {0};
    static bool GotProcessName = false;

    if (!GotProcessName)
    {
        ELAGetProcessName(ProcessName, sizeof(ProcessName));
        GotProcessName = true;
    }

    if (NetworkLoggingSocket == -1)
    {
        memset(&NetworkLoggingDestAddr, 0, sizeof(NetworkLoggingDestAddr));
        NetworkLoggingDestAddr.sin_family = AF_INET;
        NetworkLoggingDestAddr.sin_port = htons(NetLoggingPort);
        NetworkLoggingDestAddr.sin_addr.s_addr = inet_addr(NetLoggingAddress);

        NetworkLoggingSocket = socket(PF_INET, SOCK_DGRAM, 0);

        if (NetworkLoggingSocket != -1)
        {
            // Socket created OK
            // Set the file descriptor to not be inherited by child proceses
            int Flags = fcntl(NetworkLoggingSocket, F_GETFD);
            Flags |= FD_CLOEXEC;
            fcntl(NetworkLoggingSocket, F_SETFD, Flags);
        }
    }

    if (NetworkLoggingSocket != -1)
    {
        const int c_MaxDebugLength = 1024;
        char DebugString[c_MaxDebugLength];

        // Decode variable argument list to produce the output debugging string.
        vsnprintf(DebugString, c_MaxDebugLength, pFormatString, arg);

        // Null Terminate Buffer in case of max buffer size write
        DebugString[c_MaxDebugLength - 1] = 0;

        char DebugPacket[c_MaxDebugLength];
        snprintf(DebugPacket, c_MaxDebugLength, "[%s] [%s] %s:%d (%s): %s", ProcessName, ELADebugLevelString(Level), ELAFileName(pFileName), LineNumber, pFunctionName, DebugString);

        sendto(NetworkLoggingSocket, DebugPacket, strlen(DebugPacket) + 1, 0, (sockaddr *)&NetworkLoggingDestAddr, sizeof(NetworkLoggingDestAddr));
    }
}
#endif

bool ELAUpdateStatistics(unsigned int level)
{
    bool outputStatistics = false;

    if ((level < c_MaxDebugFilters) && (level <= RAYLOG_LEVEL_WARNING))
    {
        LogLevelCounters[level]++;
    }

    return outputStatistics;
}


unsigned int ELAGetLogLevelCounter(unsigned int level)
{
    unsigned int count = 0;

    if (pthread_mutex_lock(&ELAMutex) == 0)
    {
        if ((level < c_MaxDebugFilters) && (level <= RAYLOG_LEVEL_WARNING))
        {
            count = LogLevelCounters[level];
        }
    }
    pthread_mutex_unlock(&ELAMutex);

    return count;
}

