/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef LOGGER_H_INCLUDED
#define LOGGER_H_INCLUDED

#include "RayAssert.h"

#include <syslog.h>
#include <stdio.h>
#include <stdbool.h>
#include <stdint.h>

static const uint32_t c_SyncCallTimeWarning = 1000;

#ifdef RAY_UNIT_TEST
# ifndef _NO_DIAGNOSTICS
#  define _NO_DIAGNOSTICS
# endif
#endif

// Functions use C calling convention an are un-mangled to allow them to be
// called from 'C' modules.
// As this file has to compile under C, we cannot use fully qualified names for
// types etc. scoped into a namespace so we must explicity use all required namespaces.
#ifdef __cplusplus
//using namespace EventLogTypes;
extern "C"
{
#endif

/// This function is used to build a basic debug string from the format string
/// and parameters. It is used by processes whose output is swallowed by CProcessHelper
/// in the MFD - it will retransmit output to stderr, but that output will already
/// be in a RAYDEBUG, so we don't need all the extra bits.
///
/// \param[in]  pFileName       The name of the file, from __FILE__.
/// \param[in]  LineNumber      The line number, from __LINE__.
/// \param[in]  pFunctionName   The name of the function, from __PRETTY_FUNCTION__.
/// \param[in]  pFormatString   The string to ouput, with a printf-style list of
///                             substitution arguments following it.
void ELADebugBasic(const char* pFileName, int LineNumber, const char* pFunctionName, const char* pFormatString, ...);

/// This function is used to build a debug string from the format string
/// and parameters and to forward it to the event log.
///
/// \param[in]  Level           The level of the output to log.
/// \param[in]  pFileName       The name of the file, from __FILE__.
/// \param[in]  LineNumber      The line number, from __LINE__.
/// \param[in]  pFunctionName   The name of the function, from __PRETTY_FUNCTION__.
/// \param[in]  pFormatString   The string to ouput, with a printf-style list of
///                             substitution arguments following it.
///
/// \return Success of failure.
bool ELADebugEx(unsigned int Level, const char* pFileName,
                int LineNumber, const char* pFunctionName, const char* pFormatString,
                ...) /*__attribute((__format__(printf, 5, 6)))*/;

unsigned int ELAGetLogLevelCounter(unsigned int level);

#ifdef __cplusplus
}
#endif


# define RAYLOG_LEVEL_TRACE     9
# define RAYLOG_LEVEL_DETAIL    8
# define RAYLOG_LEVEL_DEBUG     LOG_DEBUG
# define RAYLOG_LEVEL_INFO      LOG_INFO
# define RAYLOG_LEVEL_NOTICE    LOG_NOTICE
# define RAYLOG_LEVEL_WARNING   LOG_WARNING
# define RAYLOG_LEVEL_ERROR     LOG_ERR
# define RAYLOG_LEVEL_FATAL     LOG_CRIT
# define RAYLOG_LEVEL_OFF       0

#ifndef _NO_DIAGNOSTICS
// The printf ensures the format string and arguments are error checked,
// but should not impact on code size or performance. The extra space
// avoids an annoying warning that occurs when using the attribute above.

#ifndef RAYLOG_INHIBIT_FORMAT_CHECK
#define RAYLOG_CHECK_FORMAT(format, ...) while (false) { printf(" " format, ##__VA_ARGS__); }
#else
#define RAYLOG_CHECK_FORMAT(format, ...)
#endif

#ifdef REMOVE_NONCRITICAL_DIAGNOSTICS

# define RAYTRACE(format, ...)
# define RAYDETAIL(format, ...)
# define RAYDEBUG(format, ...)
# define RAYINFO(format, ...)    do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_INFO,    __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYNOTICE(format, ...)  do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_NOTICE,  __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYWARNING(format, ...) do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_WARNING, __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYERROR(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_ERROR,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYFATAL(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_FATAL,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); RAYASSERT(false && "RAYFATAL: " format); } while (false) // NOLINT
# define RAYASSERT_LOG(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_FATAL,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT

#else

# define RAYTRACE(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_TRACE,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYDETAIL(format, ...)  do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_DETAIL,  __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYINFO(format, ...)    do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_INFO,    __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYNOTICE(format, ...)  do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_NOTICE,  __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYWARNING(format, ...) do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_WARNING, __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# define RAYERROR(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_ERROR,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT

// Processes run through CProcessHelper need only basic debugging, as it swallows and re-outputs their logging.
# ifdef RAY_PROCESS_HELPER
#  define RAYDEBUG(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugBasic(__FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
#  define RAYFATAL(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugBasic(__FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); RAYASSERT(false && "RAYFATAL: " format); } while (false) // NOLINT
#  define RAYASSERT_LOG(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugBasic(__FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# else
#  define RAYDEBUG(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_DEBUG,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
#  define RAYFATAL(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_FATAL,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); RAYASSERT(false && "RAYFATAL: " format); } while (false) // NOLINT
#  define RAYASSERT_LOG(format, ...)   do { RAYLOG_CHECK_FORMAT(format, ##__VA_ARGS__); ELADebugEx(RAYLOG_LEVEL_FATAL,   __FILE__, __LINE__, __FUNCTION__, format, ##__VA_ARGS__); } while (false) // NOLINT
# endif

#endif

#else // _NO_DIAGNOSTICS

# define RAYTRACE(format, ...)
# define RAYDETAIL(format, ...)
# define RAYDEBUG(format, ...)
# define RAYINFO(format, ...)
# define RAYNOTICE(format, ...)
# define RAYWARNING(format, ...)
# define RAYERROR(format, ...)
# define RAYFATAL(format, ...)

#endif // _NO_DIAGNOSTICS

#endif // LOGGER_H_INCLUDED
