/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "IpInfo.h"
#include "RayAssert.h"
#include "EventLogger.h"
#include "Framework/PlatformAbstraction/TCPIP/RaySocketTypes.h"

#include <unistd.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/if.h>
#include <net/if_arp.h>
#include <unistd.h>
#include <linux/sockios.h>
#include <linux/ethtool.h>
#include <cstring>
#include <cassert>
#include <cstdio>
#include <fstream>

//#define _IPINFO_DEBUG

static constexpr char const* c_LoopbackErr = "Error - we don't have loopback interface. Check your network card config.";
static constexpr char const* c_InterfaceConfigErr = "Error - you must have both IP addresses on the same interface (198.18.x.x/21 and 10.x.x.x/11). "
                                                   "(ip address add [ip]/[mask-digits] dev [nic]). Check your network card config.";

#ifdef __ANDROID__
static const std::string c_WifiClientInterface = "wlan0";
#else
static const std::string c_WifiClientInterface = "wlan1";
#endif

// for interface up and configured verification
static const unsigned int c_PollMaxRetries = 30;
static const unsigned int c_UpdatePollRetries = 10;
static const unsigned int c_PollSleepUsec = 1000000;

const int CIpInfo::c_markValue = 1234; // SO_MARK value

CIpInfo::CIpInfo() :
    m_IpAddress(0),
    m_NetMask(0),
    m_LegacyNetworkIpAddress(0),
    m_LegacyNetworkNetMask(0),
    m_LoopbackIpAddress(0),
    m_LoopbackNetMask(0),
    m_WirelessIpAddress(0),
    m_WirelessNetMask(0),
    m_DefaultGateway(0),
    m_InterfaceName(""),
    m_LegacyNetworkInterfaceName(""),
    m_LoopbackInterfaceName(""),
    m_isLegacyFirst(false)
//    m_UpdateIntInfoFD(-1),
//    m_GetNetStatusFD(-1)
{
    _PopulateBasicNetworkInfoRetry(c_PollMaxRetries);

    FILE* file = fopen("/proc/net/route", "r");
    RAYASSERT(file);

    if (file)
    {
        unsigned long int d, g, m;
        int flgs, ref, use, metric, mtu, win, ir;

        if (fscanf(file, "%*[^\n]\n") >= 0)
        {
            while (feof(file) == 0)
            {
                char devname[64];
                // Skip the first line
                int count = fscanf(file, "%63s%10lx%10lx%10X%11d%11d%11d%10lx%11d%11d%11d\n",
                            devname, &d, &g, &flgs, &ref, &use, &metric, &m, &mtu, &win, &ir);

                if (count == 11)
                {
                    if (d == 0)
                    {
                        struct in_addr  sin_addr;
                        sin_addr.s_addr = g;
                        m_DefaultGateway = ntohl(g);

                        // This is the default gateway
                        char Buffer[INET_ADDRSTRLEN] = {0};
                        const char* result = inet_ntop(AF_INET, &sin_addr, Buffer, sizeof(Buffer));
                        RAYASSERT(result);
                        if (result != 0)
                        {
                            m_DefaultGatewayString = result;
                        }
                    }
                }
            }
        }

        fclose(file);
    }

    const char* c_WirelessIpAddress = "192.168.0.1";
    const char* c_WirelessSubnetMask = "255.255.255.0";

    m_WirelessIPAddressString = c_WirelessIpAddress;
    m_WirelessNetMaskString = c_WirelessSubnetMask;

    unsigned int FlippedAddress = 0u;
    if (inet_pton(AF_INET, c_WirelessIpAddress, &FlippedAddress) == 0)
    {
        RAYWARNING("Failed to convert %s to inet form", c_WirelessIpAddress);
    }
    else
    {
        m_WirelessIpAddress = ntohl(FlippedAddress);
        RAYTRACE("Wireless IP Address %s converted to 0x%08x", c_WirelessIpAddress, m_WirelessIpAddress);
    }

    if (inet_pton(AF_INET, c_WirelessSubnetMask, &FlippedAddress) == 0)
    {
        RAYWARNING("Failed to convert %s to inet form", c_WirelessIpAddress);
    }
    else
    {
        m_WirelessNetMask = ntohl(FlippedAddress);
        RAYTRACE("Wireless Subnet mask %s converted to 0x%08x", c_WirelessSubnetMask, m_WirelessNetMask);
    }

    std::ifstream addressFile(std::string("/sys/class/net/") + c_WifiClientInterface + "/address");
    if (addressFile.is_open())
    {
        std::getline(addressFile, m_WirelessMacAddressString);
    }

    RAYINFO("Using: IP=%s 0x%08x NetMask=%s 0x%08x; Legacy IP=%s 0x%08x NetMask=%s 0x%08x; Gateway=%s MAC=%s",
            m_IPAddressString.c_str(), m_IpAddress,
            m_NetMaskString.c_str(), m_NetMask,
            m_LegacyNetworkIPAddressString.c_str(), m_LegacyNetworkIpAddress,
            m_LegacyNetworkNetMaskString.c_str(), m_LegacyNetworkNetMask,
            m_DefaultGatewayString.c_str(),
            m_MacAddressString.c_str());
}

CIpInfo::~CIpInfo()
{
//    close(m_UpdateIntInfoFD);
//    close(m_GetNetStatusFD);
}

#if !defined(PLATFORM_WORKSTATION)

bool CIpInfo::_PollInterfaceAvailability(int fd, std::string interfaceName, unsigned int pollRetries, bool* pConfigured)
{
    bool found = false;
    struct ifreq ifr;
    *pConfigured = false;

    if ((fd > 0) && !interfaceName.empty())
    {
        unsigned int retries = 0;
        bool interfaceUp = false;

        // check for interface to be up
        while (!interfaceUp && (retries < pollRetries))
        {
            retries++;

            memset(&ifr, 0, sizeof(struct ifreq));
            strncpy(ifr.ifr_name, interfaceName.c_str(), IFNAMSIZ - 1);
            found = (ioctl(fd, SIOCGIFFLAGS, &ifr) == 0);
            if (!found)
            {
                break; // while (!interfaceUp && retries < pollRetries)
            }
            interfaceUp = (ifr.ifr_flags & IFF_UP);
            if (!interfaceUp)
            {
                usleep(c_PollSleepUsec);
            }
        }

        // check for interface to be configured
        if (interfaceUp)
        {
            // do not reset counter, just continue from wher it was left off
            while (!(*pConfigured) && (retries < pollRetries))
            {
                memset(&ifr, 0, sizeof(struct ifreq));
                strncpy(ifr.ifr_name, interfaceName.c_str(), IFNAMSIZ - 1);
                found = (ioctl(fd, SIOCGIFADDR, &ifr) == 0);
                if (!found)
                {
                    break; // while (!(*pConfigured) && (retries < pollRetries))
                }
                *pConfigured = ((reinterpret_cast<struct sockaddr_in*>(&(ifr.ifr_addr))->sin_addr.s_addr) != 0);
                if (!(*pConfigured))
                {
                    usleep(c_PollSleepUsec);
                }
            }
        }
    }
    return found;
}


bool CIpInfo::UpdateInterfaceInfo(RaySocketTypes::EthernetDevice_t deviceType)
{
    bool ifrOk = false;
    bool updated = false;

    std::string interfaceName;
    uint32_t* pIpAddress;
    uint32_t* pNetMask;
    std::string* pIPAddressString;
    std::string* pNetMaskString;

    if ((RaySocketTypes::e_DevWired == deviceType) || (RaySocketTypes::e_DevWiredLegacy == deviceType))
    {
        if (RaySocketTypes::e_DevWired == deviceType)
        {
            interfaceName = m_InterfaceName;
            pIpAddress = &m_IpAddress;
            pNetMask = &m_NetMask;
            pIPAddressString = &m_IPAddressString;
            pNetMaskString = &m_NetMaskString;
        }
        else
        {
            interfaceName = m_LegacyNetworkInterfaceName;
            pIpAddress = &m_LegacyNetworkIpAddress;
            pNetMask = &m_LegacyNetworkNetMask;
            pIPAddressString = &m_LegacyNetworkIPAddressString;
            pNetMaskString = &m_LegacyNetworkNetMaskString;
        }

        if (!interfaceName.empty())
        {
//            if (m_UpdateIntInfoFD == -1)
//            {
//                m_UpdateIntInfoFD = socket(AF_INET, SOCK_DGRAM, 0);
//                RAYASSERT(m_UpdateIntInfoFD > 0);
//            }
            int fd = socket(AF_INET, SOCK_DGRAM, 0);
            RAYASSERT(fd > 0);

            _PollInterfaceAvailability(fd, interfaceName, c_UpdatePollRetries, &ifrOk);

            if (ifrOk)
            {
                int err = 0;
                struct ifconf ifc;
                struct ifreq *pifr;

                _initializeIfConf(fd, &ifc);
                RAYASSERT(ifc.ifc_buf != NULL);
                pifr = ifc.ifc_req;

                int ifc_num = ifc.ifc_len / sizeof(struct ifreq);

                // Iterate trought all ifreq and verify ip changed
                for (int i = 0;  i < ifc_num; ++i)
                {
                    // Get IP Address
                    err = ioctl(fd, SIOCGIFADDR, &pifr[i]);
                    RAYASSERT(err == 0);
                    uint32_t ipAddress = ntohl(reinterpret_cast<struct sockaddr_in*>(&(pifr[i].ifr_addr))->sin_addr.s_addr);

                    // Get Subnet Mask
                    err = ioctl(fd, SIOCGIFNETMASK, &pifr[i]);
                    RAYASSERT(err == 0);
                    uint32_t netmaskAddress = ntohl(reinterpret_cast<struct sockaddr_in*>(&(pifr[i].ifr_addr))->sin_addr.s_addr);

                    if (RaySocketTypes::e_DevWired == deviceType && m_IpAddress != ipAddress
                            && (ipAddress & netmaskAddress) == c_RaymarineSThsNetworkAddress
                             && netmaskAddress == c_RaymarineSThsNetworkNetMask)
                    {
                        updated = true;
                        RAYDEBUG("Interface %s has SThs ip !", pifr[i].ifr_name);
                    }
                    else if (RaySocketTypes::e_DevWiredLegacy == deviceType && m_LegacyNetworkIpAddress != ipAddress
                             && (ipAddress & netmaskAddress) == c_RaymarineSThsLegacyNetworkAddress
                             && netmaskAddress == c_RaymarineSThsLegacyNetworkNetMask)
                    {
                        updated = true;
                        RAYDEBUG("Interface %s has legacy ip !", pifr[i].ifr_name);
                    }

                    if (updated)
                    {
                        *pIpAddress = ipAddress;
                        _IpAddressToString(*pIpAddress, pIPAddressString);

                        *pNetMask = netmaskAddress;
                        _IpAddressToString(*pNetMask, pNetMaskString);

                        break;
                    }
                }

                if (close(fd) == -1)
                {
                    RAYERROR("Close socket %d failed! Error: %d", fd, errno);
                }

                free(ifc.ifc_buf);
            }


        }
    }
    else
    {
        RAYWARNING("Wired Interface %s unknown", interfaceName.c_str());
    }

    return updated;
}
#endif

// we assume that ipAddress is in host byte order
void CIpInfo::_IpAddressToString(uint32_t ipAddress, std::string* pString)
{
    char buffer[INET_ADDRSTRLEN] = {0};
    uint32_t ipAddressNetwork = htonl(ipAddress); // inet_ntop wants network byte order
    const char* result = inet_ntop(AF_INET, &ipAddressNetwork, buffer, sizeof(buffer));
    RAYASSERT(result);
    *pString = result;
}


bool CIpInfo::_PopulateBasicNetworkInfoRetry(int nbRetries)
{
    bool success = false;

    // Try to update network infor for nbRetries times
    while(nbRetries > 0 && success == false) {
        success = _PopulateBasicNetworkInfo(nbRetries == 1 ? true : false);
        nbRetries--;
        // wait before next request
        usleep(c_PollSleepUsec);
    }

    return success;
}

// Don't forget to release memory of ifc->ifc_buf
void CIpInfo::_initializeIfConf(int fd, struct ifconf *ifc)
{
    int BufferSize = 10;
    int err = 0;

    ifc->ifc_buf = 0;
    ifc->ifc_len = sizeof(struct ifreq) * BufferSize;

    while ((err >= 0) && (ifc->ifc_len == static_cast<int>((sizeof(struct ifreq) * BufferSize))))
    {
        ifc->ifc_len = sizeof(struct ifreq) * BufferSize;
        ifc->ifc_buf = (char*) realloc(ifc->ifc_buf, ifc->ifc_len);

        err = ioctl(fd, SIOCGIFCONF, ifc);

        if ((err >= 0) && (ifc->ifc_len == static_cast<int>(sizeof(struct ifreq) * BufferSize)))
        {
            // assume it overflowed so increase buffer
            BufferSize += 10;
        }
    }
}

#ifdef TARGET_DESKTOP
bool CIpInfo::_PopulateBasicNetworkInfo(bool abort)
{
    (void)abort;
    bool success = false;
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    RAYASSERT(fd > 0);

    int err = 0;
    struct ifconf ifc;
    struct ifreq *pifr;

    _initializeIfConf(fd, &ifc);
    RAYASSERT(ifc.ifc_buf != NULL);
    pifr = ifc.ifc_req;

    int sthsNetworkInterfaceIndex = -1;
    int legacyNetworkInterfaceIndex = -1;
    int loopBackInterfaceIndex = -1;
    int hwIndex = -1;
    int ifc_num = ifc.ifc_len / sizeof(struct ifreq);
    InterfaceInfo interfaceInfo[ifc_num];

    // Iterate trought all ifreq and save interface type, ip and netmask
    for (int i = 0;  i < ifc_num; ++i)
    {
        // Get IP Address
        err = ioctl(fd, SIOCGIFADDR, &pifr[i]);
        RAYASSERT(err == 0);
        interfaceInfo[i].ipAddress = ntohl(reinterpret_cast<struct sockaddr_in*>(&(pifr[i].ifr_addr))->sin_addr.s_addr);

        // Get Subnet Mask
        err = ioctl(fd, SIOCGIFNETMASK, &pifr[i]);
        RAYASSERT(err == 0);
        interfaceInfo[i].netmaskAddress = ntohl(reinterpret_cast<struct sockaddr_in*>(&(pifr[i].ifr_addr))->sin_addr.s_addr);

        // Flags - For loopback
        err = ioctl(fd, SIOCGIFFLAGS, &pifr[i]);
        RAYASSERT(err == 0);

        // -----  debug  -----
        std::string IPAddressString;
        std::string NetMaskString;

        _IpAddressToString(interfaceInfo[i].ipAddress, &IPAddressString);
        _IpAddressToString(interfaceInfo[i].netmaskAddress, &NetMaskString);

        RAYDEBUG("Network Interface Found : %s, IP=%s, NetMask=%s", pifr[i].ifr_name, IPAddressString.c_str(), NetMaskString.c_str());
        // ----- end debug -----

        if (pifr[i].ifr_flags & IFF_LOOPBACK)
        {
            interfaceInfo[i].interfaceType = RaySocketTypes::e_DevLoopback;
            loopBackInterfaceIndex = i;
            RAYDEBUG("%s is loopback interface !", pifr[i].ifr_name);
        }
        else if (static_cast<uint32_t>(interfaceInfo[i].ipAddress & interfaceInfo[i].netmaskAddress) == c_RaymarineSThsNetworkAddress
                 && interfaceInfo[i].netmaskAddress == c_RaymarineSThsNetworkNetMask)
        {
            sthsNetworkInterfaceIndex = i;
            interfaceInfo[i].interfaceType = RaySocketTypes::e_DevWired;

            RAYDEBUG("Interface %s ip new subnet !", pifr[i].ifr_name);

            if (legacyNetworkInterfaceIndex != -1 && strcmp(pifr[i].ifr_name, pifr[legacyNetworkInterfaceIndex].ifr_name) != 0)
            {
                legacyNetworkInterfaceIndex = -1;
            }
        }
        else if (static_cast<uint32_t>(interfaceInfo[i].ipAddress & interfaceInfo[i].netmaskAddress) == c_RaymarineSThsLegacyNetworkAddress
                 && interfaceInfo[i].netmaskAddress == c_RaymarineSThsLegacyNetworkNetMask) // set it once and then just skip, usually first is the ethernet
        {
            interfaceInfo[i].interfaceType = RaySocketTypes::e_DevWiredLegacy;

            legacyNetworkInterfaceIndex = i;

            RAYDEBUG("Interface %s ip legacy !", pifr[i].ifr_name);

            if (sthsNetworkInterfaceIndex != -1 && strcmp(pifr[i].ifr_name, pifr[sthsNetworkInterfaceIndex].ifr_name) != 0)
            {
                sthsNetworkInterfaceIndex = -1;
            }
        }
    }

    // If we have loopback iface save data into class members (ip, netmask, iface)
    if (loopBackInterfaceIndex != -1)
    {
        m_LoopbackInterfaceName = pifr[loopBackInterfaceIndex].ifr_name;
        m_LoopbackIpAddress = interfaceInfo[loopBackInterfaceIndex].ipAddress;
        m_LoopbackNetMask = interfaceInfo[loopBackInterfaceIndex].netmaskAddress;
        _IpAddressToString(m_LoopbackIpAddress, &m_LoopbackIPAddressString);
        _IpAddressToString(m_LoopbackNetMask, &m_LoopbackNetMaskString);

        RAYDEBUG("Loopback Network Interface Found : %s, IP=%s, NetMask=%s",
                 m_LoopbackInterfaceName.c_str(), m_LoopbackIPAddressString.c_str(), m_LoopbackNetMaskString.c_str());

        hwIndex = loopBackInterfaceIndex;
    }

    // If we have both ip's save data into class members (ip, netmask, iface)
    if (sthsNetworkInterfaceIndex != -1 && legacyNetworkInterfaceIndex != -1)
    {
        m_InterfaceName = pifr[sthsNetworkInterfaceIndex].ifr_name;
        m_IpAddress = interfaceInfo[sthsNetworkInterfaceIndex].ipAddress;
        m_NetMask = interfaceInfo[sthsNetworkInterfaceIndex].netmaskAddress;
        _IpAddressToString(m_IpAddress, &m_IPAddressString);
        _IpAddressToString(m_NetMask, &m_NetMaskString);

        RAYDEBUG("SThs Network Interface set as : %s, IP=%s, NetMask=%s",
                 m_InterfaceName.c_str(), m_IPAddressString.c_str(), m_NetMaskString.c_str());

        m_LegacyNetworkInterfaceName = pifr[legacyNetworkInterfaceIndex].ifr_name;
        m_LegacyNetworkIpAddress = interfaceInfo[legacyNetworkInterfaceIndex].ipAddress;
        m_LegacyNetworkNetMask = interfaceInfo[legacyNetworkInterfaceIndex].netmaskAddress;
        _IpAddressToString(m_LegacyNetworkIpAddress, &m_LegacyNetworkIPAddressString);
        _IpAddressToString(m_LegacyNetworkNetMask, &m_LegacyNetworkNetMaskString);

        RAYDEBUG("Legacy Network Interface set as : %s, IP=%s, NetMask=%s",
                 m_LegacyNetworkInterfaceName.c_str(), m_LegacyNetworkIPAddressString.c_str(), m_LegacyNetworkNetMaskString.c_str());

        hwIndex = sthsNetworkInterfaceIndex;
    }

    // Get the MAC Address
    if (hwIndex != -1)
    {
        err = ioctl(fd, SIOCGIFHWADDR, &pifr[hwIndex]);
        RAYASSERT(err == 0);

        memcpy(m_MacAddress.Address, pifr[hwIndex].ifr_ifru.ifru_hwaddr.sa_data, MAC_ADDRESS_LENGTH);

        const int c_MacAddressStringLen = 64;
        char MacAddressString[c_MacAddressStringLen];
        snprintf(MacAddressString, c_MacAddressStringLen, "%02x:%02x:%02x:%02x:%02x:%02x",
                 m_MacAddress.Address[0], m_MacAddress.Address[1], m_MacAddress.Address[2],
                m_MacAddress.Address[3], m_MacAddress.Address[4], m_MacAddress.Address[5]);
        m_MacAddressString = MacAddressString;

        RAYDEBUG("MACAddress = %s", m_MacAddressString.c_str());
    }

    // done
    free(ifc.ifc_buf);
    close(fd);

    success = true;

    return success;
}
#else
bool CIpInfo::_PopulateBasicNetworkInfo(bool abort)
{
    bool success = false;
    int fd = socket(AF_INET, SOCK_DGRAM, 0);
    RAYASSERT(fd > 0);

    int err = 0;
    struct ifconf ifc;
    struct ifreq *pifr;

    _initializeIfConf(fd, &ifc);
    RAYASSERT(ifc.ifc_buf != NULL);
    pifr = ifc.ifc_req;

    int sthsNetworkInterfaceIndex = -1;
    int legacyNetworkInterfaceIndex = -1;
    int loopBackInterfaceIndex = -1;
    int hwIndex = -1;
    int ifc_num = ifc.ifc_len / sizeof(struct ifreq);
    InterfaceInfo interfaceInfo[ifc_num];

    int countNewSubnetAddresses = 0;

    // Iterate trought all ifreq and save interface type, ip and netmask
    for (int i = 0;  i < ifc_num; ++i)
    {
        // Get IP Address
        err = ioctl(fd, SIOCGIFADDR, &pifr[i]);
        RAYASSERT(err == 0);
        interfaceInfo[i].ipAddress = ntohl(reinterpret_cast<struct sockaddr_in*>(&(pifr[i].ifr_addr))->sin_addr.s_addr);


        // Get Subnet Mask
        err = ioctl(fd, SIOCGIFNETMASK, &pifr[i]);
        RAYASSERT(err == 0);
        interfaceInfo[i].netmaskAddress = ntohl(reinterpret_cast<struct sockaddr_in*>(&(pifr[i].ifr_addr))->sin_addr.s_addr);

        // Flags - For loopback
        err = ioctl(fd, SIOCGIFFLAGS, &pifr[i]);
        RAYASSERT(err == 0);

        // -----  debug  -----
        std::string IPAddressString;
        std::string NetMaskString;

        _IpAddressToString(interfaceInfo[i].ipAddress, &IPAddressString);
        _IpAddressToString(interfaceInfo[i].netmaskAddress, &NetMaskString);

        RAYINFO("Network Interface Found : %s, IP=%s, NetMask=%s", pifr[i].ifr_name, IPAddressString.c_str(), NetMaskString.c_str());
        // ----- end debug -----

        if (pifr[i].ifr_flags & IFF_LOOPBACK)
        {
            interfaceInfo[i].interfaceType = RaySocketTypes::e_DevLoopback;
            loopBackInterfaceIndex = i;
            RAYDEBUG("%s is loopback interface !", pifr[i].ifr_name);
        }
        else if ((interfaceInfo[i].ipAddress & interfaceInfo[i].netmaskAddress) == c_RaymarineSThsNetworkAddress
                 && interfaceInfo[i].netmaskAddress == c_RaymarineSThsNetworkNetMask)
        {
            interfaceInfo[i].interfaceType = RaySocketTypes::e_DevWired;
            RAYDEBUG("Interface %s has SThs ip !", pifr[i].ifr_name);
            countNewSubnetAddresses++;
        }
        else if ((interfaceInfo[i].ipAddress & interfaceInfo[i].netmaskAddress) == c_RaymarineSThsLegacyNetworkAddress
                 && interfaceInfo[i].netmaskAddress == c_RaymarineSThsLegacyNetworkNetMask)
        {
            interfaceInfo[i].interfaceType = RaySocketTypes::e_DevWiredLegacy;
            RAYDEBUG("Interface %s has legacy ip !", pifr[i].ifr_name);
            if (countNewSubnetAddresses <= 0 && !m_isLegacyFirst)
            {
                m_isLegacyFirst = true;
            }
        }
    }

    RAYDEBUG("Is legacy address listed first? -> %d", static_cast<int>(m_isLegacyFirst));

    // Loop trought interfaceType vector and verify if we have both legacy and SThs ip's
    for (int i = 0; i < ifc_num && sthsNetworkInterfaceIndex == -1; i++)
    {
        if (interfaceInfo[i].interfaceType == RaySocketTypes::e_DevWired)
        {
            for (int j = 0; j < ifc_num; j++)
            {
                if (interfaceInfo[j].interfaceType == RaySocketTypes::e_DevWiredLegacy
                    && strcmp(pifr[i].ifr_name, pifr[j].ifr_name) == 0)
                {
                    sthsNetworkInterfaceIndex = i;
                    legacyNetworkInterfaceIndex = j;
                    RAYDEBUG("Interface %s has SThs and legacy ip !", pifr[i].ifr_name);
                    break;
                }
            }
        }
    }

    // If we have loopback iface save data into class members (ip, netmask, iface)
    if (loopBackInterfaceIndex != -1)
    {
        m_LoopbackInterfaceName = pifr[loopBackInterfaceIndex].ifr_name;
        m_LoopbackIpAddress = interfaceInfo[loopBackInterfaceIndex].ipAddress;
        m_LoopbackNetMask = interfaceInfo[loopBackInterfaceIndex].netmaskAddress;
        _IpAddressToString(m_LoopbackIpAddress, &m_LoopbackIPAddressString);
        _IpAddressToString(m_LoopbackNetMask, &m_LoopbackNetMaskString);

        RAYDEBUG("Loopback Network Interface Found : %s, IP=%s, NetMask=%s",
                 m_LoopbackInterfaceName.c_str(), m_LoopbackIPAddressString.c_str(), m_LoopbackNetMaskString.c_str());

        hwIndex = loopBackInterfaceIndex;
    }

    // If we have both ip's save data into class members (ip, netmask, iface)
    if (sthsNetworkInterfaceIndex != -1 && legacyNetworkInterfaceIndex != -1)
    {
        m_InterfaceName = pifr[sthsNetworkInterfaceIndex].ifr_name;
        m_IpAddress = interfaceInfo[sthsNetworkInterfaceIndex].ipAddress;
        m_NetMask = interfaceInfo[sthsNetworkInterfaceIndex].netmaskAddress;
        _IpAddressToString(m_IpAddress, &m_IPAddressString);
        _IpAddressToString(m_NetMask, &m_NetMaskString);

        RAYDEBUG("SThs Network Interface Found : %s, IP=%s, NetMask=%s",
                 m_InterfaceName.c_str(), m_IPAddressString.c_str(), m_NetMaskString.c_str());

        m_LegacyNetworkInterfaceName = pifr[legacyNetworkInterfaceIndex].ifr_name;
        m_LegacyNetworkIpAddress = interfaceInfo[legacyNetworkInterfaceIndex].ipAddress;
        m_LegacyNetworkNetMask = interfaceInfo[legacyNetworkInterfaceIndex].netmaskAddress;
        _IpAddressToString(m_LegacyNetworkIpAddress, &m_LegacyNetworkIPAddressString);
        _IpAddressToString(m_LegacyNetworkNetMask, &m_LegacyNetworkNetMaskString);

        RAYDEBUG("Legacy Network Interface Found : %s, IP=%s, NetMask=%s",
                 m_LegacyNetworkInterfaceName.c_str(), m_LegacyNetworkIPAddressString.c_str(), m_LegacyNetworkNetMaskString.c_str());

        hwIndex = sthsNetworkInterfaceIndex;
    }

#if defined(PRODUCT_HORNET)
    // Set Loopback interface for Hornet
    for (int i = 0; i < ifc_num; i++)
    {
        if (interfaceInfo[i].interfaceType == RaySocketTypes::e_DevLoopback)
        {
            hwIndex = loopBackInterfaceIndex = i;

            m_InterfaceName = m_LegacyNetworkInterfaceName = pifr[i].ifr_name;
            m_IpAddress = m_LegacyNetworkIpAddress = interfaceInfo[i].ipAddress;
            m_NetMask = m_LegacyNetworkNetMask = interfaceInfo[i].netmaskAddress;

            _IpAddressToString(m_LegacyNetworkIpAddress, &m_LegacyNetworkIPAddressString);
            _IpAddressToString(m_LegacyNetworkNetMask, &m_LegacyNetworkNetMaskString);

            m_IPAddressString = m_LegacyNetworkIPAddressString;
            m_NetMaskString = m_LegacyNetworkNetMaskString;


            RAYDEBUG("SThs Network Interface Found : %s, IP=%s, NetMask=%s",
                     m_InterfaceName.c_str(), m_IPAddressString.c_str(), m_NetMaskString.c_str());

            RAYDEBUG("Legacy Network Interface Found : %s, IP=%s, NetMask=%s",
                     m_LegacyNetworkInterfaceName.c_str(), m_LegacyNetworkIPAddressString.c_str(), m_LegacyNetworkNetMaskString.c_str());

            break;
        }
    }
#endif

    // Get the MAC Address
    if (hwIndex != -1)
    {
        err = ioctl(fd, SIOCGIFHWADDR, &pifr[hwIndex]);
        RAYASSERT(err == 0);

        memcpy(m_MacAddress.Address, pifr[hwIndex].ifr_ifru.ifru_hwaddr.sa_data, MAC_ADDRESS_LENGTH);

        const int c_MacAddressStringLen = 64;
        char MacAddressString[c_MacAddressStringLen];
        snprintf(MacAddressString, c_MacAddressStringLen, "%02x:%02x:%02x:%02x:%02x:%02x",
                 m_MacAddress.Address[0], m_MacAddress.Address[1], m_MacAddress.Address[2],
                m_MacAddress.Address[3], m_MacAddress.Address[4], m_MacAddress.Address[5]);
        m_MacAddressString = MacAddressString;

        RAYDETAIL("MACAddress = %s", m_MacAddressString.c_str());
    }

    // done
    free(ifc.ifc_buf);
    close(fd);

#if defined(PRODUCT_HORNET)
    if (loopBackInterfaceIndex == -1)
    {
         if (abort == true)
         {
             RAYFATAL("%s", c_LoopbackErr);
         }
         else
         {
             RAYDEBUG("%s", c_LoopbackErr);
         }
    }
    else
    {
        success = true;
    }
#else
    if ((m_IpAddress & m_NetMask) != c_RaymarineSThsNetworkAddress || (m_LegacyNetworkIpAddress & m_LegacyNetworkNetMask) != c_RaymarineSThsLegacyNetworkAddress)
    {
        // The IP addr on the interface is not 198.18.x.x and 10.x.x.x
        if (abort == true)
        {
            RAYFATAL("%s", c_InterfaceConfigErr);
        }
        else
        {
            RAYDEBUG("%s", c_InterfaceConfigErr);
        }
    }
    else
    {
        success = true;
    }
#endif

    return success;
}
#endif

bool CIpInfo::isLegacyAddressFirst()
{
    return m_isLegacyFirst;
}

uint32_t CIpInfo::GetIpAddress(RaySocketTypes::EthernetDevice_t deviceType) const
{
    uint32_t ip = 0;


    switch (deviceType) {
    case RaySocketTypes::e_DevWired:
        ip = m_IpAddress;
        break;
    case RaySocketTypes::e_DevWiredLegacy:
        ip = m_LegacyNetworkIpAddress;
        break;
    case RaySocketTypes::e_DevLoopback:
        ip = m_LoopbackIpAddress;
        break;
    case RaySocketTypes::e_DevWireless:
        ip = m_WirelessIpAddress;
        break;
    case RaySocketTypes::e_Unknown:
    default:
        break;
    }
    return ip;
}

uint32_t CIpInfo::GetSubnetMask(RaySocketTypes::EthernetDevice_t deviceType) const
{
    uint32_t subnet = 0;

    switch (deviceType) {
    case RaySocketTypes::e_DevWired:
        subnet = m_NetMask;
        break;
    case RaySocketTypes::e_DevWiredLegacy:
        subnet = m_LegacyNetworkNetMask;
        break;
    case RaySocketTypes::e_DevLoopback:
        subnet = m_LoopbackNetMask;
        break;
    case RaySocketTypes::e_DevWireless:
        subnet = m_WirelessNetMask;
        break;
    case RaySocketTypes::e_Unknown:
    default:
        break;
    }

    return subnet;
}

const std::string CIpInfo::GetFormattedIPAddress(RaySocketTypes::EthernetDevice_t deviceType) const
{
    std::string ipStr("");

    switch (deviceType) {
    case RaySocketTypes::e_DevWired:
        ipStr = m_IPAddressString;
        break;
    case RaySocketTypes::e_DevWiredLegacy:
        ipStr = m_LegacyNetworkIPAddressString;
        break;
    case RaySocketTypes::e_DevLoopback:
        ipStr = m_LoopbackIPAddressString;
        break;
    case RaySocketTypes::e_DevWireless:
        ipStr = m_WirelessIPAddressString;
        break;
    case RaySocketTypes::e_Unknown:
    default:
        break;
    }

    return ipStr;
}

const std::string CIpInfo::GetFormattedSubnetMask(RaySocketTypes::EthernetDevice_t deviceType) const
{

    std::string subnetStr("");

    switch (deviceType) {
    case RaySocketTypes::e_DevWired:
        subnetStr = m_NetMaskString;
        break;
    case RaySocketTypes::e_DevWiredLegacy:
        subnetStr = m_LegacyNetworkNetMaskString;
        break;
    case RaySocketTypes::e_DevLoopback:
        subnetStr = m_LoopbackNetMaskString;
        break;
    case RaySocketTypes::e_DevWireless:
        subnetStr = m_WirelessNetMaskString;
        break;
    case RaySocketTypes::e_Unknown:
    default:
        break;
    }

    return subnetStr;
}

CIpInfo::e_NetworkStatus CIpInfo::GetNetworkStatus(bool useLegacyNetwork)
{
    e_NetworkStatus NetworkStatus = e_NetworkStatusLinkUnknown;

    // Get file descriptor.
//    if (m_GetNetStatusFD == -1)
//    {
//        m_GetNetStatusFD = socket(PF_INET, SOCK_DGRAM, 0);
//    }
    int fd = socket(PF_INET, SOCK_DGRAM, 0);

    if (fd != -1)
    {
        struct ifreq ifr;
        struct ethtool_value edata;

        memset(&ifr, 0, sizeof(ifr));

        // Specify which data we are interested in.
        edata.cmd = ETHTOOL_GLINK;

        // Default to link down if the ioctl() fails.
        edata.data = 0;

        // And copy name and data type to struct we are sending over.
        strncpy(ifr.ifr_name, (!useLegacyNetwork) ? m_InterfaceName.c_str() : m_LegacyNetworkInterfaceName.c_str(), IFNAMSIZ - 1);
        ifr.ifr_data = (char *) &edata;

        // Call the ioctl
        int err = ioctl(fd, SIOCETHTOOL, &ifr);

        if (err == 0)
        {
            NetworkStatus = edata.data ? e_NetworkStatusLinkUp : e_NetworkStatusLinkDown;
        }

        if (close(fd) == -1)
        {
            RAYERROR("Closing socket %d failed! Error: %d", fd, errno);
        }
    }

    return NetworkStatus;
}

void CIpInfo::GetMACAddress(MacAddress_t& rMACAddress) const
{
    memcpy(rMACAddress.Address, m_MacAddress.Address, sizeof(rMACAddress.Address));
}

CIpInfo::MacAddress_t::MacAddress_t()
{
    memset(Address, 0, sizeof(Address));
}

bool CIpInfo::isLegacyAddressIp(uint32_t tIpAddress)
{
    return (tIpAddress & c_RaymarineSThsLegacyNetworkNetMask) == c_RaymarineSThsLegacyNetworkAddress;
}

bool CIpInfo::isLegacyAddressIp(std::string tIpAddress)
{
    return (ntohl(inet_addr(tIpAddress.c_str())) & c_RaymarineSThsLegacyNetworkNetMask) == c_RaymarineSThsLegacyNetworkAddress;
}

bool CIpInfo::isNewAddressIp(uint32_t tIpAddress)
{
    return (tIpAddress & c_RaymarineSThsNetworkNetMask) == c_RaymarineSThsNetworkAddress;
}

bool CIpInfo::isNewAddressIp(std::string tIpAddress)
{
    return (ntohl(inet_addr(tIpAddress.c_str())) & c_RaymarineSThsNetworkNetMask) == c_RaymarineSThsNetworkAddress;
}

bool CIpInfo::isLocalhost(std::string tIpAddress)
{
    return tIpAddress.find("127.0") == 0;
}

void CIpInfo::apply10xNetworkFlagsOnSocket(int socketFd, std::string ipAddress, uint16_t port)
{
    if (socketFd != - 1 && !ipAddress.empty())
    {
        bool isLegacyIpToConnect = isLegacyAddressIp(ipAddress);
        if (isLegacyIpToConnect)
        {
            int val = CIpInfo::c_markValue;
            /* setting socket option to use MARK value */
            int e = setsockopt (socketFd, SOL_SOCKET, SO_MARK, &val, sizeof (val));
            RAYASSERT(e != -1);
        }

        //check if it is legacy or new subnet to bind to the right network
        if (isLegacyIpToConnect || isNewAddressIp(ipAddress))
        {
            int opt = 1;
            int e = setsockopt(socketFd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
            if (e < 0)
            {
                RAYERROR("Error set SO_REUSEADDR");
            }

            sockaddr_in localAddr;
            memset(&localAddr, 0, sizeof(localAddr));
            localAddr.sin_family = AF_INET;         /* Set family to internet */
            localAddr.sin_addr.s_addr = htonl(GetIpAddress(isLegacyIpToConnect ? RaySocketTypes::e_DevWiredLegacy : RaySocketTypes::e_DevWired));/* Set the local IP address */
            localAddr.sin_port = htons(port);

            int nRetBind = 0;
            nRetBind = bind(socketFd, (struct sockaddr*)&(localAddr), sizeof(localAddr));
            if (nRetBind < 0)
            {
                RAYERROR("Error bind ip addr to socket");
            }
        }
    }
}
