/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef IP_INFO_H_INCLUDED
#define IP_INFO_H_INCLUDED

#include <string>
#include <stdint.h>
#include <cstring>
#include "Framework/PlatformAbstraction/TCPIP/RaySocketTypes.h"

#define MAC_ADDRESS_LENGTH  6

static const uint32_t c_LoopBackNetworkAddress = 0x7F000000;
static const uint32_t c_RaymarineSThsNetworkAddress = 0xC6120000;
static const uint32_t c_RaymarineSThsNetworkNetMask = 0xFFFFF800;
static const uint32_t c_RaymarineSThsLegacyNetworkAddress = 0x0A000000;
static const uint32_t c_RaymarineSThsLegacyNetworkNetMask = 0xFFE00000;

class CIpInfo
{
    public:

        /// Structure to hold a 6 byte Ethernet MAC Address
        struct MacAddress_t
        {
            MacAddress_t();
            unsigned char Address[MAC_ADDRESS_LENGTH];
        };

        /// Enumeration of network status
        enum e_NetworkStatus
        {
            e_NetworkStatusLinkUp,      ///< Link up and working
            e_NetworkStatusLinkDown,    ///< Link down / disconnected
            e_NetworkStatusLinkUnknown  ///< Unable to determine link status
        };

        CIpInfo ();
        virtual ~CIpInfo();

        /// Get the Unit's wired / wireless IP address as a number in Host Byte Order
        /// @return IP address
        uint32_t GetIpAddress(RaySocketTypes::EthernetDevice_t deviceType = RaySocketTypes::e_DevWired) const;

        /// Get network link status
        /// @return link status
        e_NetworkStatus GetNetworkStatus(bool useLegacyNetwork = false);

#if !defined(PLATFORM_WORKSTATION)
        /// Update the Unit's wired / wireless IP address and Net Mask
        /// @return Interface found, up and configured
        bool UpdateInterfaceInfo(RaySocketTypes::EthernetDevice_t deviceType);
#endif

        /// Get the MAC address of the network interface as a number
        /// @param rMACAddress  Reference to the MAC address to be populated
        void GetMACAddress(MacAddress_t& rMACAddress) const;

        /// Get the wired / wireless subnet mask of the network interface as a number in Host Byte Order
        /// @return Net mask
        uint32_t GetSubnetMask(RaySocketTypes::EthernetDevice_t deviceType = RaySocketTypes::e_DevWired) const;

        /// Get the default gateway address as a number in Host Byte Order
        /// @return Default gateway
        uint32_t GetDefaultGateway() const
        {
            return m_DefaultGateway;
        }

        /// Get the IP address formatted as a string
        /// @return IP Address string
        const std::string GetFormattedIPAddress(RaySocketTypes::EthernetDevice_t deviceType = RaySocketTypes::e_DevWired) const;

        /// Get the IP subnet mask formatted as a string
        /// @return Subnet mask string
        const std::string GetFormattedSubnetMask(RaySocketTypes::EthernetDevice_t deviceType = RaySocketTypes::e_DevWired) const;

        /// Get the MAC address formatted as a string
        /// @return MAC Address string
        const std::string& GetFormattedMacAddress() const
        {
            return m_MacAddressString;
        }

        /// Get the Default Gateway address formatted as a string
        /// @return Default Gateway Address string
        const std::string& GetFormattedDefaultGateway() const
        {
            return m_DefaultGatewayString;
        }

        /// Get the WiFi Client MAC address formatted as a string
        /// @return MAC Address string
        const std::string& GetFormattedWiFiMacAddress() const
        {
            return m_WirelessMacAddressString;
        }


        /// Get the interface name
        /// @return interface name
        const std::string& GetInterfaceName() const
        {
            return m_InterfaceName;
        }

        bool isLegacyAddressFirst();

        /// Check if the ip address is from legacy
        /// @return true if the address is in the legacy format, false otherwise
        static bool isLegacyAddressIp(uint32_t tIpAddress);
        static bool isLegacyAddressIp(std::string tIpAddress);

        /// Check if the ip address is from the new networking system
        /// @return true if the address is in the new format, false otherwise
        static bool isNewAddressIp(uint32_t tIpAddress);
        static bool isNewAddressIp(std::string tIpAddress);
        void apply10xNetworkFlagsOnSocket(int socketFd, std::string ipAddress, uint16_t port);

        static bool isLocalhost(std::string tIpAddress);

        /// Socket packet MARK constant
        static const int c_markValue;
    protected:

        CIpInfo(const CIpInfo& IpInfo);
        CIpInfo& operator = (const CIpInfo& IpInfo);

        uint32_t        m_IpAddress;
        uint32_t        m_NetMask;

        uint32_t        m_LegacyNetworkIpAddress;
        uint32_t        m_LegacyNetworkNetMask;

        uint32_t        m_LoopbackIpAddress;
        uint32_t        m_LoopbackNetMask;

        uint32_t        m_WirelessIpAddress;
        uint32_t        m_WirelessNetMask;
        uint32_t        m_DefaultGateway;
        MacAddress_t   m_MacAddress;

        std::string     m_InterfaceName;
        std::string     m_IPAddressString;
        std::string     m_NetMaskString;

        std::string     m_LegacyNetworkInterfaceName;
        std::string     m_LegacyNetworkIPAddressString;
        std::string     m_LegacyNetworkNetMaskString;

        std::string     m_LoopbackInterfaceName;
        std::string     m_LoopbackIPAddressString;
        std::string     m_LoopbackNetMaskString;

        std::string     m_WirelessIPAddressString;
        std::string     m_WirelessNetMaskString;
        std::string     m_DefaultGatewayString;
        std::string     m_MacAddressString;
        std::string     m_WirelessMacAddressString;

        //this member is used to determine if the legacy ip address is listed first
        bool m_isLegacyFirst;

//        int m_UpdateIntInfoFD;
//        int m_GetNetStatusFD;
private:
        struct InterfaceInfo
        {
                RaySocketTypes::EthernetDevice_t interfaceType;
                uint32_t ipAddress;
                uint32_t netmaskAddress;

                InterfaceInfo()
                : interfaceType(RaySocketTypes::e_Unknown),
                  ipAddress(0),
                  netmaskAddress(0)
                {}
        };

        bool _PollInterfaceAvailability(int fd, std::string interfaceName, unsigned int pollRetries, bool* pConfigured);
        void _IpAddressToString(uint32_t ipAddress, std::string* pString);
        bool _PopulateBasicNetworkInfoRetry(int nbRetries);
        bool _PopulateBasicNetworkInfo(bool abort = true);
        void _initializeIfConf(int fd, struct ifconf *ifc);
};

#endif // IP_INFO_H_INCLUDED
