/*
Copyright (c) 2015, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "MemoryEventLogger.h"

#include <pthread.h>
#include <cstring>
#include <cstdio>
#include <cstdlib>

#include <unistd.h>
#include <sys/types.h>

#include "GetThreadId.h"

namespace MemoryEventLogger
{
#ifdef ENABLE_MEMORY_EVENT_LOGGER

size_t BufferSize = 0;
char* LogBuffer = 0;
size_t LogBufferIndex = 0;
bool LogBufferEnabled = false;

pthread_mutex_t Mutex = PTHREAD_MUTEX_INITIALIZER;

static const char* FileName(const char* pFilePath)
{
    // strip off any directories before the file name
    const char* pFileName = strrchr(pFilePath, '/');
    if (pFileName == 0)
    {
        return pFilePath;
    }
    else
    {
        return pFileName + 1;
    }
}

void Init(size_t Size)
{
    if (LogBuffer == 0 && Size > 0)
    {
        BufferSize = Size;
        LogBuffer = (char *)malloc(BufferSize);
        if (LogBuffer)
        {
            memset(LogBuffer, 0, BufferSize);
            LogBufferEnabled = true;
        }
    }
}

void LogToMemoryBuffer(const char* pFileName,
                       int LineNumber,
                       const char* pFunctionName,
                       const char* pFormatString,
                       va_list arg )
{
    if (LogBufferEnabled)
    {
        if (pthread_mutex_lock(&Mutex) == 0)
        {
            const int c_MaxDebugLength = 1024;
            char DebugString[c_MaxDebugLength];

            // Decode variable argument list to produce the output debugging string.
            vsnprintf(DebugString, c_MaxDebugLength, pFormatString, arg);

            // Null Terminate Buffer in case of max buffer size write
            DebugString[c_MaxDebugLength - 1] = 0;

            struct timespec tp;

            // Get Current time
            clock_gettime(CLOCK_MONOTONIC, &tp);
            double Time = static_cast<double>(tp.tv_sec) + static_cast<double>(tp.tv_nsec / 1000000) / 1000.0;

            char FullDebugString[c_MaxDebugLength];
            snprintf(FullDebugString, c_MaxDebugLength, "%.3f [%d] %s:%d (%s): %s", Time, GetThreadId(), FileName(pFileName), LineNumber, pFunctionName, DebugString);

            size_t len = strlen(FullDebugString);

            if ((LogBufferIndex + len) < BufferSize)
            {
                // string fits in buffer
                memcpy(LogBuffer + LogBufferIndex, FullDebugString, len + 1);
                LogBufferIndex += (len + 1);
                if (LogBufferIndex >= BufferSize)
                {
                    LogBufferIndex = 0;
                }
            }
            else
            {
                // String too long to fit at end of buffer, fill to end of buffer with 0 and then write string at start of buffer
                memset(LogBuffer + LogBufferIndex, 0, BufferSize - LogBufferIndex);
                memcpy(LogBuffer, FullDebugString, len + 1);
                LogBufferIndex = len + 2;
            }

            pthread_mutex_unlock(&Mutex);
        }
    }
}


size_t ReadIndex = 0;
char* Buffer = 0;

void InitMemoryBufferRead()
{
    LogBufferEnabled = false;
    ReadIndex = LogBufferIndex;
    Buffer = (char *)malloc(BufferSize);
    if (Buffer && LogBuffer)
    {
        // Find first NULL character marking end of earliest log entry (probably incomplete so ignore it)
        while (ReadIndex < BufferSize && Buffer[ReadIndex] != 0)
        {
            ++ReadIndex;
        }
        if (ReadIndex >= BufferSize -1)
        {
            // The end of buffer was reached before finding a NULL, so just use the buffer starting from the beginning
            memcpy(Buffer, LogBuffer, LogBufferIndex - 1);
            Buffer[LogBufferIndex - 1] = 0;
        }
        else
        {
            ++ReadIndex; // Skip over NULL
            size_t SkippedBytes = ReadIndex - LogBufferIndex;
            size_t NumBytesAtEndOfBuffer = BufferSize - ReadIndex;
            memcpy(Buffer, LogBuffer + ReadIndex, NumBytesAtEndOfBuffer);
            memcpy(Buffer + NumBytesAtEndOfBuffer, LogBuffer, BufferSize - (NumBytesAtEndOfBuffer + SkippedBytes));
            Buffer[BufferSize - (SkippedBytes + 1)] = 0;
        }
        ReadIndex = 0;
    }
}

char* GetNextMemoryBufferEntry()
{
    char* result = 0;

    if (Buffer != 0)
    {
        while (ReadIndex < BufferSize)
        {
            size_t Len = strlen(Buffer + ReadIndex);

            if (Len != 0)
            {
                result = Buffer + ReadIndex;
                ReadIndex += (Len + 1);
                break;
            }
            ReadIndex += (Len + 1);
        }
    }
    return result;
}
#endif
} // namespace
