/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// Originally developed from eCrash published in Linux Jornal by David Frascone
// http://www.linuxjournal.com/article/8724

#ifndef _OMAPCRASH_H_
#define _OMAPCRASH_H_

#include <stdio.h>
#include <signal.h>
#include <QString>

#define MAX_LINE_LEN 2048

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

#define ECRASH_DEFAULT_STACK_DEPTH 10
#define ECRASH_DEFAULT_BACKTRACE_SIGNAL SIGRTMAX
#define ECRASH_DEFAULT_THREAD_WAIT_TIME_MS 4000
#define ECRASH_MAX_NUM_SIGNALS 30

/** \struct eCrashSymbol
 *  \brief Function Name / Address pair
 *
 *  This is used in conjunction with eCrashSymbolTable to
 *  provide an alternative to backtrace_symbols.
 */
typedef struct {
    char          *function;
    void          *address;
} eCrashSymbol;
/** \struct eCrashSymbolTable
 *  \brief Symbol table used to avoid backtrace_symbols()
 *
 *  This structure is used to provide a alternative to
 *  backtrace_symbols which is not async signal safe.
 *
 *  The symbol table should be sorted by address (it will
 *  be either binary or linearly searched)
 */
typedef struct {
    int           numSymbols;
    eCrashSymbol *symbols;
} eCrashSymbolTable;

#define MAX_FUNCTION_NAME_LENGTH 50

typedef struct sCallParams {
    volatile unsigned int valid;
    unsigned int Func;
    char sName[MAX_FUNCTION_NAME_LENGTH];
    unsigned int pc;
    unsigned int sp;
    unsigned int fp;
    unsigned int ra;
    unsigned int a0;
    unsigned int a1;
    unsigned int a2;
    unsigned int a3;
} tCallParams;



#define ECRASH_DEBUG_ENABLE  /* undef to turn off debug */

#ifdef ECRASH_DEBUG_ENABLE
# define ECRASH_DEBUG_VERY_VERBOSE 1
# define ECRASH_DEBUG_VERBOSE      2
# define ECRASH_DEBUG_INFO         3
# define ECRASH_DEBUG_WARN         4
# define ECRASH_DEBUG_ERROR        5
# define ECRASH_DEBUG_OFF          6
# define ECRASH_DEBUG_DEFAULT     (ECRASH_DEBUG_ERROR)
# define DPRINTF(level, fmt...) \
        if (level >= gbl_params.debugLevel) { printf(fmt); fflush(stdout); }
#else /* ECRASH_DEBUG_ENABLE */
# define DPRINTF(level, fmt...)
#endif /* ECRASH_DEBUG_ENABLE */

typedef int (*OutputFunctionPtr_t)(const char*, ...);
typedef void (*CrashLoggerOutputFunctionPtr_t)(OutputFunctionPtr_t);

/** \struct eCrashParameters
 *  \brief eCrash Initialization Parameters
 *
 *  This structure contains all the global initialization functions
 *  for eCrash.
 *
 *  @see eCrash_Init
 */
typedef struct {


    /*  OUTPUT OPTIONS */
        /** Filename to output to, or NULL */
    const char *filename;
        /** FILE * to output to or NULL */
    FILE *filep;
        /** fd to output to or -1 */
    int fd;

        /** If a filename is specified here, then if that
          * file exists when a crash occurs, a core dump
          * will be generated. On G-Series typically
          * the tester will have a special USB memory stick with
          * a file called DoDump.txt which will cause the
          * core dump to be written to the USB memory stick.
          *
          * The location the core dump is written to is
          * configured in /proc/sys/kernel/core_pattern
          */
    const char* doCoreDumpIndicatorFile;

        /** If a filename is specified, then a dump of the
          * main window is written to the filename.
          */
    const char* windowDumpFile;

    int debugLevel;

        /** If true, all registered threads will
         *   be dumped
         */
    bool dumpAllThreads;

        /** How far to backtrace each stack */
    int maxStackDepth;

        /** Default signal to use to tell a thread to drop it's
         *  stack.
         */
    int defaultBacktraceSignal;

        /** How long to wait for all threads
         *  dump
         */
    unsigned int threadWaitTime;

        /** To avoid the issues with backtrace_symbols (see comments above)
          * the caller can supply it's own symbol table, containing function
          * names and start addresses.  This table can be created using
          * a script, or a static table.
          *
          * If this variable is not null, it will be used, instead of
          * backtrace_symbols, reguardless of the setting
          * of useBacktraceSymbols.
          */
    eCrashSymbolTable *symbolTable;

        /** This will cause a nanosleep() to be done _BEFORE_ the
         *  final abort() is called. This allows us to attach a
         *  debugger to diagnose the cause of the fault */
    bool delayBeforeFinalAbort;

    /**
     * The application included in the log.
     */
    const char  *applicationName;

        /** The software version string is included in the log, so that
          * the log may be tied back to a particular version of software.
          */
    const char  *softwareVersionString;

        /** Array of crash signals to catch,
         *  ending in 0  I would have left it a [] array, but I
         *  do a static copy, so I needed a set size.
         */
    int signals[ECRASH_MAX_NUM_SIGNALS];

    CrashLoggerOutputFunctionPtr_t pLNetDiagnosticsOutputFunction;

    CrashLoggerOutputFunctionPtr_t pFunctionTraceOutputFunction;

} eCrashParameters;

/*!
 * Method to pass carthography library version numbers
 */
void eCrash_PassCartLibVersions(const std::string& jeppesenLibVersion, const std::string& navionicsLibVersion);

/*!
 * Initialize eCrash.
 *
 * This function must be called before calling any other eCrash
 * functions.  It sets up the global behavior of the system.
 *
 * @param params Our input parameters.  The passed in structure will be copied.
 *
 * @return Zero on success.
 */
int eCrash_Init(eCrashParameters *params);
/*!
 * UnInitialize eCrash.
 *
 * This function may be called to de-activate eCrash, release the
 * signal handlers, and free any memory allocated by eCrash.
 *
 * @return Zero on success.
 */
int eCrash_Uninit();

/*!
 * Register a thread for backtracing on crash.
 *
 * This function must be called by any thread wanting it's stack
 * dumped in the event of a crash.  The thread my specify what
 * signal should be used, or the default, SIGUSR1 will be used.
 *
 * @param name String used to refer to us in crash dumps
 * @param signo Signal to use to generate dump (default: SIGUSR1)
 *
 * @return Zero on success.
 */
int eCrash_RegisterThread(const char *name, int signo);

/*!
 * Un-register a thread for stack dumps.
 *
 * This function may be called to un-register any previously
 * registered thread.
 *
 * @return Zero on success.
 */
int eCrash_UnregisterThread();

void eCrash_RegisterQmlThread();
/*!
 * Abort method.
 * This performs the same backtrace on crash, as a signal crash, but
 * adds the details given in the printf style format to the log file.
 * This should be called when things in the application are not
 * recoverable.
 */
void eCrash_Abort(const char *format, ...) __THROW __attribute__ ((__noreturn__));

/*!
 * Abort method.
 * This performs the same backtrace on crash, as a signal crash, but
 * adds the details given an abort message style.
 * This should be called when things in the application are not
 * recoverable.
 */
void eCrash_Abort(const char* err_str, int err_no, const char *file, int line) __THROW __attribute__ ((__noreturn__));

/*!
 * assert method.
 * This performs the same backtrace on crash, as a signal crash, but
 * adds the details given in the assert to the log file.
 */
extern "C" void eCrash_assert(const char *cond, const char* file, int line)  __THROW __attribute__ ((__noreturn__));

#endif /* _OMAPCRASH_H_ */
