/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "POSIXDirectory.h"
#include "RayAssert.h"
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <algorithm>

//--------------------------------------------------------------------------------
CPOSIXDirectory::CPOSIXDirectory() :
    m_options(NO_OPTIONS)
{

}


//--------------------------------------------------------------------------------
CPOSIXDirectory::~CPOSIXDirectory()
{
    Close();
}

//--------------------------------------------------------------------------------
bool    CPOSIXDirectory::Open(const char* folderName, OPEN_OPTIONS   options)
{
    bool result = false;
    
    //
    // Start of empty, in case the callers has failed to call Close().
    //
    Close();
    
    m_folderName    = folderName;
    m_options       = options;
    
    DIR*    directory = opendir(folderName);
    if (directory)
    {
        result = true;
        LoadFilenameList(directory, options);
        if (IsOptionSet(options, SORT_BY_NAME))
        {
            m_filenameList.sort();
        }
        
        closedir(directory);
    }
    return result;
}

//--------------------------------------------------------------------------------
void    CPOSIXDirectory::Reload()
{
    m_filenameList.clear();
    m_currentFilename = m_filenameList.end();
    
    //
    // It is necessary to take a copy of the foldername as this is
    // cleared by Open which does a close to clear things up before doing
    // the open work.
    //
    std::string folderName(m_folderName);
    
    Open(folderName.c_str(), m_options);
}

//--------------------------------------------------------------------------------
void    CPOSIXDirectory::LoadFilenameList(DIR*   directory, OPEN_OPTIONS options)
{
    RAYASSERT(directory);
    
    struct dirent*  entry = 0;
    do 
    {
        entry = readdir(directory);
        if (entry)
        {
            struct stat StatData;
            std::string DirName = entry->d_name;
            std::string FullName = BuildFullname(DirName);
            if (stat(FullName.c_str(), &StatData) == 0)
            {
                if (S_ISDIR(StatData.st_mode))
                {
                    if (IsOptionSet(options, RETURN_DIRECTORIES))
                    {
                        if (IsDotOrDotDot(entry->d_name))
                        {
                            if (IsOptionSet(options, RETURN_DOT_AND_DOT_DOT))
                            {
                                m_filenameList.push_back(DirName);
                            }
                        }
                        else
                        {
                            m_filenameList.push_back(DirName);
                        }
                    }
                }
                else
                {
                    m_filenameList.push_back(DirName);
                }
            }
        }
    }
    while (entry);
}

//--------------------------------------------------------------------------------
void    CPOSIXDirectory::Close()
{
    m_filenameList.clear();
    m_folderName.clear();
    m_currentFilename = m_filenameList.end();
}

//--------------------------------------------------------------------------------
void    CPOSIXDirectory::ResetGetNextFilename()
{
    m_currentFilename = m_filenameList.begin();
}

//--------------------------------------------------------------------------------
bool    CPOSIXDirectory::GetNextFilename(std::string& filename)
{
    bool result = false;
    
    if (m_currentFilename != m_filenameList.end())
    {
        filename = *m_currentFilename;
        ++m_currentFilename;
        result = true;
    }
    return result;
}

//--------------------------------------------------------------------------------
bool CPOSIXDirectory::DeleteFile(const std::string& filename)
{
    bool result = false;
    
    int r = unlink(BuildFullname(filename).c_str());
     
    if (r == 0)
    {
        RemoveFilenameFromList(filename);
        result = true;
    }
    
    return result;
}

//--------------------------------------------------------------------------------
unsigned  CPOSIXDirectory::GetNumberFiles()
{
    return m_filenameList.size();
}

//--------------------------------------------------------------------------------
bool    CPOSIXDirectory::IsOptionSet(OPEN_OPTIONS    options, OPEN_OPTIONS   optionToTest)
{
    return (options & optionToTest) != 0;
}

//--------------------------------------------------------------------------------
bool    CPOSIXDirectory::IsDotOrDotDot(const char *filename)
{
    return ((strcmp(filename, ".") == 0) || (strcmp(filename, "..") == 0));
}

//--------------------------------------------------------------------------------
std::string  CPOSIXDirectory::BuildFullname(const std::string& filename)
{
    // build full filename.
    std::string  fullName;
    
    fullName.append(m_folderName);
    fullName.append("/");
    fullName.append(filename);
    
    return fullName;
}
//--------------------------------------------------------------------------------
bool CPOSIXDirectory::RemoveFilenameFromList(const std::string& filename)
{
    bool result = false;
    
    FilenameList_t::iterator i = find(m_filenameList.begin(), m_filenameList.end(), filename);    
    
    if (i != m_filenameList.end())
    {
        //
        // Ensure current iterator is left valid.
        //
        if (i == m_currentFilename)
        {
            ++m_currentFilename;
        }
        
        m_filenameList.erase(i);
        result = true;
    }
    return result;

}

//--------------------------------------------------------------------------------
void    CPOSIXDirectory::DumpFilelist(FilenameList_t  filenameList)
{
    printf("+++ Dump file list\n");
    FilenameList_t::iterator i;
    for (i = filenameList.begin(); i != filenameList.end(); ++i)
    {
        printf("%s\n", i->c_str());
    }
    printf("+++ End Dump\n");
    
}

//--------------------------------------------------------------------------------
// EOF
//

