/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef POSIXDIRECTORY_H_INCLUDED
#define POSIXDIRECTORY_H_INCLUDED

#include <list>
#include <string>
#include <dirent.h>

//--------------------------------------------------------------------------------
// This class is responsible for providing iteration over the filenames
// in a given directory. Optionally the names can be sorted.
//
// The class is called POSIX Directory as it is implemented using the POSIX
// standard interface methods. Please, note there is a CDirectory class already
// in the RL90 code base for use with compact flash cards. It was first created for the MMIDemonstrator
// on G-Series.
//
class CPOSIXDirectory{
public:

    CPOSIXDirectory();
    virtual ~CPOSIXDirectory();

    //
    // Options available when the directory is opened.
    //
    enum OPEN_OPTIONS
    {
        NO_OPTIONS              = 0,
        SORT_BY_NAME            = 0x0001,
        RETURN_DOT_AND_DOT_DOT  = 0x0002,       // i.e. return '.' and '..' files. default is not to return them.
        RETURN_DIRECTORIES      = 0x0004
    };

    bool        Open(const char* folderName, OPEN_OPTIONS   options = NO_OPTIONS);
    void        Close();

    void        ResetGetNextFilename();
    bool        GetNextFilename(std::string& filename);

    unsigned    GetNumberFiles();

    std::string BuildFullname(const std::string& filename);

    bool        DeleteFile(const std::string& filename);

    //
    // If files are added to the directory then the directory needs to be reloaded.
    //
    void    Reload();

    const std::string& GetFolderName() const
    {
        return m_folderName;
    };

private:

    typedef std::list<std::string>   FilenameList_t;

    FilenameList_t                   m_filenameList;
    FilenameList_t::iterator         m_currentFilename;

    std::string                     m_folderName;
    OPEN_OPTIONS                    m_options;

    bool    IsOptionSet(OPEN_OPTIONS    options, OPEN_OPTIONS   optionToTest);
    void    LoadFilenameList(DIR*   directory, OPEN_OPTIONS options);

    bool    IsDotOrDotDot(const char *filename);

    bool    RemoveFilenameFromList(const std::string& filename);

    void    DumpFilelist(FilenameList_t  filenameList);

private:
    //--------------------------------------------------------------------------
    // Disallow object copying by providing a private copy constuctor and
    // assignment operator.
    //
    CPOSIXDirectory(const CPOSIXDirectory& rOther);

    CPOSIXDirectory&
    operator = (const CPOSIXDirectory& rOther);
};

#endif


