/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "ProcessWatchdogAndroid.h"
#include "EventLogger.h"

#include <QtAndroid>
#include <QAndroidJniObject>
#include <QCoreApplication>

#include <unistd.h>
#include <sys/syscall.h>
#include <sys/types.h>
#include <signal.h>
#include <android/log.h>


extern "C" void android_set_abort_message(const char* msg);

static constexpr int c_ToggleRate_ms = 1000;

static int MainThreadId = -1;

#ifdef _PROCESS_USES_OSS
#include "Framework/Logging/DumpMutexList.h"

static int output_fn(const char* format, ...)
{
    va_list ap;
    int return_value=0;
    va_start(ap, format);
    return_value = __android_log_vprint(ANDROID_LOG_ERROR, QCoreApplication::applicationName().toLatin1().data(), format, ap);
    va_end(ap);
    return return_value;
}
#endif

extern "C" JNIEXPORT void JNICALL Java_com_raymarine_axiom_mfd_ProcessMonitor_timeout(JNIEnv*,
                                                                                      jobject)
{
#ifdef _PROCESS_USES_OSS
    DumpMutexList(output_fn);
#endif
    __android_log_print(ANDROID_LOG_FATAL, QCoreApplication::applicationName().toLatin1().data(), "Watchdog timeout detected");
    android_set_abort_message("Watchdog timeout detected");
    if (MainThreadId != -1)
    {
        // Send abort signal to main thread so we get backtrace from there
        syscall(SYS_tgkill, (int)getpid(), MainThreadId, SIGABRT);
        // Abort here as well just in case
        sleep(2);
    }
    abort();
}


CProcessWatchdogAndroid::CProcessWatchdogAndroid()
    : m_pProcessMonitor(nullptr)
    , m_WriterTimer()
{
    MainThreadId = gettid();
    RAYINFO("%d", MainThreadId);
    QAndroidJniObject mainClass = QtAndroid::androidActivity();
    if (!mainClass.isValid())
    {
        mainClass = QtAndroid::androidService();
    }
    if (mainClass.isValid())
    {
        m_pProcessMonitor = new QAndroidJniObject(mainClass.getObjectField("mProcessMonitor", "Lcom/raymarine/axiom/mfd/ProcessMonitor;"));
    }

    if (m_pProcessMonitor && m_pProcessMonitor->isValid())
    {
        m_WriterTimer.setInterval(c_ToggleRate_ms);
        m_WriterTimer.setSingleShot(false);
        bool Connected = connect(&m_WriterTimer, &QTimer::timeout, this, &CProcessWatchdogAndroid::Toggle);
        RAYASSERT(Connected);

        m_WriterTimer.start();
    }
    else
    {
        RAYERROR("Java ProcessMonitor object not found");
    }
}

CProcessWatchdogAndroid::~CProcessWatchdogAndroid()
{
    delete m_pProcessMonitor;
    m_pProcessMonitor = nullptr;
    RAYINFO("");
}

unsigned int CProcessWatchdogAndroid::getToggleRate() const
{
    return c_ToggleRate_ms;
}

void CProcessWatchdogAndroid::SetWatchdogTimeout(int Timeout)
{
    if (m_pProcessMonitor && m_pProcessMonitor->isValid())
    {
        RAYINFO("Interval %ds", Timeout);
        m_pProcessMonitor->callMethod<void>("setTimeout", "(I)V", Timeout * 1000);

        _reset(Timeout);
    }
}

/// Timeout function for timer. resets Java monitor
void CProcessWatchdogAndroid::Toggle()
{
    if (m_pProcessMonitor && m_pProcessMonitor->isValid())
    {
        // Only toggle the watchdog at the toggle rate, more frequent toggle attempts are ignored
        //
        long long currentLateTime_ms = 0;
        long long worstLateTime_ms = 0;

        if (_canToggle(currentLateTime_ms, worstLateTime_ms))
        {
            /// @todo remove this debug once we have sorted out the watchdog timeout
            /// problems at startup
            static size_t count = 0;
            if (++count < 200)
            {
                RAYINFO("Writing to watchdog");
            }

            m_pProcessMonitor->callMethod<void>("reset");

            _updateToggleTime();

            RAYTRACE("Watchdog write complete, %lldms late, worst: %lldms late", currentLateTime_ms, worstLateTime_ms);
        }
    }
}

