/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "ProcessWatchdogFifoFile.h"
#include "EventLogger.h"
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fstream>
#include <errno.h>
#include "GetThreadId.h"

static const int c_ToggleRate_ms = 1000; // 1 second
static const int c_DefaultTimeout = 20;

CProcessWatchdogFifoFile::CProcessWatchdogFifoFile(const std::string& WatchdogFilename) :
    m_WatchdogFilename(WatchdogFilename),
    m_Timeout(),
    m_WriterTimer()
{
    RAYINFO("%s", WatchdogFilename.c_str());
    // Create the FIFO used to signal that the mfdapp is alive
    mkfifo(m_WatchdogFilename.c_str(), S_IWUSR | S_IRUSR | S_IRGRP | S_IROTH);

    SetWatchdogTimeout(c_DefaultTimeout);

    m_WriterTimer.setInterval(c_ToggleRate_ms);
    m_WriterTimer.setSingleShot(false);

    RAYTRACE("%s Interval %dms", WatchdogFilename.c_str(), c_ToggleRate_ms);
    bool Connected = QObject::connect(&m_WriterTimer, &QTimer::timeout, this, &CProcessWatchdogFifoFile::Toggle);
    RAYASSERT(Connected);

    m_WriterTimer.start();
}

CProcessWatchdogFifoFile::~CProcessWatchdogFifoFile()
{
    RAYTRACE("");
    unlink(m_WatchdogFilename.c_str());
}

unsigned int CProcessWatchdogFifoFile::getToggleRate() const
{
    return c_ToggleRate_ms;
}

void CProcessWatchdogFifoFile::SetWatchdogTimeout(int Timeout)
{
    RAYINFO("Interval %ds", Timeout);
    char buffer[32];
    sprintf(buffer, "%d\n", Timeout);
    m_Timeout = buffer;

    _reset(Timeout);
}

/// Timeout function for timer.  Writes to the file
void CProcessWatchdogFifoFile::Toggle()
{
    // Only toggle the watchdog at the toggle rate, more frequent toggle attempts are ignored
    //
    long long currentLateTime_ms = 0;
    long long worstLateTime_ms = 0;

    if (_canToggle(currentLateTime_ms, worstLateTime_ms))
    {
        int fd = open(m_WatchdogFilename.c_str(), O_WRONLY | O_NONBLOCK);
        if (fd > 0)
        {
            RAYTRACE("Thread: %d, Writing to watchdog", GetThreadId());
            if (write(fd, m_Timeout.c_str(), m_Timeout.length()) == -1)
            {
                RAYERROR("Thread: %d, Write error: %s", GetThreadId(), strerror(errno));
            }
            else
            {
                _updateToggleTime();

                RAYTRACE("Watchdog write complete, %lldms late, worst: %lldms late", currentLateTime_ms, worstLateTime_ms);
            }
            close(fd);
            fsync(fd);
        }
        else
        {
            RAYERROR("Thread: %d, Error: %s opening: %s", GetThreadId(), strerror(errno), m_WatchdogFilename.c_str());
        }
    }
}

