//-------------------------------------------------------------------------//
// Copyright (c) Raymarine UK Limited 2019
//
// Reproduction or transmission in whole or in part (whether by photocopying or
// storing in any medium by electronic means or otherwise) without the written
// permission of Raymarine UK Limited is prohibited.
//
// Confidential
//-------------------------------------------------------------------------//

#include "ProcessWatchdogMonitor.h"
#include "EventLogger.h"
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fstream>
#include <errno.h>

static const int c_SlowWarningRate_ms = 5000; // 5 seconds
static const int c_DefaultTimeout = 30; // 30 seconds

static const unsigned long c_nsPerSecond = 1000000000;
static const unsigned long c_usPerSecond = 1000000;
static const unsigned int  c_msPerSecond = 1000;

///
/// \brief timeDelta
///
/// substracts time1 from time2 and sets rTimeDelta for the difference
///
/// \param time1
/// \param time2
/// \param rTimeDelta   - time difference
///
void timeDelta(const struct timespec& time1, const struct timespec& time2, struct timespec& rTimeDelta)
{
    rTimeDelta.tv_nsec = time2.tv_nsec - time1.tv_nsec;
    rTimeDelta.tv_sec  = time2.tv_sec - time1.tv_sec;
    if (rTimeDelta.tv_sec > 0 && rTimeDelta.tv_nsec < 0)
    {
        rTimeDelta.tv_nsec += c_nsPerSecond;
        rTimeDelta.tv_sec--;
    }
    else if (rTimeDelta.tv_sec < 0 && rTimeDelta.tv_nsec > 0)
    {
        rTimeDelta.tv_nsec -= c_nsPerSecond;
        rTimeDelta.tv_sec++;
    }
}

///
/// \brief time_ms
/// Converts a timespec to milliseconds
///
/// \param time
///
/// \return timespec in milliseconds
///
unsigned long time_ms(const struct timespec& time)
{
    return (time.tv_sec * c_msPerSecond) + (time.tv_nsec / 1000000);
}

ProcessWatchdogMonitor::ProcessWatchdogMonitor()
    : m_lastToggleTime()
    , m_Timeout_ms(0)
    , m_worstWatchdogUpdateTime_ms(0)
{
    _reset(c_DefaultTimeout);
}

ProcessWatchdogMonitor::~ProcessWatchdogMonitor()
{
    RAYTRACE("");
}

///
/// \brief ProcessWatchdogMonitor::_reset
/// Resets the worst time, should be called when the watchdog timeout time
/// is changed
///
/// \param Timeout
///
void ProcessWatchdogMonitor::_reset(int Timeout)
{
    m_Timeout_ms = Timeout * c_msPerSecond;
    m_worstWatchdogUpdateTime_ms = 0;
}

///
/// \brief ProcessWatchdogMonitor::_canToggle
///
/// Determines if it is time to toggle the watchdog and updates the monitor times
///
/// \param currentLateTime_ms   - set to the number of milliseconds that have passed since this toggle should have occurred
/// \param worstLateTime_ms     - set the worst late toggle time
///
/// \return true if it is time to toggle the watchdog
///
bool ProcessWatchdogMonitor::_canToggle(long long& currentLateTime_ms, long long& worstLateTime_ms)
{
    bool canToggle = false;

    struct timespec now, delta;
    bool timeWentBackwards  = false;
    bool timeJumpedForwards = false;

    // Calculate the time between now and the last toggle
    //
    clock_gettime(CLOCK_REALTIME, &now);
    timeDelta(m_lastToggleTime, now, delta);

    unsigned long elapsedTime_ms    = time_ms(delta);
    unsigned long now_ms            = time_ms(now);
    unsigned long lastToggle_ms     = time_ms(m_lastToggleTime);

    // Detect if the time jumps forwards or backwards
    //
    if (now_ms < lastToggle_ms)
    {
        RAYDETAIL("Time jumped backwards by: %ldms", (now_ms - lastToggle_ms));
        timeWentBackwards = true;
    }
    else if (now_ms > lastToggle_ms + m_Timeout_ms)
    {
        RAYDETAIL("Time jumped forwards by: %ldms", (lastToggle_ms - now_ms));
        timeJumpedForwards = true;
    }

    // Only toggle the watchdog at the toggle rate, more frequent toggle attempts are ignored
    //
    if (elapsedTime_ms > getToggleRate())
    {
        canToggle = true;

        // Only update the monitor times of the time hasn't jumped
        //
        if (!timeWentBackwards && !timeJumpedForwards)
        {
            bool outputWatchdogSlowWarning = false;
            if (elapsedTime_ms > c_SlowWarningRate_ms)
            {
                outputWatchdogSlowWarning = true;
            }

            if (elapsedTime_ms > m_worstWatchdogUpdateTime_ms)
            {
                m_worstWatchdogUpdateTime_ms = elapsedTime_ms;
                outputWatchdogSlowWarning = true;
            }

            currentLateTime_ms = elapsedTime_ms - getToggleRate();
            worstLateTime_ms   = m_worstWatchdogUpdateTime_ms - getToggleRate();

            if (outputWatchdogSlowWarning)
            {
                RAYWARNING("System slow, this toggle: %lldms late, worst: %lldms late", currentLateTime_ms, worstLateTime_ms);
            }
        }
    }

    return canToggle;
}

///
/// \brief ProcessWatchdogMonitor::_updateToggleTime
///
/// Sets the toggle time to the current time.
///
void ProcessWatchdogMonitor::_updateToggleTime()
{
    struct timespec now;
    clock_gettime(CLOCK_REALTIME, &now);

    m_lastToggleTime = now;
}
