/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "CameraProxy.h"
#include "soapDeviceBindingProxy.h"
#include "soapImagingBindingProxy.h"
#include "soapMediaBindingProxy.h"
#include "IpCameraStore.h"
#include "FrameworkCommon/EventLogger.h"
#include "FrameworkCommon/IpInfo.h"
#include "DiscoveryService.h"
#include <cstdio>
#include <QMutex>

//#define DUMP_ENCODER_CONFIGS
//#define ERASE_EXISTING_PROFILES

static const int c_NamePollPeriodMs = 10000;
static const int c_MinNamePollPeriodMs = 2000;
static const char c_MinNameChar = 'a';  ///< Minimum encoded name character
static const char c_MaxNameChar = 'p';  ///< Maximum encoded name character
/// Prefix in encoded name.  This lets us detect whether we set this hostname or not
static const std::string c_EncodingPrefix = "rrrr";

struct EncoderPreferences_t
{
    EncoderPreferences_t(const char* Name,
                         int Width,
                         int Height,
                         float Quality,
                         int Bitrate,
                         int Framerate,
                         int MinHeight) :
         Name(Name),
         Width(Width),
         Height(Height),
         Quality(Quality),
         Bitrate(Bitrate),
         Framerate(Framerate),
         MinHeight(MinHeight)
    {
    }
    std::string     Name;
    int             Width;
    int             Height;
    float           Quality;
    int             Bitrate;
    int             Framerate;
    int             MinHeight;
};

static const EncoderPreferences_t c_HiResProfile("RymH264Hi", 1280, 720, 0.7, 4000, 25, 480);
static const EncoderPreferences_t c_LoResProfile("RymH264Lo", 640, 360, 0.7, 1200, 25, 240);

void LogSoapError(struct soap* pointer)
{
    std::vector<char> buffer(2048);
    soap_sprint_fault(pointer, buffer.data(), buffer.size()-1);
    buffer.back() = 0;
    RAYERROR("%s", buffer.data());
}


#ifndef WITH_NOIO
#ifndef PALM_1
static const char*
soap_strerror(struct soap *soap)
{ int err = soap->errnum;
  *soap->msgbuf = '\0';
  if (err)
  {
#ifndef WIN32
# ifdef HAVE_STRERROR_R
#  if defined(_GNU_SOURCE) && !defined(__ANDROID__)
    return strerror_r(err, soap->msgbuf, sizeof(soap->msgbuf)); /* GNU-specific */
#  else
    strerror_r(err, soap->msgbuf, sizeof(soap->msgbuf)); /* XSI-compliant */
#  endif
# else
    return strerror(err);
# endif
#else
#ifndef UNDER_CE
    DWORD len;
    *soap->msgbuf = '\0';
    len = FormatMessageA(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, err, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), (LPSTR)soap->msgbuf, (DWORD)sizeof(soap->msgbuf), NULL);
#else
    DWORD i, len;
    *soap->msgbuf = '\0';
    len = FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS, NULL, err, 0, (LPTSTR)soap->msgbuf, (DWORD)(sizeof(soap->msgbuf)/sizeof(TCHAR)), NULL);
    for (i = 0; i <= len; i++)
    { if (((TCHAR*)soap->msgbuf)[i] < 0x80)
        soap->msgbuf[i] = (char)((TCHAR*)soap->msgbuf)[i];
      else
        soap->msgbuf[i] = '?';
    }
#endif
#endif
  }
  else
  { int rt = soap->recv_timeout, st = soap->send_timeout;
#ifndef WITH_LEAN
    int ru = ' ', su = ' ';
#endif
    soap_strcpy(soap->msgbuf, sizeof(soap->msgbuf), "message transfer interrupted");
    if (rt || st)
      soap_strcpy(soap->msgbuf + 28, sizeof(soap->msgbuf) - 28, " or timed out");
#ifndef WITH_LEAN
    if (rt < 0)
    { rt = -rt;
      ru = 'u';
    }
    if (st < 0)
    { st = -st;
      su = 'u';
    }
    if (rt)
    { size_t l = strlen(soap->msgbuf);
      (SOAP_SNPRINTF(soap->msgbuf + l, sizeof(soap->msgbuf) - l, 36), " (%d%cs recv delay)", rt, ru);
    }
    if (st)
    { size_t l = strlen(soap->msgbuf);
      (SOAP_SNPRINTF(soap->msgbuf + l, sizeof(soap->msgbuf) - l, 36), " (%d%cs send delay)", st, su);
    }
#endif
  }
  return soap->msgbuf;
}
#endif
#endif

#ifndef WITH_NOIO
#ifndef WITH_LEAN
static const struct soap_code_map h_error_codes[] =
{
#ifdef HOST_NOT_FOUND
  { HOST_NOT_FOUND, "Host not found" },
#endif
#ifdef TRY_AGAIN
  { TRY_AGAIN, "Try Again" },
#endif
#ifdef NO_RECOVERY
  { NO_RECOVERY, "No Recovery" },
#endif
#ifdef NO_DATA
  { NO_DATA, "No Data" },
#endif
#ifdef NO_ADDRESS
  { NO_ADDRESS, "No Address" },
#endif
  { 0, NULL }
};
#endif
#endif

#ifndef WITH_NOIO
#ifndef PALM_1
static const char*
tcp_error(struct soap *soap)
{ const char *msg = NULL;
  switch (soap->errmode)
  { case 0:
      msg = soap_strerror(soap);
      break;
    case 1:
      msg = "WSAStartup failed";
      break;
    case 2:
    {
#ifndef WITH_LEAN
      msg = soap_code_str(h_error_codes, soap->errnum);
      if (!msg)
#endif
      { (SOAP_SNPRINTF(soap->msgbuf, sizeof(soap->msgbuf), 37), "TCP/UDP IP error %d", soap->errnum);
        msg = soap->msgbuf;
      }
    }
  }
  return msg;
}
#endif
#endif

#define SOAP_TCP_SELECT_RCV 0x1
#define SOAP_TCP_SELECT_SND 0x2
#define SOAP_TCP_SELECT_ERR 0x4
#define SOAP_TCP_SELECT_ALL 0x7

#define SOAP_SOCKBLOCK(fd) fcntl(fd, F_SETFL, fcntl(fd, F_GETFL)&~O_NONBLOCK);
#define SOAP_SOCKNONBLOCK(fd) fcntl(fd, F_SETFL, fcntl(fd, F_GETFL)|O_NONBLOCK);

#ifndef WITH_NOIO
#ifndef PALM_1
static int
tcp_select(struct soap *soap, SOAP_SOCKET sk, int flags, int timeout)
{ int r;
  struct timeval tv;
  fd_set fd[3], *rfd, *sfd, *efd;
  int retries = 0;
  int eintr = SOAP_MAXEINTR;
  soap->errnum = 0;
#ifndef WIN32
#if !defined(FD_SETSIZE) || defined(__QNX__) || defined(QNX)
  /* no FD_SETSIZE or select() is not MT safe on some QNX: always poll */
  if (1)
#else
  /* if fd max set size exceeded, use poll() */
  if ((int)sk >= (int)FD_SETSIZE)
#endif
#ifdef HAVE_POLL
  { struct pollfd pollfd;
    pollfd.fd = (int)sk;
    pollfd.events = 0;
    if (flags & SOAP_TCP_SELECT_RCV)
      pollfd.events |= POLLIN;
    if (flags & SOAP_TCP_SELECT_SND)
      pollfd.events |= POLLOUT;
    if (flags & SOAP_TCP_SELECT_ERR)
      pollfd.events |= POLLERR;
    if (timeout <= 0)
      timeout /= -1000; /* -usec -> ms */
    else
    { retries = timeout - 1;
      timeout = 1000;
    }
    do
    { r = poll(&pollfd, 1, timeout);
      if (r < 0 && (soap->errnum = soap_socket_errno(sk)) == SOAP_EINTR && eintr-- > 0)
        r = 0;
      else if (retries-- <= 0)
        break;
    } while (r == 0);
    if (r > 0)
    { r = 0;
      if ((flags & SOAP_TCP_SELECT_RCV) && (pollfd.revents & POLLIN))
        r |= SOAP_TCP_SELECT_RCV;
      if ((flags & SOAP_TCP_SELECT_SND) && (pollfd.revents & POLLOUT))
        r |= SOAP_TCP_SELECT_SND;
      if ((flags & SOAP_TCP_SELECT_ERR) && (pollfd.revents & POLLERR))
        r |= SOAP_TCP_SELECT_ERR;
    }
    else if (r == 0)
      soap->errnum = 0;
    return r;
  }
#else
  { soap->error = SOAP_FD_EXCEEDED;
    return -1;
  }
#endif
#endif
  if (timeout > 0)
    retries = timeout - 1;
  do
  { rfd = sfd = efd = NULL;
    if (flags & SOAP_TCP_SELECT_RCV)
    { rfd = &fd[0];
      FD_ZERO(rfd);
      FD_SET(sk, rfd);
    }
    if (flags & SOAP_TCP_SELECT_SND)
    { sfd = &fd[1];
      FD_ZERO(sfd);
      FD_SET(sk, sfd);
    }
    if (flags & SOAP_TCP_SELECT_ERR)
    { efd = &fd[2];
      FD_ZERO(efd);
      FD_SET(sk, efd);
    }
    if (timeout <= 0)
    { tv.tv_sec = -timeout / 1000000;
      tv.tv_usec = -timeout % 1000000;
    }
    else
    { tv.tv_sec = 1;
      tv.tv_usec = 0;
    }
    r = select((int)sk + 1, rfd, sfd, efd, &tv);
    if (r < 0 && (soap->errnum = soap_socket_errno(sk)) == SOAP_EINTR && eintr-- > 0)
      r = 0;
    else if (retries-- <= 0)
      break;
  } while (r == 0);
  if (r > 0)
  { r = 0;
    if ((flags & SOAP_TCP_SELECT_RCV) && FD_ISSET(sk, rfd))
      r |= SOAP_TCP_SELECT_RCV;
    if ((flags & SOAP_TCP_SELECT_SND) && FD_ISSET(sk, sfd))
      r |= SOAP_TCP_SELECT_SND;
    if ((flags & SOAP_TCP_SELECT_ERR) && FD_ISSET(sk, efd))
      r |= SOAP_TCP_SELECT_ERR;
  }
  else if (r == 0)
    soap->errnum = 0;
  return r;
}
#endif
#endif

#ifndef WITH_NOIO
#ifndef PALM_1
static int
tcp_init(struct soap *soap)
{ soap->errmode = 1;
#ifdef WIN32
  if (tcp_done)
    return 0;
  else
  { WSADATA w;
    if (WSAStartup(MAKEWORD(1, 1), &w))
      return -1;
    tcp_done = 1;
  }
#endif
  return 0;
}
#endif
#endif

#ifndef WITH_NOIO
#ifndef PALM_1
static SOAP_SOCKET
tcp_custom_connect(struct soap *soap, const char *endpoint, const char *host, int port)
{
#ifdef WITH_IPV6
  struct addrinfo hints, *res, *ressave;
#endif
  SOAP_SOCKET sk;
  int err = 0;
#ifndef WITH_LEAN
  int set = 1;
#endif
#if !defined(WITH_LEAN) || defined(WITH_OPENSSL) || defined(WITH_GNUTLS) || defined(WITH_SYSTEMSSL)
  int retries;
#endif
  if (soap_valid_socket(soap->socket))
    soap->fclosesocket(soap, soap->socket);
  soap->socket = SOAP_INVALID_SOCKET;
  if (tcp_init(soap))
  { soap->errnum = 0;
    soap_set_receiver_error(soap, tcp_error(soap), "TCP init failed in tcp_connect()", SOAP_TCP_ERROR);
    return SOAP_INVALID_SOCKET;
  }
  soap->errmode = 0;
#ifdef WITH_IPV6
  memset((void*)&hints, 0, sizeof(hints));
#ifdef WITH_IPV6_V6ONLY
  hints.ai_family = PF_INET6;
#else
  hints.ai_family = PF_UNSPEC;
#endif
#ifndef WITH_LEAN
  if ((soap->omode & SOAP_IO_UDP))
    hints.ai_socktype = SOCK_DGRAM;
  else
#endif
    hints.ai_socktype = SOCK_STREAM;
  soap->errmode = 2;
  if (soap->proxy_host)
    err = getaddrinfo(soap->proxy_host, soap_int2s(soap, soap->proxy_port), &hints, &res);
  else
    err = getaddrinfo(host, soap_int2s(soap, port), &hints, &res);
  if (err || !res)
  { soap_set_receiver_error(soap, SOAP_GAI_STRERROR(err), "getaddrinfo failed in tcp_connect()", SOAP_TCP_ERROR);
    return SOAP_INVALID_SOCKET;
  }
  ressave = res;
again:
  sk = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
  soap->error = SOAP_OK;
  soap->errmode = 0;
#else
#ifndef WITH_LEAN
again:
#endif
#ifndef WITH_LEAN
  if ((soap->omode & SOAP_IO_UDP))
    sk = socket(AF_INET, SOCK_DGRAM, 0);
  else
#endif
    sk = socket(AF_INET, SOCK_STREAM, 0);
#endif
  //raymarine 10.x networking fix
    uint32_t ipAddrToBind = 0;
    if (host != nullptr && sk != -1)
    {
        bool isLegacy = false;
        if (CIpInfo::isLegacyAddressIp(host))
        {
          int flag = CIpInfo::c_markValue;
          if (-1 == setsockopt(sk, SOL_SOCKET, SO_MARK, &flag, sizeof(flag)))
          {
              RAYERROR("setsockopt fail SO_MARK: %d", errno);
          }

          std::string eth0 = "eth0";

          if (-1 == setsockopt(sk, SOL_SOCKET, SO_BINDTODEVICE, eth0.c_str(), eth0.size()))
          {
              RAYERROR("setsockopt fail SO_BINDTODEVICE: %d", errno);
          }

          isLegacy = true;
        }

       uint32_t IpNewSubnet = 0;
       uint32_t IpLegacySubnet = 0;

       if (CDiscoveryService::m_LocalIpAddress.size() > 0 && CDiscoveryService::m_LocalIpAddressLegacy.size() > 0)
       {
           IpNewSubnet = ntohl(inet_addr(CDiscoveryService::m_LocalIpAddress.c_str()));
           IpLegacySubnet = ntohl(inet_addr(CDiscoveryService::m_LocalIpAddressLegacy.c_str()));
       }

        if ((IpLegacySubnet != 0 && isLegacy) || (IpNewSubnet != 0 && CIpInfo::isNewAddressIp(host)))
        {
            int opt = 1;
            int e = setsockopt(sk, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt));
            if (e < 0)
            {
                RAYERROR("Error set SO_REUSEADDR");
            }

            ipAddrToBind = isLegacy ? IpLegacySubnet : IpNewSubnet;

            if (soap->client_port >= 0) //we will do bind later
            {
                sockaddr_in localAddr;
                memset(&localAddr, 0, sizeof(localAddr));
                localAddr.sin_family = AF_INET;         /* Set family to internet */
                localAddr.sin_addr.s_addr = htonl(ipAddrToBind);/* Set the local IP address */
                localAddr.sin_port = htons(port);

                int nRetBind = 0;
                nRetBind = bind(sk, (struct sockaddr*)&(localAddr), sizeof(localAddr));
                if (nRetBind < 0)
                {
                    RAYERROR("Error bind ip addr to socket");
                }
            }
        }
    }
  //raymarine 10.x networking fix

  if (!soap_valid_socket(sk))
  {
#ifdef WITH_IPV6
    if (res->ai_next)
    { res = res->ai_next;
      goto again;
    }
#endif
    soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "socket failed in tcp_connect()", SOAP_TCP_ERROR);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
#ifdef SOCKET_CLOSE_ON_EXIT
#ifdef WIN32
#ifndef UNDER_CE
  SetHandleInformation((HANDLE)sk, HANDLE_FLAG_INHERIT, 0);
#endif
#else
  fcntl(sk, F_SETFD, 1);
#endif
#endif
#ifndef WITH_LEAN
  if ((soap->connect_flags & SO_LINGER))
  { struct linger linger;
    memset((void*)&linger, 0, sizeof(linger));
    linger.l_onoff = 1;
    linger.l_linger = soap->linger_time;
    if (setsockopt(sk, SOL_SOCKET, SO_LINGER, (char*)&linger, sizeof(struct linger)))
    { soap->errnum = soap_socket_errno(sk);
      soap_set_receiver_error(soap, tcp_error(soap), "setsockopt SO_LINGER failed in tcp_connect()", SOAP_TCP_ERROR);
      soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
      freeaddrinfo(ressave);
#endif
      return SOAP_INVALID_SOCKET;
    }
  }
  if ((soap->connect_flags & ~SO_LINGER) && setsockopt(sk, SOL_SOCKET, soap->connect_flags & ~SO_LINGER, (char*)&set, sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
#ifndef UNDER_CE
  if ((soap->keep_alive || soap->tcp_keep_alive) && setsockopt(sk, SOL_SOCKET, SO_KEEPALIVE, (char*)&set, sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt SO_KEEPALIVE failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
  if (soap->sndbuf > 0 && setsockopt(sk, SOL_SOCKET, SO_SNDBUF, (char*)&soap->sndbuf, sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt SO_SNDBUF failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
  if (soap->rcvbuf > 0 && setsockopt(sk, SOL_SOCKET, SO_RCVBUF, (char*)&soap->rcvbuf, sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt SO_RCVBUF failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
#ifdef TCP_KEEPIDLE
  if (soap->tcp_keep_idle && setsockopt((SOAP_SOCKET)sk, IPPROTO_TCP, TCP_KEEPIDLE, (char*)&(soap->tcp_keep_idle), sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt TCP_KEEPIDLE failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, (SOAP_SOCKET)sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
#endif
#ifdef TCP_KEEPINTVL
  if (soap->tcp_keep_intvl && setsockopt((SOAP_SOCKET)sk, IPPROTO_TCP, TCP_KEEPINTVL, (char*)&(soap->tcp_keep_intvl), sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt TCP_KEEPINTVL failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, (SOAP_SOCKET)sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
#endif
#ifdef TCP_KEEPCNT
  if (soap->tcp_keep_cnt && setsockopt((SOAP_SOCKET)sk, IPPROTO_TCP, TCP_KEEPCNT, (char*)&(soap->tcp_keep_cnt), sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt TCP_KEEPCNT failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, (SOAP_SOCKET)sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
#endif
#ifdef TCP_NODELAY
  if (!(soap->omode & SOAP_IO_UDP) && setsockopt(sk, IPPROTO_TCP, TCP_NODELAY, (char*)&set, sizeof(int)))
  { soap->errnum = soap_socket_errno(sk);
    soap_set_receiver_error(soap, tcp_error(soap), "setsockopt TCP_NODELAY failed in tcp_connect()", SOAP_TCP_ERROR);
    soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
    freeaddrinfo(ressave);
#endif
    return SOAP_INVALID_SOCKET;
  }
#endif
#ifdef WITH_IPV6
  if ((soap->omode & SOAP_IO_UDP) && soap->ipv6_multicast_if)
  { struct sockaddr_in6 *in6addr = (struct sockaddr_in6*)res->ai_addr;
    in6addr->sin6_scope_id = soap->ipv6_multicast_if;
  }
#endif
#ifdef IP_MULTICAST_TTL
  if ((soap->omode & SOAP_IO_UDP))
  { if (soap->ipv4_multicast_ttl)
    { unsigned char ttl = soap->ipv4_multicast_ttl;
      if (setsockopt(sk, IPPROTO_IP, IP_MULTICAST_TTL, (char*)&ttl, sizeof(ttl)))
      { soap->errnum = soap_socket_errno(sk);
        soap_set_receiver_error(soap, tcp_error(soap), "setsockopt IP_MULTICAST_TTL failed in tcp_connect()", SOAP_TCP_ERROR);
        soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
        freeaddrinfo(ressave);
#endif
        return SOAP_INVALID_SOCKET;
      }
    }
    if ((soap->omode & SOAP_IO_UDP) && soap->ipv4_multicast_if && !soap->ipv6_multicast_if)
    { if (setsockopt(sk, IPPROTO_IP, IP_MULTICAST_IF, (char*)soap->ipv4_multicast_if, sizeof(struct in_addr)))
#ifndef WINDOWS
      { soap->errnum = soap_socket_errno(sk);
        soap_set_receiver_error(soap, tcp_error(soap), "setsockopt IP_MULTICAST_IF failed in tcp_connect()", SOAP_TCP_ERROR);
        soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
        freeaddrinfo(ressave);
#endif
        return SOAP_INVALID_SOCKET;
      }
#else
#ifndef IP_MULTICAST_IF
#define IP_MULTICAST_IF 2
#endif
      if (setsockopt(sk, IPPROTO_IP, IP_MULTICAST_IF, (char*)soap->ipv4_multicast_if, sizeof(struct in_addr)))
      { soap->errnum = soap_socket_errno(sk);
        soap_set_receiver_error(soap, tcp_error(soap), "setsockopt IP_MULTICAST_IF failed in tcp_connect()", SOAP_TCP_ERROR);
        soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
        freeaddrinfo(ressave);
#endif
        return SOAP_INVALID_SOCKET;
      }
#endif
    }
  }
#endif
#endif
#endif
  DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Opening socket=%d to host='%s' port=%d\n", (int)sk, host, port));
#ifndef WITH_IPV6
  soap->peerlen = sizeof(soap->peer.in);
  memset((void*)&soap->peer.in, 0, sizeof(soap->peer.in));
  soap->peer.in.sin_family = AF_INET;
  if (soap->client_port >= 0)
  { struct sockaddr_in addr;
    memset((void*)&addr, 0, sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_port = htons(soap->client_port);
    //raymarine 10x arhitecture
    if (ipAddrToBind != 0)
    {
        addr.sin_addr.s_addr = htonl(ipAddrToBind);/* Set the local IP address */
    }
    //raymarine 10x arhitecture
    if (bind(sk, (struct sockaddr*)&addr, sizeof(addr)))
    { soap->errnum = soap_socket_errno(sk);
      DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Could not bind before connect\n"));
      soap_set_receiver_error(soap, tcp_error(soap), "bind failed in tcp_connect()", SOAP_TCP_ERROR);
      soap->fclosesocket(soap, sk);
      return SOAP_INVALID_SOCKET;
    }
    soap->client_port = -1; /* disable bind before connect, so explicitly need to set client_port before next connect */
  }
  soap->errmode = 2;
  if (soap->proxy_host)
  { if (soap->fresolve(soap, soap->proxy_host, &soap->peer.in.sin_addr))
    { soap_set_receiver_error(soap, tcp_error(soap), "get proxy host by name failed in tcp_connect()", SOAP_TCP_ERROR);
      soap->fclosesocket(soap, sk);
      return SOAP_INVALID_SOCKET;
    }
    soap->peer.in.sin_port = htons((short)soap->proxy_port);
  }
  else
  { if (soap->fresolve(soap, host, &soap->peer.in.sin_addr))
    { soap_set_receiver_error(soap, tcp_error(soap), "get host by name failed in tcp_connect()", SOAP_TCP_ERROR);
      soap->fclosesocket(soap, sk);
      return SOAP_INVALID_SOCKET;
    }
    soap->peer.in.sin_port = htons((short)port);
  }
  soap->errmode = 0;
#ifndef WITH_LEAN
  if ((soap->omode & SOAP_IO_UDP))
    return sk;
#endif
#else
  if (soap->client_port >= 0)
  { struct sockaddr_in6 addr;
    memset((void*)&addr, 0, sizeof(addr));
    addr.sin6_family = AF_INET6;
    addr.sin6_port = htons(soap->client_port);
    if (bind(sk, (struct sockaddr*)&addr, sizeof(addr)))
    { soap->errnum = soap_socket_errno(sk);
      DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Could not bind before connect\n"));
      soap_set_receiver_error(soap, tcp_error(soap), "bind failed in tcp_connect()", SOAP_TCP_ERROR);
      soap->fclosesocket(soap, sk);
      return SOAP_INVALID_SOCKET;
    }
    soap->client_port = -1; /* disable bind before connect, so explicitly need to set client_port before next connect */
  }
  if ((soap->omode & SOAP_IO_UDP))
  { if (soap_memcpy((void*)&soap->peer.storage, sizeof(soap->peer.storage), (const void*)res->ai_addr, res->ai_addrlen))
    { soap->error = SOAP_EOM;
      soap->fclosesocket(soap, sk);
      sk = SOAP_INVALID_SOCKET;
    }
    soap->peerlen = res->ai_addrlen;
    freeaddrinfo(ressave);
    return sk;
  }
#endif
#ifndef WITH_LEAN
  if (soap->connect_timeout)
    SOAP_SOCKNONBLOCK(sk)
  else
    SOAP_SOCKBLOCK(sk)
  retries = 10;
#endif
  for (;;)
  {
#ifdef WITH_IPV6
    if (connect(sk, res->ai_addr, (int)res->ai_addrlen))
#else
    if (connect(sk, &soap->peer.addr, sizeof(soap->peer.in)))
#endif
    { err = soap_socket_errno(sk);
#ifdef WITH_IPV6
      if (err == SOAP_ECONNREFUSED && res->ai_next)
      { soap->fclosesocket(soap, sk);
        res = res->ai_next;
        goto again;
      }
#endif
#ifndef WITH_LEAN
      if (err == SOAP_EADDRINUSE)
      { soap->fclosesocket(soap, sk);
        if (retries-- > 0)
          goto again;
      }
      else if (soap->connect_timeout && (err == SOAP_EINPROGRESS || err == SOAP_EAGAIN || err == SOAP_EWOULDBLOCK))
      { SOAP_SOCKLEN_T k;
        for (;;)
        { int r;
          r = tcp_select(soap, sk, SOAP_TCP_SELECT_SND, soap->connect_timeout);
          if (r > 0)
            break;
          if (!r)
          { DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Connect timeout\n"));
            soap_set_receiver_error(soap, "Timeout", "connect failed in tcp_connect()", SOAP_TCP_ERROR);
            soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
            if (res->ai_next)
            { res = res->ai_next;
              goto again;
            }
            freeaddrinfo(ressave);
#endif
            return SOAP_INVALID_SOCKET;
          }
          r = soap->errnum = soap_socket_errno(sk);
          if (r != SOAP_EINTR)
          { DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Could not connect to host\n"));
            soap_set_receiver_error(soap, tcp_error(soap), "connect failed in tcp_connect()", SOAP_TCP_ERROR);
            soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
            if (res->ai_next)
            { res = res->ai_next;
              goto again;
            }
            freeaddrinfo(ressave);
#endif
            return SOAP_INVALID_SOCKET;
          }
        }
        k = (SOAP_SOCKLEN_T)sizeof(soap->errnum);
        if (!getsockopt(sk, SOL_SOCKET, SO_ERROR, (char*)&soap->errnum, &k) && !soap->errnum)   /* portability note: see SOAP_SOCKLEN_T definition in stdsoap2.h */
          break;
        DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Could not connect to host\n"));
        if (!soap->errnum)
          soap->errnum = soap_socket_errno(sk);
        soap_set_receiver_error(soap, tcp_error(soap), "connect failed in tcp_connect()", SOAP_TCP_ERROR);
        soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
        if (res->ai_next)
        { res = res->ai_next;
          goto again;
        }
        freeaddrinfo(ressave);
#endif
        return SOAP_INVALID_SOCKET;
      }
#endif
#ifdef WITH_IPV6
      if (res->ai_next)
      { res = res->ai_next;
        soap->fclosesocket(soap, sk);
        goto again;
      }
#endif
      if (err && err != SOAP_EINTR)
      { soap->errnum = err;
        DBGLOG(TEST, SOAP_MESSAGE(fdebug, "Could not connect to host\n"));
        soap_set_receiver_error(soap, tcp_error(soap), "connect failed in tcp_connect()", SOAP_TCP_ERROR);
        soap->fclosesocket(soap, sk);
#ifdef WITH_IPV6
        freeaddrinfo(ressave);
#endif
        return SOAP_INVALID_SOCKET;
      }
    }
    else
      break;
  }
#ifdef WITH_IPV6
  soap->peerlen = 0; /* IPv6: already connected so use send() */
  freeaddrinfo(ressave);
#endif
  soap->socket = sk;
  soap->imode &= ~SOAP_ENC_SSL;
  soap->omode &= ~SOAP_ENC_SSL;
  if (endpoint && !soap_tag_cmp(endpoint, "https:*"))
  {
    soap->fclosesocket(soap, sk);
    soap->error = SOAP_SSL_ERROR;
    return SOAP_INVALID_SOCKET;
  }
  if (soap->recv_timeout || soap->send_timeout)
    SOAP_SOCKNONBLOCK(sk)
  else
    SOAP_SOCKBLOCK(sk)
  return sk;
}
#endif
#endif

CCameraProxy::CCameraProxy() :
    m_NamePollListIter(m_NodeMap.begin()),
    m_ForceReconfigure(getenv("ONVIF_FORCE_RECONFIGURE") != 0)
{
    m_CameraNamePollTimer.setSingleShot(false);
    bool Connected = connect(&m_CameraNamePollTimer, SIGNAL(timeout()), SLOT(_PollCameraName()));
    RAYASSERT(Connected);
}

bool CCameraProxy::GetMacAddress(const std::string& OnvifNode, std::string& rMacAddress)
{
    RAYTRACE("");
    // MAC Address
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        OnvifNode_t& rNode = Iter->second;
        _tds__GetNetworkInterfaces NetworkInfo;
        _tds__GetNetworkInterfacesResponse NetworkInfoResponse;

        rNode.m_pDeviceService->reset();

        _SetTimeouts(rNode.m_pDeviceService.get());
        _ApplySoapObjProps(rNode.m_pDeviceService.get(), OnvifNode);


        int rc = rNode.m_pDeviceService->GetNetworkInterfaces(OnvifNode.c_str(), 0, &NetworkInfo, NetworkInfoResponse);
        if (rc == SOAP_OK)
        {
            for (tt__NetworkInterface* pNetworkInfo : NetworkInfoResponse.NetworkInterfaces)
            {
                rMacAddress = pNetworkInfo->Info->HwAddress;
                Success = true;
            }
        }
        else
        {
            RAYERROR("%s: Error getting MAC Address: %d", OnvifNode.c_str(), rc);
            LogSoapError(rNode.m_pDeviceService.get());
        }
    }
    else
    {
        RAYERROR("Unknown Onvif node : %s", OnvifNode.c_str());
    }
    return Success;
}

bool CCameraProxy::GetIPAddress(const std::string& OnvifNode, std::string& rIPAddress)
{
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        OnvifNode_t& rNode = Iter->second;
        _tds__GetNetworkInterfaces NetworkInfo;
        _tds__GetNetworkInterfacesResponse NetworkInfoResponse;

        rNode.m_pDeviceService->reset();
        _SetTimeouts(rNode.m_pDeviceService.get());
        _ApplySoapObjProps(rNode.m_pDeviceService.get(), OnvifNode);

        int rc = rNode.m_pDeviceService->GetNetworkInterfaces(OnvifNode.c_str(), 0, &NetworkInfo, NetworkInfoResponse);
        if (rc == SOAP_OK)
        {
            for (tt__NetworkInterface* pNetworkInfo : NetworkInfoResponse.NetworkInterfaces)
            {
                rIPAddress = pNetworkInfo->IPv4->Config->FromDHCP->Address;
                Success = true;
            }
        }
        else
        {
            RAYERROR("%s: Error getting IP Address: %d", OnvifNode.c_str(), rc);
            LogSoapError(rNode.m_pDeviceService.get());
        }
    }
    else
    {
        RAYERROR("Unknown Onvif node : %s", OnvifNode.c_str());
    }
    return Success;
}

bool CCameraProxy::GetManufacturer(const std::string& OnvifNode, std::string& rManufacturer)
{
    RAYTRACE("");
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        rManufacturer = Iter->second.m_Manufacturer;
        Success = true;
    }
    return Success;
}

bool CCameraProxy::GetModel(const std::string& OnvifNode, std::string& rModel)
{
    RAYTRACE("");
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        rModel = Iter->second.m_Model;
        Success = true;
    }
    return Success;

}
bool CCameraProxy::GetFirmwareVersion(const std::string& OnvifNode, std::string& rFirmwareVersion)
{
    RAYTRACE("");
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        rFirmwareVersion = Iter->second.m_FirmwareVersion;
        Success = true;
    }
    return Success;

}
bool CCameraProxy::GetSerialNumber(const std::string& OnvifNode, std::string& rSerialNumber)
{
    RAYTRACE("");
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        rSerialNumber = Iter->second.m_SerialNumber;
        Success = true;
    }
    return Success;

}

bool CCameraProxy::GetBrightnessColourContrast(const std::string& OnvifNode, int& rBrightness, int& rColour, int& rContrast)
{
    RAYTRACE("");
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        OnvifNode_t& rNode = Iter->second;

        _timg__GetImagingSettings ImagingSettingsGet;
        ImagingSettingsGet.VideoSourceToken = rNode.m_VideoSourceToken;

        _timg__GetImagingSettingsResponse ImagingSettingsGetResponse;

        rNode.m_pImagingService->reset();
       _SetTimeouts(rNode.m_pImagingService.get());
       _ApplySoapObjProps(rNode.m_pImagingService.get(), OnvifNode);

        if (rNode.m_pImagingService->GetImagingSettings(OnvifNode.c_str(), 0, &ImagingSettingsGet, ImagingSettingsGetResponse) == SOAP_OK)
        {
            tt__ImagingSettings20* pImagingSettings = ImagingSettingsGetResponse.ImagingSettings;
            if (pImagingSettings != 0)
            {
                if (pImagingSettings->Brightness != 0)
                {
                    rBrightness = static_cast<int>(_ScaleRangeToPercent(*pImagingSettings->Brightness, rNode.m_BrightnessRange));
                    Success = true;
                }
                if (pImagingSettings->ColorSaturation != 0)
                {
                    rColour = static_cast<int>(_ScaleRangeToPercent(*pImagingSettings->ColorSaturation, rNode.m_ColourRange));
                    Success = true;
                }
                if (pImagingSettings->Contrast != 0)
                {
                    rContrast = static_cast<int>(_ScaleRangeToPercent(*pImagingSettings->Contrast, rNode.m_ContrastRange));
                    Success = true;
                }
            }
        }
    }

    return Success;
}


 bool CCameraProxy::GetName(const std::string& OnvifNode, std::string& rName)
{
    RAYTRACE("");
    // hostname for now
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        OnvifNode_t& rNode = Iter->second;

        _tds__GetHostname GetHostName;
        _tds__GetHostnameResponse GetHostNameResponse;
        rNode.m_pDeviceService->reset();
        _SetTimeouts(rNode.m_pDeviceService.get());
        _ApplySoapObjProps(rNode.m_pDeviceService.get(), OnvifNode);
        int rc = rNode.m_pDeviceService->GetHostname(OnvifNode.c_str(), 0, &GetHostName, GetHostNameResponse);
        if (rc == SOAP_OK || rc == 200)
        {
            if (GetHostNameResponse.HostnameInformation)
            {
                std::string EncodedName = *GetHostNameResponse.HostnameInformation->Name;
                std::string DecodedName;
                if (DecodeName(EncodedName, DecodedName))
                {
                    rName = DecodedName;
                    Success = true;
                }
                else
                {
                    RAYERROR("Can't decode name %s", EncodedName.c_str());
                }
            }
            else
            {
                RAYERROR("GetHostNameResponse.HostnameInformation is null");
            }
        }
        else
        {
            RAYERROR("%s: Error getting hostname: %d", OnvifNode.c_str(), rc);
            LogSoapError(rNode.m_pDeviceService.get());
        }
    }
    else
    {
        RAYERROR("Unknown Onvif node : %s", OnvifNode.c_str());
    }

    return Success;
}

static float GetNearestResolution(const std::vector<tt__VideoResolution * >& Resolutions, int PreferredWidth, int PreferredHeight, int MinHeight, int& rNearestWidth, int& rNearestHeight)
{
    RAYTRACE("");
    double Delta = 10.0;
    double PreferredAspectRatio = static_cast<double>(PreferredWidth) / static_cast<double>(PreferredHeight);
    for (tt__VideoResolution* pRes : Resolutions)
    {
        if (pRes)
        {
            double WidthRatio = fabs(PreferredWidth - pRes->Width) / static_cast<double>(PreferredWidth);
            double HeightRatio = fabs(PreferredHeight - pRes->Height) / static_cast<double>(PreferredHeight);
            double AspectRatio = static_cast<double>(pRes->Width) / static_cast<double>(pRes->Height);
            double AspectRatioRatio = fabs(PreferredAspectRatio - AspectRatio) / PreferredAspectRatio;
            double ThisDelta = WidthRatio + HeightRatio  + AspectRatioRatio;
            if (pRes->Height < MinHeight)
            {
                ThisDelta += 0.5;
            }
            if (ThisDelta < Delta )
            {
                rNearestWidth = pRes->Width;
                rNearestHeight = pRes->Height;
                Delta = ThisDelta;
            }
        }
    }
    return static_cast<float>(Delta);
}

bool CCameraProxy::GetRtspUri(const std::string& OnvifNode, bool AutoConfigure, std::string& rHiResRtspUri, std::string& rLoResRtspUri)
{
    RAYTRACE("");

    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        OnvifNode_t& rNode = Iter->second;

        // Get Profiles
        _trt__GetProfiles ProfilesGet;
        _trt__GetProfilesResponse ProfilesGetResponse;

        rNode.m_pMediaService->reset();
        _SetTimeouts(rNode.m_pMediaService.get());
        _ApplySoapObjProps(rNode.m_pMediaService.get(), OnvifNode);
        int rc = rNode.m_pMediaService->GetProfiles(OnvifNode.c_str(), 0, &ProfilesGet, ProfilesGetResponse);
        if (rc == SOAP_OK)
        {
            // Clear any previous selections
            rNode.m_HiResProfileToken.clear();
            rNode.m_LoResProfileToken.clear();
            rNode.m_HiResVideoEncoderToken.clear();
            rNode.m_LoResVideoEncoderToken.clear();

            // Check if we have already configured profiles
            for (tt__Profile* pProfile : ProfilesGetResponse.Profiles)
            {
                RAYDEBUG("Profile: %s, Token: %s, fixed=%d", pProfile->Name.c_str(), pProfile->token.c_str(), pProfile->fixed ? *(pProfile->fixed) : -1);

                if (m_ForceReconfigure && AutoConfigure)
                {
                    RAYDEBUG("Forced reconfigure of video encoders");
                }
                else
                {
                    if (!_CheckForExistingProfile(pProfile, c_HiResProfile.Name, rNode.m_HiResProfileToken, rNode.m_HiResVideoEncoderToken) &&
                        !_CheckForExistingProfile(pProfile, c_LoResProfile.Name, rNode.m_LoResProfileToken, rNode.m_LoResVideoEncoderToken))
                    {
#ifdef ERASE_EXISTING_PROFILES
                        if (AutoConfigure && (!pProfile->fixed || !*(pProfile->fixed)))
                        {
                            _trt__DeleteProfile DeleteProfile;
                            _trt__DeleteProfileResponse DeleteProfileResponse;
                            DeleteProfile.ProfileToken = (*itr)->token;
                            if (rNode.m_pMediaService->DeleteProfile(&DeleteProfile, &DeleteProfileResponse) == SOAP_OK)
                            {
                                delete pProfile;
                                *itr = 0;
                            }
                            else
                            {
                                RAYERROR("Failed to delete profile %s", DeleteProfile.ProfileToken.c_str());
                                LogSoapError(rNode.m_pMediaService.get());
                            }
                        }
#endif
                    }
                }
            }

            if (AutoConfigure)
            {
                // Configure a Hi Res Profile
                if (rNode.m_HiResProfileToken.empty())
                {
                    _ConfigureProfile(OnvifNode, rNode, c_HiResProfile, ProfilesGetResponse.Profiles, rNode.m_HiResProfileToken, rNode.m_HiResVideoEncoderToken);
                }

                // Configure a Lo Res Profile
                if (rNode.m_LoResProfileToken.empty())
                {
                    _ConfigureProfile(OnvifNode, rNode, c_LoResProfile, ProfilesGetResponse.Profiles, rNode.m_LoResProfileToken, rNode.m_LoResVideoEncoderToken);
                }
            }

            if (!rNode.m_HiResProfileToken.empty() && !rNode.m_LoResProfileToken.empty())
            {
                if (rNode.m_HiResVideoEncoderToken == rNode.m_LoResVideoEncoderToken)
                {
                    RAYFATAL("m_HiResVideoEncoderToken=%s, LoResVideoEncoderToken=%s", rNode.m_HiResVideoEncoderToken.c_str(), rNode.m_LoResVideoEncoderToken.c_str());
                }
            }

            if (rNode.m_HiResVideoEncoderToken.empty())
            {
                // Fallback to just finding best existing H264 profile
                const int c_MaxHeight = 1080;
                const int c_MinHeight = 480;
                rNode.m_HiResProfileToken = _FindBestH264Profile(ProfilesGetResponse.Profiles, c_MinHeight, c_MaxHeight, rNode.m_HiResVideoEncoderToken);
            }

            RAYDEBUG("HiRes profile: %s, encoder: %s", rNode.m_HiResProfileToken.c_str(), rNode.m_HiResVideoEncoderToken.c_str());
            RAYDEBUG("LoRes profile: %s, encoder: %s", rNode.m_LoResProfileToken.c_str(), rNode.m_LoResVideoEncoderToken.c_str());
        }
        else
        {
            RAYERROR("%s: Error getting profiles: %d", OnvifNode.c_str(), rc);
        }

        // Get RTSP URIs
        if (!rNode.m_HiResProfileToken.empty())
        {
            rHiResRtspUri = _GetStreamUri(OnvifNode, *(rNode.m_pMediaService), rNode.m_HiResProfileToken);
            Success = !rHiResRtspUri.empty();
        }
        if (!rNode.m_LoResProfileToken.empty())
        {
            rLoResRtspUri = _GetStreamUri(OnvifNode, *(rNode.m_pMediaService), rNode.m_LoResProfileToken);
            Success = Success || !rLoResRtspUri.empty();
        }
        rNode.m_pDeviceService->reset();
        _ApplySoapObjProps(rNode.m_pDeviceService.get(), OnvifNode);
        rNode.m_pMediaService->reset();
        _ApplySoapObjProps(rNode.m_pMediaService.get(), OnvifNode);
        rNode.m_pImagingService->reset();
        _ApplySoapObjProps(rNode.m_pImagingService.get(), OnvifNode);
        _SetTimeouts(rNode.m_pDeviceService.get());
        _SetTimeouts(rNode.m_pMediaService.get());
        _SetTimeouts(rNode.m_pImagingService.get());
    }
    else
    {
        RAYERROR("Unknown Onvif node : %s", OnvifNode.c_str());
    }


    return Success;
}

bool CCameraProxy::AddNode(const std::string& OnvifNode)
{
    RAYTRACE("");
    bool Success = false;

    bool isNewNode = (m_NodeMap.find(OnvifNode) == m_NodeMap.end());
    if (!isNewNode) return false;

    RAYDEBUG("Node: %s", OnvifNode.c_str());

    OnvifNode_t Node;
    Node.m_pDeviceService.reset(new DeviceBindingProxy(OnvifNode.c_str()));
    _ApplySoapObjProps(Node.m_pDeviceService.get(), OnvifNode);
    Node.m_pMediaService.reset(new MediaBindingProxy(OnvifNode.c_str()));
    _ApplySoapObjProps(Node.m_pMediaService.get(), OnvifNode);
    Node.m_pImagingService.reset(new ImagingBindingProxy(OnvifNode.c_str()));
    _ApplySoapObjProps(Node.m_pImagingService.get(), OnvifNode);
    _SetLongTimeouts(Node.m_pDeviceService.get());
    _SetLongTimeouts(Node.m_pMediaService.get());
    _SetLongTimeouts(Node.m_pImagingService.get());

    _tds__GetDeviceInformation DeviceInfoGet;
    _tds__GetDeviceInformationResponse DeviceInfoGetResponse;

    Node.m_pDeviceService->reset();
    _SetTimeouts(Node.m_pDeviceService.get());
    _ApplySoapObjProps(Node.m_pDeviceService.get(), OnvifNode);

    int GetDeviceInformationReturnCode = Node.m_pDeviceService->GetDeviceInformation(OnvifNode.c_str(), 0, &DeviceInfoGet, DeviceInfoGetResponse);
    if (GetDeviceInformationReturnCode == SOAP_OK)
    {
        Node.m_Manufacturer = DeviceInfoGetResponse.Manufacturer;
        Node.m_Model = DeviceInfoGetResponse.Model;
        Node.m_FirmwareVersion = DeviceInfoGetResponse.FirmwareVersion;
        Node.m_SerialNumber = DeviceInfoGetResponse.SerialNumber;
        RAYDEBUG("Detected Manufacturer: %s, Model: %s", Node.m_Manufacturer.c_str(), Node.m_Model.c_str());
    }
    else
    {
        RAYERROR("GetDeviceInformation returned %d", GetDeviceInformationReturnCode);
        LogSoapError(Node.m_pDeviceService.get());
    }

    _trt__GetVideoSources mediaInfo;
    _trt__GetVideoSourcesResponse mediaInfoResponse;

    if (Node.m_pMediaService->GetVideoSources(OnvifNode.c_str(), 0, &mediaInfo, mediaInfoResponse) == SOAP_OK)
    {
        for (auto& item : mediaInfoResponse.VideoSources)
        {
            RAYDEBUG("Found Video Source %s, Resolution: %dx%d", item->token.c_str(), item->Resolution->Width, item->Resolution->Height);
        }

        tt__VideoSource* pSource = mediaInfoResponse.VideoSources.front();
        if (pSource)
        {
            Node.m_VideoSourceToken = pSource->token;
            RAYDEBUG("Using Video Source %s", Node.m_VideoSourceToken.c_str());
        }
    }

    _trt__GetGuaranteedNumberOfVideoEncoderInstances GetGuaranteedNumberOfVideoEncoderInstances;
    GetGuaranteedNumberOfVideoEncoderInstances.ConfigurationToken = Node.m_VideoSourceToken;
    _trt__GetGuaranteedNumberOfVideoEncoderInstancesResponse GetGuaranteedNumberOfVideoEncoderInstancesResponse;
    if (Node.m_pMediaService->GetGuaranteedNumberOfVideoEncoderInstances(OnvifNode.c_str(), 0, &GetGuaranteedNumberOfVideoEncoderInstances, GetGuaranteedNumberOfVideoEncoderInstancesResponse) == SOAP_OK)
    {
        if (GetGuaranteedNumberOfVideoEncoderInstancesResponse.H264)
        {
            Node.m_MaxH264Instances = *GetGuaranteedNumberOfVideoEncoderInstancesResponse.H264;
            RAYINFO("MaxEncoders: %d, MaxH264Encoders: %zu", GetGuaranteedNumberOfVideoEncoderInstancesResponse.TotalNumber, Node.m_MaxH264Instances);
        }
        else
        {
            Node.m_MaxH264Instances = GetGuaranteedNumberOfVideoEncoderInstancesResponse.TotalNumber;
            RAYINFO("MaxEncoders: %d", GetGuaranteedNumberOfVideoEncoderInstancesResponse.TotalNumber);
        }
    }

#ifdef FUTURE
    _trt__GetServiceCapabilities mediaServiceCapabilities;
    _trt__GetServiceCapabilitiesResponse mediaServiceCapabilitiesResponse;
    if (Node.m_pMediaService->GetServiceCapabilities(OnvifNode.c_str(), 0, &mediaServiceCapabilities, mediaServiceCapabilitiesResponse) == SOAP_OK)
    {
        if (mediaServiceCapabilitiesResponse.Capabilities)
        {
            if (mediaServiceCapabilitiesResponse.Capabilities->ProfileCapabilities &&
                mediaServiceCapabilitiesResponse.Capabilities->ProfileCapabilities->MaximumNumberOfProfiles)
            {
                Node.m_MaxProfiles = *(mediaServiceCapabilitiesResponse.Capabilities->ProfileCapabilities->MaximumNumberOfProfiles);
                RAYDEBUG("MaximumNumberOfProfiles=%d", Node.m_MaxProfiles);
            }
            else
            {
                RAYERROR("MaximumNumberOfProfiles capability not reported");
            }
        }
        else
        {
            RAYERROR("Media GetServiceCapabilities NULL Capabilities");
        }
    }
    else
    {
        RAYERROR("Media GetServiceCapabilities failed");
    }
#endif

#ifdef DUMP_ENCODER_CONFIGS
    _DumpEncoders(*Node.m_pMediaService);
#endif

    // Get imaging options to correctly scale brightness, colour and contrast settings
    _timg__GetOptions imagingOptions;
    imagingOptions.VideoSourceToken = Node.m_VideoSourceToken;
    _timg__GetOptionsResponse imagingOptionsResponse;
    if (Node.m_pImagingService->GetOptions(OnvifNode.c_str(), 0, &imagingOptions, imagingOptionsResponse) == SOAP_OK)
    {
        if (imagingOptionsResponse.ImagingOptions)
        {
            if (imagingOptionsResponse.ImagingOptions->Brightness)
            {
                RAYDEBUG("Supports Brightness Range: %0.1f - %0.1f", imagingOptionsResponse.ImagingOptions->Brightness->Min, imagingOptionsResponse.ImagingOptions->Brightness->Max);
                Node.m_BrightnessRange.Supported = true;
                Node.m_BrightnessRange.Min = imagingOptionsResponse.ImagingOptions->Brightness->Min;
                Node.m_BrightnessRange.Max = imagingOptionsResponse.ImagingOptions->Brightness->Max;
            }
            if (imagingOptionsResponse.ImagingOptions->Contrast)
            {
                RAYDEBUG("Supports Contrast Range: %0.1f - %0.1f", imagingOptionsResponse.ImagingOptions->Contrast->Min, imagingOptionsResponse.ImagingOptions->Contrast->Max);
                Node.m_ContrastRange.Supported = true;
                Node.m_ContrastRange.Min = imagingOptionsResponse.ImagingOptions->Contrast->Min;
                Node.m_ContrastRange.Max = imagingOptionsResponse.ImagingOptions->Contrast->Max;
            }
            if (imagingOptionsResponse.ImagingOptions->ColorSaturation)
            {
                RAYDEBUG("Supports Colour Range: %0.1f - %0.1f", imagingOptionsResponse.ImagingOptions->ColorSaturation->Min, imagingOptionsResponse.ImagingOptions->ColorSaturation->Max);
                Node.m_ColourRange.Supported = true;
                Node.m_ColourRange.Min = imagingOptionsResponse.ImagingOptions->ColorSaturation->Min;
                Node.m_ColourRange.Max = imagingOptionsResponse.ImagingOptions->ColorSaturation->Max;
            }
        }
    }
    else
    {
        RAYERROR("Imaging GetOptions failed");
        LogSoapError(Node.m_pImagingService.get());
    }

    Node.m_pDeviceService->reset();
    _ApplySoapObjProps(Node.m_pDeviceService.get(), OnvifNode);
    Node.m_pMediaService->reset();
    _ApplySoapObjProps(Node.m_pMediaService.get(), OnvifNode);
    Node.m_pImagingService->reset();
    _ApplySoapObjProps(Node.m_pImagingService.get(), OnvifNode);
    _SetTimeouts(Node.m_pDeviceService.get());
    _SetTimeouts(Node.m_pMediaService.get());
    _SetTimeouts(Node.m_pImagingService.get());
    m_NodeMap[OnvifNode] = Node;
    m_NamePollListIter = m_NodeMap.begin();
    _SetNamePollTimer();

    return true;
}

void CCameraProxy::SetNodeList(const NodeList_t& NamePollList)
{
    for (NodeList_t::const_iterator Iter = NamePollList.begin(); Iter != NamePollList.end(); ++Iter)
    {
        if (m_NodeMap.find(*Iter) == m_NodeMap.end())
        {
            AddNode(*Iter);
        }
    }

    NodeMap_t::const_iterator Iter = m_NodeMap.begin();
    while(Iter != m_NodeMap.end())
    {
        if (NamePollList.find(Iter->first) == NamePollList.end())
        {
            const std::string& Node = Iter->first;
            ++Iter;
            RemoveNode(Node);
        }
        else
        {
            ++Iter;
        }
    }
}

bool CCameraProxy::RemoveNode(const std::string& OnvifNode)
{
    RAYTRACE("");
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        bool resetIterator = (Iter == m_NamePollListIter);
        m_NodeMap.erase(Iter);
        if (resetIterator)
        {
            m_NamePollListIter = m_NodeMap.begin();
        }

        _SetNamePollTimer();
        return true;
    }

    return false;
}

// SETTERS
bool CCameraProxy::SetBrightnessColourContrast(const std::string& OnvifNode, int Brightness, int Colour, int Contrast)
{
    RAYTRACE("OnvifNode: %s, Brightness: %d, Colour: %d, Contrast: %d",
             OnvifNode.c_str(), Brightness, Colour, Contrast);
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        OnvifNode_t& rNode = Iter->second;
        _timg__SetImagingSettings ImagingSettingsSet;
        _timg__SetImagingSettingsResponse ImagingSettingsSetResponse;

        rNode.m_pImagingService->reset();
        _SetTimeouts(rNode.m_pImagingService.get());
        _ApplySoapObjProps(rNode.m_pImagingService.get(), OnvifNode);

        float NewBrightness = _ScalePercentToRange(Brightness, rNode.m_BrightnessRange);
        float NewColour = _ScalePercentToRange(Colour, rNode.m_ColourRange);
        float NewContrast = _ScalePercentToRange(Contrast, rNode.m_ContrastRange);

        tt__ImagingSettings20 ImagingSettings;
        ImagingSettings.Brightness = &NewBrightness;
        ImagingSettings.ColorSaturation = &NewColour;
        ImagingSettings.Contrast = &NewContrast;

        ImagingSettingsSet.ImagingSettings = &ImagingSettings;
        ImagingSettingsSet.VideoSourceToken = rNode.m_VideoSourceToken;

        int rc = rNode.m_pImagingService->SetImagingSettings(OnvifNode.c_str(), 0, &ImagingSettingsSet, ImagingSettingsSetResponse);

        Success = (rc == SOAP_OK);
        RAYDEBUG("OnvifNode: %s, Brightness: %d, Colour: %d, Contrast: %d - rc: %d",
                  OnvifNode.c_str(), Brightness, Colour, Contrast, rc);
    }
    else
    {
        RAYERROR("Unknown Onvif node : %s", OnvifNode.c_str());
    }

    return Success;
}


bool CCameraProxy::SetName(const std::string& OnvifNode, const std::string& NewName)
{
    RAYTRACE("");
    bool Success = false;
    NodeMap_t::iterator Iter = m_NodeMap.find(OnvifNode);
    if (Iter != m_NodeMap.end())
    {
        OnvifNode_t& rNode = Iter->second;
        _tds__SetHostname SetHostName;
        _tds__SetHostnameResponse SetHostNameResponse;

        rNode.m_pDeviceService->reset();
        _SetTimeouts(rNode.m_pDeviceService.get());
        _ApplySoapObjProps(rNode.m_pDeviceService.get(), OnvifNode);

        std::string EncodedName;
        if (EncodeName(NewName, EncodedName))
        {
            SetHostName.Name = EncodedName;

            RAYDEBUG("Node:%s ProcessedName:%s", OnvifNode.c_str(), EncodedName.c_str());
            int rc = rNode.m_pDeviceService->SetHostname(OnvifNode.c_str(), 0, &SetHostName, SetHostNameResponse);
            Success = (rc == SOAP_OK);
        }
    }
    else
    {
        RAYERROR("Unknown Onvif node : %s", OnvifNode.c_str());
    }

    return Success;
}

void CCameraProxy::_PollCameraName()
{
    RAYTRACE("");
    // Make sure the names of the cameras ars up to date
    // We poll one camera at a time to ensure we don't block the event loop
    // for long period of time waiting for SOAP call timeouts
    if (m_NamePollListIter != m_NodeMap.end())
    {
        const std::string& OnvifNode = m_NamePollListIter->first;
        // We've already polled for this one in the _GetCameraDetails function
        // so skip it
        if (OnvifNode.size() > 0)
        {
            std::string Name;
            if (GetName(OnvifNode, Name))
            {
                Q_EMIT CameraNameUpdated(OnvifNode, Name);
            }
        }

        ++m_NamePollListIter;
    }

    if (m_NamePollListIter == m_NodeMap.end())
    {
        m_NamePollListIter = m_NodeMap.begin();
    }
}

std::string CCameraProxy::_GetStreamUri(const std::string& OnvifNodeUrl, MediaBindingProxy& MediaService, const std::string& ProfileToken)
{
    RAYTRACE("");
    std::string URI;

    _trt__GetStreamUri StreamUri;
    _trt__GetStreamUriResponse StreamUriResponse;

    tt__Transport Transport;
    tt__StreamSetup StreamSetup;

    StreamUri.StreamSetup = &StreamSetup;
    StreamSetup.Transport = &Transport;

    StreamUri.StreamSetup->Stream = tt__StreamType__RTP_Unicast;
    StreamUri.StreamSetup->Transport->Protocol = tt__TransportProtocol__RTSP;
    StreamUri.StreamSetup->Transport->Tunnel = 0;
    StreamUri.ProfileToken = ProfileToken;
    RAYDEBUG("Getting RTSP URI for %s", ProfileToken.c_str());
    MediaService.reset();
    _SetTimeouts(&MediaService);
    _ApplySoapObjProps(&MediaService, OnvifNodeUrl);

    int rc = MediaService.GetStreamUri(OnvifNodeUrl.c_str(), 0, &StreamUri, StreamUriResponse);
    if (rc == SOAP_OK)
    {
        if (StreamUriResponse.MediaUri)
        {
            URI = StreamUriResponse.MediaUri->Uri;
            RAYDEBUG("Found RTSP URI: %s", URI.c_str());
        }
        else
        {
            RAYERROR("Received NULL MediaUri");
        }
    }
    else
    {
        // Workaround for an HD-SDI encoder that only gives the RTSP URI if you ask for a UDP protocol.
        RAYDEBUG("Error getting RTSP URI: %d. Trying UDP transport", rc);
        StreamUri.StreamSetup->Transport->Protocol = tt__TransportProtocol__UDP;
        rc = MediaService.GetStreamUri(OnvifNodeUrl.c_str(), 0, &StreamUri, StreamUriResponse);
        if (rc == SOAP_OK)
        {
            if (StreamUriResponse.MediaUri)
            {
                if (StreamUriResponse.MediaUri->Uri.find("rtsp:") == 0)
                {
                    URI = StreamUriResponse.MediaUri->Uri;
                    RAYINFO("Found UDP RTSP URI: %s", URI.c_str());
                }
                else
                {
                    RAYWARNING("Error didn't get RTSP URI (as UDP transport): %s", StreamUriResponse.MediaUri->Uri.c_str());
                }
            }
            else
            {
                RAYERROR("Received NULL MediaUri");
            }
        }
        else
        {
            RAYWARNING("Error getting RTSP URI (as UDP transport): %d", rc);
        }
    }
    return URI;
}

static bool ExceedMinimumResolution(const std::vector<tt__VideoResolution * >& Resolutions, int MinWidth, int MinHeight)
{
    for (tt__VideoResolution* resolution : Resolutions)
    {
        if (resolution->Width >= MinWidth && resolution->Height >= MinHeight)
        {
            return true;
        }
    }
    return false;
}

bool CCameraProxy::_CheckForExistingProfile(const tt__Profile* pProfile, const std::string& Name, std::string& rProfileToken, std::string& rEncoderToken)
{
    bool Existing = false;

    // Check if a profile already exists that uses an encoder with the correct name
    if (pProfile->VideoEncoderConfiguration &&
        pProfile->VideoEncoderConfiguration->Name == Name &&
        pProfile->VideoEncoderConfiguration->H264)
    {
        rProfileToken = pProfile->token;
        rEncoderToken = pProfile->VideoEncoderConfiguration->token;
        Existing = true;
        RAYINFO("Found existing profile %s with Encoder %s", rProfileToken.c_str(), rEncoderToken.c_str());
    }

    return Existing;
}

bool CCameraProxy::_ConfigureProfile(const std::string& OnvifNodeUrl, OnvifNode_t& Node, const EncoderPreferences_t& EncoderPreferences, std::vector<class tt__Profile*>& Profiles, std::string& rProfileToken, std::string& rEncoderToken)
{
    RAYTRACE("");
    bool Success = false;

    // Try using an existing Profile of the correct name, implying half completed configuration
    for (tt__Profile* pProfile : Profiles)
    {
        // Check for a matching name
        if (pProfile && pProfile->Name == EncoderPreferences.Name)
        {
            std::string EncoderToken = _ConfigureEncoderForProfile(OnvifNodeUrl, Node, pProfile->token, EncoderPreferences);
            if (!EncoderToken.empty())
            {
                rProfileToken = pProfile->token;
                rEncoderToken = EncoderToken;
                Success = true;
                break;
            }
        }
    }

    if (!Success && (Node.m_MaxH264Instances > Profiles.size() || Profiles.size() < 2u))
    {
        // Try to create a new Profile
        Success = _CreateNewProfile(OnvifNodeUrl, Node, EncoderPreferences, rProfileToken, rEncoderToken);
        if (!Success)
        {
            RAYINFO("Unable to create new profile for %s, trying to re-use an existing one", EncoderPreferences.Name.c_str());
        }
    }

    if (!Success)
    {
        // Just try any existing profile
        for (tt__Profile* pProfile : Profiles)
        {
            // Skip an already selected profile
            if (pProfile &&
                pProfile->token != Node.m_HiResProfileToken &&
                pProfile->token != Node.m_LoResProfileToken)
            {
                std::string EncoderToken = _ConfigureEncoderForProfile(OnvifNodeUrl, Node, pProfile->token, EncoderPreferences);
                if (!EncoderToken.empty())
                {
                    rProfileToken = pProfile->token;
                    rEncoderToken = EncoderToken;
                    Success = true;
                    break;
                }
            }
        }
    }
    return Success;
}

bool CCameraProxy::_CreateNewProfile(const std::string& OnvifNodeUrl, OnvifNode_t& Node, const EncoderPreferences_t& EncoderPreferences, std::string& rProfileToken, std::string& rEncoderToken)
{
    RAYTRACE("");
    bool Success = false;

    _trt__CreateProfile CreateProfile;
    _trt__CreateProfileResponse CreateProfileResponse;
    CreateProfile.Name = EncoderPreferences.Name;
    std::string Token = EncoderPreferences.Name;
    CreateProfile.Token = &Token;
    if (Node.m_pMediaService->CreateProfile(&CreateProfile, CreateProfileResponse) == SOAP_OK)
    {
        tt__Profile* Profile = CreateProfileResponse.Profile;
        _trt__AddVideoSourceConfiguration AddVideoSourceConfiguration;
        _trt__AddVideoSourceConfigurationResponse AddVideoSourceConfigurationResponse;
        AddVideoSourceConfiguration.ProfileToken = Profile->token;
        AddVideoSourceConfiguration.ConfigurationToken = Node.m_VideoSourceToken;
        if (Node.m_pMediaService->AddVideoSourceConfiguration(OnvifNodeUrl.c_str(), 0, &AddVideoSourceConfiguration, AddVideoSourceConfigurationResponse) == SOAP_OK)
        {
            std::string EncoderToken = _ConfigureEncoderForProfile(OnvifNodeUrl, Node, Profile->token, EncoderPreferences);
            if (!EncoderToken.empty())
            {
                rProfileToken = Profile->token;
                rEncoderToken = EncoderToken;
                Success = true;
                RAYINFO("Successfully created new profile %s with Encoder %s for %s", rProfileToken.c_str(), rEncoderToken.c_str(), EncoderPreferences.Name.c_str());
            }
        }
        else
        {
            RAYERROR("AddVideoSourceConfiguration failed for Profile %s, Source %s", Profile->token.c_str(), Node.m_VideoSourceToken.c_str());
            LogSoapError(Node.m_pMediaService.get());
        }

        if (!Success)
        {
            // If we have encountered an error try to clean up the incomplete profile
            _trt__DeleteProfile DeleteProfile;
            _trt__DeleteProfileResponse DeleteProfileResponse;
            DeleteProfile.ProfileToken = Profile->token;
            Node.m_pMediaService->DeleteProfile(&DeleteProfile, DeleteProfileResponse);
            Node.m_pMediaService->reset();
            _SetTimeouts(Node.m_pMediaService.get());
            _ApplySoapObjProps(Node.m_pMediaService.get(), OnvifNodeUrl);
        }
    }
    else
    {
        RAYERROR("CreateProfile for Name %s", EncoderPreferences.Name.c_str());
        LogSoapError(Node.m_pMediaService.get());
    }

    return Success;
}

std::string CCameraProxy::_ConfigureEncoderForProfile(const std::string& OnvifNodeUrl, OnvifNode_t& Node, const std::string& ProfileToken, const EncoderPreferences_t& EncoderPreferences)
{
    RAYTRACE("");
    std::string EncoderToken;

    _trt__GetCompatibleVideoEncoderConfigurations GetCompatibleVideoEncoderConfigurations;
    _trt__GetCompatibleVideoEncoderConfigurationsResponse GetCompatibleVideoEncoderConfigurationsResponse;
    GetCompatibleVideoEncoderConfigurations.ProfileToken = ProfileToken;
    if (Node.m_pMediaService->GetCompatibleVideoEncoderConfigurations(OnvifNodeUrl.c_str(), 0, &GetCompatibleVideoEncoderConfigurations, GetCompatibleVideoEncoderConfigurationsResponse) == SOAP_OK)
    {
        const size_t c_NumberOfEncoders = GetCompatibleVideoEncoderConfigurationsResponse.Configurations.size();
        if (c_NumberOfEncoders > 0)
        {
            struct EncoderRanking_t
            {
                bool Valid = false;
                float Ranking = 0.0f;
                int BestResolutionWidth = 0;
                int BestResolutionHeight = 0;
                float Quality = 0.0f;
                int Bitrate = 0;
                int EncodeInterval = 1;
                int Framerate = 0;
                tt__H264Profile H264Profile = tt__H264Profile__Baseline;
                bool MeetsMinResolution = false;
            };
            EncoderRanking_t EncoderRanking[c_NumberOfEncoders];
            for (size_t Index = 0u; Index < c_NumberOfEncoders; ++Index)
            {
                std::string ThisEncoderToken = GetCompatibleVideoEncoderConfigurationsResponse.Configurations[Index]->token;
                if (GetCompatibleVideoEncoderConfigurationsResponse.Configurations[Index]->Name == EncoderPreferences.Name)
                {
                    // Already one of the right name, probably an incomplete configuration from a previous run so use it
                    EncoderRanking[Index].Ranking += 2.0f;
                }

                if (GetCompatibleVideoEncoderConfigurationsResponse.Configurations[Index]->UseCount == 0)
                {
                    // Prefer an unused encoder
                    EncoderRanking[Index].Ranking += 1.0f;
                }

                if (GetCompatibleVideoEncoderConfigurationsResponse.Configurations[Index]->Encoding == tt__VideoEncoding__H264 &&
                    GetCompatibleVideoEncoderConfigurationsResponse.Configurations[Index]->H264)
                {
                    // Prefer an existing H264 encoder
                    EncoderRanking[Index].Ranking += 0.5f;
                }

                if (ThisEncoderToken == Node.m_HiResVideoEncoderToken ||
                    ThisEncoderToken == Node.m_LoResVideoEncoderToken)
                {
                    // Already in use so don't reconfigure
                    EncoderRanking[Index].Valid = false;
                }
                else
                {
                    _trt__GetVideoEncoderConfigurationOptions VideoEncoderConfigurationOptions;
                    _trt__GetVideoEncoderConfigurationOptionsResponse VideoEncoderConfigurationOptionsResponse;
                    VideoEncoderConfigurationOptions.ConfigurationToken = &ThisEncoderToken;
                    if (Node.m_pMediaService->GetVideoEncoderConfigurationOptions(OnvifNodeUrl.c_str(), 0, &VideoEncoderConfigurationOptions, VideoEncoderConfigurationOptionsResponse) == SOAP_OK)
                    {
                        if (VideoEncoderConfigurationOptionsResponse.Options && VideoEncoderConfigurationOptionsResponse.Options->H264)
                        {
                            EncoderRanking[Index].MeetsMinResolution = ExceedMinimumResolution(VideoEncoderConfigurationOptionsResponse.Options->H264->ResolutionsAvailable, EncoderPreferences.Width, EncoderPreferences.Height);
                            if (!EncoderRanking[Index].MeetsMinResolution)
                            {
                                // Doesn't meet minimum resolution so reduce ranking
                                EncoderRanking[Index].Ranking -= 0.5f;
                            }
                            // Adjust ranking based on nearest resolution
                            EncoderRanking[Index].Ranking -= GetNearestResolution(VideoEncoderConfigurationOptionsResponse.Options->H264->ResolutionsAvailable,
                                                                                  EncoderPreferences.Width,
                                                                                  EncoderPreferences.Height,
                                                                                  EncoderPreferences.MinHeight,
                                                                                  EncoderRanking[Index].BestResolutionWidth,
                                                                                  EncoderRanking[Index].BestResolutionHeight);
                            // Adjust ranking based on max framerate available
                            if (VideoEncoderConfigurationOptionsResponse.Options->H264->FrameRateRange)
                            {
                                int MaxFramerate = VideoEncoderConfigurationOptionsResponse.Options->H264->FrameRateRange->Max;
                                if (EncoderPreferences.Framerate > MaxFramerate)
                                {
                                    EncoderRanking[Index].Ranking -= static_cast<double>(EncoderPreferences.Framerate - MaxFramerate) / static_cast<double>(EncoderPreferences.Framerate);
                                }
                                EncoderRanking[Index].Framerate = std::min(MaxFramerate, EncoderPreferences.Framerate);
                            }
                            // Choose the nest profile
                            if (VideoEncoderConfigurationOptionsResponse.Options->H264->H264ProfilesSupported.size() > 0)
                            {
                                EncoderRanking[Index].Valid = true;
                                EncoderRanking[Index].H264Profile = VideoEncoderConfigurationOptionsResponse.Options->H264->H264ProfilesSupported[0];
                                for (size_t i = 1;i < VideoEncoderConfigurationOptionsResponse.Options->H264->H264ProfilesSupported.size(); ++i)
                                {
                                    if (VideoEncoderConfigurationOptionsResponse.Options->H264->H264ProfilesSupported[i] > EncoderRanking[Index].H264Profile)
                                    {
                                        EncoderRanking[Index].H264Profile = VideoEncoderConfigurationOptionsResponse.Options->H264->H264ProfilesSupported[i];
                                    }
                                }
                            }
                            // Select the appropriate quality value based on the range
                            if (VideoEncoderConfigurationOptionsResponse.Options->QualityRange)
                            {
                                EncoderRanking[Index].Quality = VideoEncoderConfigurationOptionsResponse.Options->QualityRange->Min + (VideoEncoderConfigurationOptionsResponse.Options->QualityRange->Max - VideoEncoderConfigurationOptionsResponse.Options->QualityRange->Min) * EncoderPreferences.Quality;
                            }
                            if (VideoEncoderConfigurationOptionsResponse.Options->Extension &&
                                VideoEncoderConfigurationOptionsResponse.Options->Extension->H264 &&
                                VideoEncoderConfigurationOptionsResponse.Options->Extension->H264->BitrateRange)
                            {
                                EncoderRanking[Index].Bitrate = std::min(EncoderPreferences.Bitrate, VideoEncoderConfigurationOptionsResponse.Options->Extension->H264->BitrateRange->Max);
                            }
                            else
                            {
                                EncoderRanking[Index].Bitrate = EncoderPreferences.Bitrate;
                            }
                            if (VideoEncoderConfigurationOptionsResponse.Options->H264->EncodingIntervalRange)
                            {
                                EncoderRanking[Index].EncodeInterval = std::min(1, VideoEncoderConfigurationOptionsResponse.Options->H264->EncodingIntervalRange->Max);
                            }
                        }
                    }
                }
            }

            // Determine the encoder index with the best ranking
            size_t BestIndex = c_NumberOfEncoders;
            float BestRanking = -10000.0f;
            for (size_t Index = 0u; Index < c_NumberOfEncoders; ++Index)
            {
                RAYDEBUG("Encoder %s Valid=%d, ranking %0.2f", GetCompatibleVideoEncoderConfigurationsResponse.Configurations[Index]->token.c_str(), EncoderRanking[Index].Valid, EncoderRanking[Index].Ranking);
                if (EncoderRanking[Index].Valid && EncoderRanking[Index].Ranking > BestRanking)
                {
                    BestIndex = Index;
                    BestRanking = EncoderRanking[Index].Ranking;
                }
            }

            // Having found the best encoder, we now need to configure it
            if (BestIndex != c_NumberOfEncoders)
            {
                RAYDEBUG("Chosen Encoder %s", GetCompatibleVideoEncoderConfigurationsResponse.Configurations[BestIndex]->token.c_str());

                _trt__SetVideoEncoderConfiguration SetVideoEncoderConfiguration;

                tt__VideoResolution VideoResolution;
                VideoResolution.Width = EncoderRanking[BestIndex].BestResolutionWidth;
                VideoResolution.Height = EncoderRanking[BestIndex].BestResolutionHeight;

                tt__VideoRateControl RateControl;

                if (Node.m_Manufacturer == "AXIS" && EncoderRanking[BestIndex].Quality != 0.0f)
                {
                    // Workaround for poor video performance when setting bitrate limit on Axis cameras
                    RateControl.BitrateLimit = 0;
                }
                else
                {
                    RateControl.BitrateLimit = EncoderRanking[BestIndex].Bitrate;
                }
                RateControl.EncodingInterval = EncoderRanking[BestIndex].EncodeInterval;
                RateControl.FrameRateLimit = EncoderRanking[BestIndex].Framerate;

                tt__H264Configuration H264Config;
#ifdef REDUCE_GOP_ON_VISION_HITECH
                if (Node.m_Manufacturer == "Vision Hitech" && Node.m_Model.find("VISION Mi") == 0)
                {
                    const int c_GoodGopValue = 15;
                    RAYDEBUG("Resetting GOP value to %d", c_GoodGopValue);
                    H264Config.GovLength = c_GoodGopValue;
                }
                else
                if (Node.m_Manufacturer == "RAYMARINE" && Node.m_Model.find("CAM200") == 0)
                {
                    H264Config.GovLength = 15;
                }
                else
#endif
                {
                    H264Config.GovLength = EncoderRanking[BestIndex].Framerate;
                }
                H264Config.H264Profile = EncoderRanking[BestIndex].H264Profile;

                tt__VideoEncoderConfiguration VideoEncoderConfiguration;
                VideoEncoderConfiguration.token = GetCompatibleVideoEncoderConfigurationsResponse.Configurations[BestIndex]->token;
                VideoEncoderConfiguration.Name = EncoderPreferences.Name;
                VideoEncoderConfiguration.Encoding = tt__VideoEncoding__H264;
                VideoEncoderConfiguration.Resolution = &VideoResolution;
                VideoEncoderConfiguration.Quality = EncoderRanking[BestIndex].Quality;
                VideoEncoderConfiguration.H264 = &H264Config;
                VideoEncoderConfiguration.MPEG4 = 0;
                VideoEncoderConfiguration.RateControl = &RateControl;
                VideoEncoderConfiguration.Multicast = GetCompatibleVideoEncoderConfigurationsResponse.Configurations[BestIndex]->Multicast;
                VideoEncoderConfiguration.Multicast->AutoStart = false;

                SetVideoEncoderConfiguration.Configuration = &VideoEncoderConfiguration;
                SetVideoEncoderConfiguration.ForcePersistence = true;
                _trt__SetVideoEncoderConfigurationResponse SetVideoEncoderConfigurationResponse;
                if (Node.m_pMediaService->SetVideoEncoderConfiguration(OnvifNodeUrl.c_str(), 0, &SetVideoEncoderConfiguration, SetVideoEncoderConfigurationResponse) == SOAP_OK)
                {
                    // Now add the video encoder configuration to the profile
                    _trt__AddVideoEncoderConfiguration AddVideoEncoderConfiguration;
                    _trt__AddVideoEncoderConfigurationResponse AddVideoEncoderConfigurationResponse;
                    AddVideoEncoderConfiguration.ProfileToken = ProfileToken;
                    AddVideoEncoderConfiguration.ConfigurationToken = SetVideoEncoderConfiguration.Configuration->token;
                    if (Node.m_pMediaService->AddVideoEncoderConfiguration(OnvifNodeUrl.c_str(), 0, &AddVideoEncoderConfiguration, AddVideoEncoderConfigurationResponse) == SOAP_OK)
                    {
                        EncoderToken = SetVideoEncoderConfiguration.Configuration->token;
                        RAYINFO("Successfully configured Encoder %s for Profile %s", AddVideoEncoderConfiguration.ConfigurationToken.c_str(), AddVideoEncoderConfiguration.ProfileToken.c_str());
#ifdef DUMP_ENCODER_CONFIGS
                        _DumpEncoder(&VideoEncoderConfiguration);
#endif
                    }
                    else
                    {
                        RAYERROR("AddVideoEncoderConfiguration failed, Configuration=%s to Profile=%s", AddVideoEncoderConfiguration.ConfigurationToken.c_str(), AddVideoEncoderConfiguration.ProfileToken.c_str());
                        LogSoapError(Node.m_pMediaService.get());
                    }
                }
                else
                {
                    RAYERROR("SetVideoEncoderConfiguration failed, Configuration=%s", SetVideoEncoderConfiguration.Configuration->token.c_str());
                    LogSoapError(Node.m_pMediaService.get());
                }
            }
            else
            {
                RAYWARNING("Failed to find suitable encoder for Profile %s", ProfileToken.c_str());
            }
        }
        else
        {
            RAYERROR("No compatible encoders for Profile %s", ProfileToken.c_str());
        }
    }
    else
    {
        RAYERROR("GetCompatibleVideoEncoderConfigurations failed for profile %s", ProfileToken.c_str());
        LogSoapError(Node.m_pMediaService.get());
    }

    return EncoderToken;
}

std::string CCameraProxy::_FindBestH264Profile(std::vector<class tt__Profile * > &Profiles, int MinHeight, int MaxHeight, std::string& rEncoderToken)
{
    RAYTRACE("");
    std::string ProfileToken;
    int BestHeight = 0;
    float BestQuality= 0.0f;

    for (std::vector<class tt__Profile * >::iterator itr = Profiles.begin(); itr != Profiles.end(); ++itr)
    {
        tt__Profile* pProfile = *itr;
        if (pProfile->VideoEncoderConfiguration != 0)
        {
            tt__VideoEncoderConfiguration* pVideoEncoder = pProfile->VideoEncoderConfiguration;
            _DumpEncoder(pVideoEncoder);

            if (pVideoEncoder->Encoding == tt__VideoEncoding__H264)
            {
                bool UseThisOne = false;

                if (ProfileToken.empty())
                {
                    UseThisOne = true;
                }
                else if (pVideoEncoder->Resolution->Height >= MinHeight)
                {
                    // Prefer all sources greater than minimum
                    if (BestHeight > MaxHeight &&
                        pVideoEncoder->Resolution->Height < BestHeight)
                    {
                        // When over maximum resolution try to pick a smaller resolution
                        UseThisOne = true;
                    }
                    else if (pVideoEncoder->Resolution->Height <= MaxHeight)
                    {
                        // If under our maximum resolution, pick the highest resolution
                        if (pVideoEncoder->Resolution->Height > BestHeight)
                        {
                            UseThisOne = true;
                        }
                        else if (pVideoEncoder->Resolution->Height == BestHeight &&
                                 pVideoEncoder->Quality > BestQuality)
                        {
                            // If the resolutions match then pick the best quality
                            UseThisOne = true;
                        }
                    }
                }
                else if (pVideoEncoder->Resolution->Height > BestHeight)
                {
                    // If all profiles are below our minimum select the biggest
                    UseThisOne = true;
                }

                if (UseThisOne)
                {
                    ProfileToken = pProfile->token;
                    rEncoderToken = pProfile->VideoEncoderConfiguration->token;
                    BestQuality = pVideoEncoder->Quality;
                    BestHeight = pVideoEncoder->Resolution->Height;
                }
            }
        }
    }

    return ProfileToken;
}

bool CCameraProxy::EncodeName(const std::string& PlainName, std::string& rEncodedName)
{
    rEncodedName = c_EncodingPrefix;
    for (unsigned int idx = 0; idx < PlainName.size(); ++idx)
    {
        // Encode this 8-bit character as two characters 'a' to 'p'
        char InChar = PlainName[idx];
        char HiWord = (InChar >> 4);
        char LoWord = InChar & 0x0f;
        rEncodedName += (c_MinNameChar + HiWord);
        rEncodedName += (c_MinNameChar + LoWord);
    }
    RAYDETAIL("Encoding name %s->%s", PlainName.c_str(), rEncodedName.c_str());
    return (rEncodedName.size() > 0) && ((rEncodedName.size() % 2) == 0);
}

bool CCameraProxy::DecodeName(const std::string& EncodedName, std::string& rPlainName)
{
    bool EncodedString = true;
    rPlainName = "";
    if (EncodedName.compare(0, c_EncodingPrefix.size(), c_EncodingPrefix) == 0)
    {
        for (unsigned int idx = c_EncodingPrefix.size(); EncodedString && (idx < EncodedName.size()); idx += 2)
        {
            // Decode these two characters 'a' to 'p' to a single 8-bit character
            // If either character is out of range then it's not one we've encoded, so bomb out
            if ((EncodedName[idx + 0] < c_MinNameChar) || (EncodedName[idx + 0] > c_MaxNameChar)
             || (EncodedName[idx + 1] < c_MinNameChar) || (EncodedName[idx + 1] > c_MaxNameChar))
            {
                RAYDEBUG("Bad character: {%c, %c}\n", EncodedName[idx + 0], EncodedName[idx + 1]);
                EncodedString = false;
            }
            else
            {
                char HiWord = EncodedName[idx + 0] - c_MinNameChar;
                char LoWord = EncodedName[idx + 1] - c_MinNameChar;
                char OutChar = (HiWord << 4) | LoWord;
                rPlainName += OutChar;
            }
        }
    }
    else
    {
        EncodedString = false;
    }
    // If we had a problem then we probably didn't encode this string so just return
    // the input
    if (!EncodedString)
    {
        rPlainName = EncodedName;
    }

    return true;
}

void CCameraProxy::_SetTimeouts(struct soap* s)
{
    s->connect_timeout = 3;
    s->send_timeout = 2;
    s->recv_timeout = 3;
    //s->keep_alive = 1;
}

/// It has been observed that some devices (e.g. CAM200) can take a long time to
/// get a SOAP response from.  The initial query of the device information is
/// critical, so we allow a longer timeout when getting this information
void CCameraProxy::_SetLongTimeouts(struct soap* s)
{
    s->connect_timeout = 6;
    s->send_timeout = 4;
    s->recv_timeout = 6;
    //s->keep_alive = 1;
}

void CCameraProxy::_SetNamePollTimer()
{
    size_t NumberOfNodes = m_NodeMap.size();
    if (NumberOfNodes > 0)
    {
        int Period = c_NamePollPeriodMs / NumberOfNodes;
        m_CameraNamePollTimer.start(Period > c_MinNamePollPeriodMs ? Period : c_MinNamePollPeriodMs);
    }
    else
    {
        m_CameraNamePollTimer.stop();
    }
}

static const char* H264ProfileToString(tt__H264Profile H264Profile)
{
    const char* Profile;
    switch (H264Profile)
    {
        case tt__H264Profile__Baseline:
            Profile = "baseline";
            break;
        case tt__H264Profile__Main:
            Profile = "main";
            break;
        case tt__H264Profile__Extended:
            Profile = "extended";
            break;
        case tt__H264Profile__High:
            Profile = "high";
            break;
        default:
            Profile = "unknown";
            break;
    }
    return Profile;
}

void CCameraProxy::_DumpEncoder(tt__VideoEncoderConfiguration* pVideoEncoder)
{
    const char* Encoding;
    switch (pVideoEncoder->Encoding)
    {
        case tt__VideoEncoding__JPEG:
            Encoding = "JPEG";
            break;
        case tt__VideoEncoding__MPEG4:
            Encoding = "MPEG4";
            break;
        case tt__VideoEncoding__H264:
            Encoding = "H264";
            break;
        default:
            Encoding = "Unknown";
            break;
    }

    RAYDEBUG("Token: %s, Name: %s, UseCount=%d, Encoding: %s, Resolution: %dx%d Quality:%.2f",
             pVideoEncoder->token.c_str(),
             pVideoEncoder->Name.c_str(),
             pVideoEncoder->UseCount,
             Encoding,
             pVideoEncoder->Resolution->Width,
             pVideoEncoder->Resolution->Height,
             pVideoEncoder->Quality );
    if (pVideoEncoder->RateControl)
    {
        RAYDEBUG("Rate Control: FrameRate=%d, EncodingInterval=%d, BitRate=%d",
                 pVideoEncoder->RateControl->FrameRateLimit,
                 pVideoEncoder->RateControl->EncodingInterval,
                 pVideoEncoder->RateControl->BitrateLimit);
    }

    if (pVideoEncoder->H264)
    {
        RAYDEBUG("H264 Profile: %s, GovLength=%d", H264ProfileToString(pVideoEncoder->H264->H264Profile), pVideoEncoder->H264->GovLength);
    }
    if (pVideoEncoder->Multicast)
    {
        RAYDEBUG("Multicast: AutoStart=%d", pVideoEncoder->Multicast->AutoStart);
    }
}

#ifdef DUMP_ENCODER_CONFIGS
void CCameraProxy::_DumpEncoderConfigOptions(tt__VideoEncoderConfigurationOptions* pEncoderConfigOptions)
{
    RAYDEBUG("Quality Range: %f-%f", pEncoderConfigOptions->QualityRange->Min, pEncoderConfigOptions->QualityRange->Max);
    if (pEncoderConfigOptions->H264)
    {
        RAYDEBUG("Supports H264");
        for (std::vector<tt__VideoResolution * >::const_iterator Iter = pEncoderConfigOptions->H264->ResolutionsAvailable.begin();
             Iter != pEncoderConfigOptions->H264->ResolutionsAvailable.end(); ++Iter)
        {
            RAYDEBUG("Supports Resolution %dx%d", (*Iter)->Width, (*Iter)->Height);
        }
        for (std::vector<tt__H264Profile>::const_iterator Iter = pEncoderConfigOptions->H264->H264ProfilesSupported.begin();
             Iter != pEncoderConfigOptions->H264->H264ProfilesSupported.end(); ++Iter)
        {
            RAYDEBUG("Supports Profile %s", H264ProfileToString(*Iter));
        }
        RAYDEBUG("GovLengthRange: %d-%d, FrameRateRange: %d-%d, EncodingIntervalRange: %d-%d",
                 pEncoderConfigOptions->H264->GovLengthRange->Min,
                 pEncoderConfigOptions->H264->GovLengthRange->Max,
                 pEncoderConfigOptions->H264->FrameRateRange->Min,
                 pEncoderConfigOptions->H264->FrameRateRange->Max,
                 pEncoderConfigOptions->H264->EncodingIntervalRange->Min,
                 pEncoderConfigOptions->H264->EncodingIntervalRange->Max);
        if (pEncoderConfigOptions->Extension && pEncoderConfigOptions->Extension->H264)
        {
            RAYDEBUG("BitrateRange: %d-%d", pEncoderConfigOptions->Extension->H264->BitrateRange->Min, pEncoderConfigOptions->Extension->H264->BitrateRange->Max);
        }
    }
}

void CCameraProxy::_DumpEncoders(MediaBindingProxy& MediaService)
{
    _trt__GetVideoEncoderConfigurations EncoderConfigurations;
    _trt__GetVideoEncoderConfigurationsResponse EncoderConfigurationsResponse;

    if (MediaService.GetVideoEncoderConfigurations(&EncoderConfigurations, EncoderConfigurationsResponse) == SOAP_OK)
    {
        RAYDEBUG("Got %zu video encoder configurations", EncoderConfigurationsResponse.Configurations.size());

        for (std::vector<tt__VideoEncoderConfiguration * >::const_iterator Iter = EncoderConfigurationsResponse.Configurations.begin();
             Iter != EncoderConfigurationsResponse.Configurations.end(); ++Iter)
        {
            _DumpEncoder(*Iter);
            _trt__GetVideoEncoderConfigurationOptions VideoEncoderConfigurationOptions;
            _trt__GetVideoEncoderConfigurationOptionsResponse VideoEncoderConfigurationOptionsResponse;
            VideoEncoderConfigurationOptions.ConfigurationToken = &((*Iter)->token);
            if (MediaService.GetVideoEncoderConfigurationOptions(&VideoEncoderConfigurationOptions, VideoEncoderConfigurationOptionsResponse) == SOAP_OK)
            {
                _DumpEncoderConfigOptions(VideoEncoderConfigurationOptionsResponse.Options);
            }
            else
            {
                RAYERROR("Failed to get video encoder configuration options for %s", (*Iter)->token.c_str());
            }
        }
    }
}
#endif

float CCameraProxy::_ScalePercentToRange(float Value, const CCameraProxy::FloatParameterRange_t& Range)
{
    float Result = Value;
    if (Range.Supported)
    {
        float Scale = Range.Max - Range.Min;
        if (Scale != 0.0f)
        {
            Result = (Value / 100.0f) * Scale + Range.Min;
        }
        RAYDETAIL("Scaling %0.1f%% to %0.1f", Value, Result);
    }

    RAYDETAIL("Scaling %0.1f%% to %0.1f", Value, Result);
    return Result;
}

float CCameraProxy::_ScaleRangeToPercent(float Value, const CCameraProxy::FloatParameterRange_t& Range)
{
    float Result = Value;
    if (Range.Supported)
    {
        float Scale = Range.Max - Range.Min;
        if (Scale != 0.0f)
        {
            Result = ((Value - Range.Min) / Scale) * 100.0f;
        }
        RAYDETAIL("Scaling %0.1f to %0.1f%%", Value, Result);
    }

    RAYDETAIL("Scaling %0.1f to %0.1f%%", Value, Result);
    return Result;

}


void CCameraProxy::_ApplySoapObjProps(soap* soapObj, std::string onvifNode)
{
    if (soapObj && !onvifNode.empty())
    {
        if (soapObj->fopen != NULL)
        {
            soapObj->fopen = &tcp_custom_connect;
        }
        else if (soapObj->fconnect != NULL)
        {
            soapObj->fconnect = &tcp_custom_connect;
        }
    }
}
