/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef CAMERAPROXY_H_INCLUDED
#define CAMERAPROXY_H_INCLUDED

#include <string>
#include <set>
#include <unordered_map>
#include <vector>
#include <QTimer>
#include <memory>
#include <QObject>
#include "FrameworkCommon/EventLogger.h"
#include "IpCameraStore.h"

class tt__VideoSource;
class tt__Profile;
class tt__VideoEncoderConfiguration;
class tt__VideoEncoderConfigurationOptions;
class DeviceBindingProxy;
class MediaBindingProxy;
class ImagingBindingProxy;
struct soap;
struct EncoderPreferences_t;

class CCameraProxy : public QObject
{
    Q_OBJECT

public:
    typedef std::set<std::string> NodeList_t;

    CCameraProxy();

    /// Get the brightness, colour and contrast for an onvif node
    /// \param[in] OnvifNode node to query
    /// \param[out] rBrightness brightness
    /// \param[out] rColour colour
    /// \param[out] rContrast contrast
    /// \return bool true for success
    bool GetBrightnessColourContrast(const std::string& OnvifNode, int& rBrightness, int& rColour, int& rContrast);

    /// Name getter
    bool GetName(const std::string& OnvifNode, std::string& rName);
    bool GetManufacturer(const std::string& OnvifNode, std::string& rManufacturer);
    bool GetModel(const std::string& OnvifNode, std::string& rModel);
    bool GetFirmwareVersion(const std::string& OnvifNode, std::string& rFirmwareVersion);
    bool GetSerialNumber(const std::string& OnvifNode, std::string& rSerialNumber);

    /// MAC Address getter
    bool GetMacAddress(const std::string& OnvifNode, std::string& rMacAddress);

    bool GetIPAddress(const std::string& OnvifNode, std::string& rIPAddress);
    /// Gets the basic details for an onif node
    /// \details Selects the RTSP URI with the H264 stream with the highest quality
    /// \param[in] OnvifNode node to query
    /// \param[in] AutoConfigure set to true of we are allowed to auto configure the camera
    /// \param[out] rHiResRtspUri RTSP Uri for the high res stream
    /// \param[out] rLoResRtspUri RTSP Uri for the low res stream

    /// \return bool true for success
    bool GetRtspUri(const std::string& OnvifNode, bool AutoConfigure, std::string& rHiResRtspUri, std::string& rLoResRtspUri);

    /// Add a node to list of nodes monitored by the proxy
    /// \param[in] OnvifNode node to add
    /// /todo Add security credentials
    bool AddNode(const std::string& OnvifNode);

    /// Sets the list of nodes.  We check for name updates for for all cameras we know about
    void SetNodeList(const NodeList_t& NodeList);

    /// Remove a node from list of nodes monitored by the proxy
    /// \param[in] OnvifNode node to add
    bool RemoveNode(const std::string& OnvifNode);

    /// Sets the three parameters on the given node
    bool SetBrightnessColourContrast(const std::string& OnvifNode, int Brightness, int Colour, int Contrast);

    /// Sets the name encoded as base16
    bool SetName(const std::string& OnvifNode, const std::string& NewName);

    /// Emitted when we've done an automatic query for a camera's name.  No change checking done here
    Q_SIGNAL void CameraNameUpdated(const std::string& OnvifNode, const std::string& Name);

    /// Name encoding function.
    /// \details we encode the input name to get around hostname restrictions splitting each
    /// 8-bit input character as two 4-bits and adding to 'a'.  This will give us a valid hostname
    /// \param[in] PlainName input plaintext
    /// \param[out] rEncodedName output ciphertext
    /// \return true if success
    static bool EncodeName(const std::string& PlainName, std::string& rEncodedName);

    /// Name decoding function.
    /// \details we decode the input name to reverse process in EncodeName.  if we detect characters
    /// outside a-p we conclude it's not one of our's and fill rPlainName with the input encoded name
    /// \param[in] EncodedName input ciphertext
    /// \param[out] rPlainName output plaintext
    /// \return true if success
    static bool DecodeName(const std::string& EncodedName, std::string& rPlainName);
private:

    struct FloatParameterRange_t
    {
        bool Supported = false;
        float Min = 0.0f;
        float Max = 0.0f;
    };

    struct OnvifNode_t
    {
        std::shared_ptr<DeviceBindingProxy>     m_pDeviceService;
        std::shared_ptr<MediaBindingProxy>      m_pMediaService;
        std::shared_ptr<ImagingBindingProxy>    m_pImagingService;
        std::string                             m_Manufacturer;
        std::string                             m_Model;
        std::string                             m_FirmwareVersion;
        std::string                             m_SerialNumber;
        std::string                             m_VideoSourceToken;
        std::string                             m_HiResProfileToken;
        std::string                             m_HiResVideoEncoderToken;
        std::string                             m_LoResProfileToken;
        std::string                             m_LoResVideoEncoderToken;
        size_t                                  m_MaxProfiles = 0;
        size_t                                  m_MaxH264Instances = 0;
        FloatParameterRange_t                   m_BrightnessRange;
        FloatParameterRange_t                   m_ColourRange;
        FloatParameterRange_t                   m_ContrastRange;
    };

    Q_SLOT void _PollCameraName();
    std::string _GetStreamUri(const std::string& OnvifNodeUrl, MediaBindingProxy& mediaService, const std::string& ProfileToken);

    bool _CheckForExistingProfile(const tt__Profile* Profile, const std::string& Name, std::string& rProfileToken, std::string& rEncoderToken);
    bool _ConfigureProfile(const std::string& OnvifNodeUrl, OnvifNode_t& Node, const EncoderPreferences_t& EncoderPreferences, std::vector<class tt__Profile*>& Profiles, std::string& rProfileToken, std::string& rEncoderToken);
    bool _CreateNewProfile(const std::string& OnvifNodeUrl, OnvifNode_t& Node, const EncoderPreferences_t& EncoderPreferences, std::string& rProfileToken, std::string& rEncoderToken);
    std::string _ConfigureEncoderForProfile(const std::string& OnvifNodeUrl, OnvifNode_t& Node, const std::string& ProfileToken, const EncoderPreferences_t& EncoderPreferences);
    std::string _FindBestH264Profile(std::vector<class tt__Profile * > &Profiles, int MinHeight, int MaxHeight, std::string& rEncoderToken);

    void _SetTimeouts(struct soap* s);
    void _SetLongTimeouts(struct soap* s);
    void _SetNamePollTimer();
    void _DumpEncoder(tt__VideoEncoderConfiguration* pVideoEncoder);
    void _DumpEncoderConfigOptions(tt__VideoEncoderConfigurationOptions* pEncoderConfigOptions);
    void _DumpEncoders(MediaBindingProxy&);
    float _ScalePercentToRange(float Value, const FloatParameterRange_t& Range);
    float _ScaleRangeToPercent(float Value, const FloatParameterRange_t& Range);
    void _ApplySoapObjProps(soap* soapObj, std::string onvifNode);


    typedef std::unordered_map<std::string, OnvifNode_t>  NodeMap_t;
    NodeMap_t m_NodeMap;  ///< list of nodes we query for names triggered by poll timer
    NodeMap_t::const_iterator m_NamePollListIter;  // Iterator into the Name Poll List
    QTimer m_CameraNamePollTimer;       ///< Poll timer

    bool m_ForceReconfigure;                       ///< Debug flag to enable forced reconfiguration

};

#endif // CAMERAPROXY_H_INCLUDED
