/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "DiscoveryService.h"
#include "IpCameraStore.h"
#include <cassert>
#include "wsddapi.h"
#include "wsaapi.h"
#include <cstdio>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/if.h>
#include <net/if_arp.h>
#include <unistd.h>
#include <regex>
#include <linux/sockios.h>
#include <linux/ethtool.h>
#include "FrameworkCommon/EventLogger.h"
#include "FrameworkCommon/IpInfo.h"
#ifdef TARGET_ATOM
    #include "FrameworkCommon/OMAPCrash.h"
#endif

const int c_ListenTimeoutS = 5; // seconds
const char* c_IpAddressRegex = "\\b(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\.(25[0-5]|2[0-4][0-9]|[01]?[0-9][0-9]?)\\b";

 std::string CDiscoveryService::m_LocalIpAddress = "";
 std::string CDiscoveryService::m_LocalIpAddressLegacy = "";

static SOAP_SOCKET (*orig_tcp_connect)(struct soap*, const char*, const char*, int);
static SOAP_SOCKET (*orig_tcp_connect_legacy)(struct soap*, const char*, const char*, int);

/// Global callback function when we receive a Hello WS-Discovery message
/// not used
void wsdd_event_Hello(struct soap *soap,
                      unsigned int InstanceId,
                      const char *SequenceId,
                      unsigned int MessageNumber,
                      const char *MessageID,
                      const char *RelatesTo,
                      const char *EndpointReference,
                      const char *Types,
                      const char *Scopes,
                      const char *MatchBy,
                      const char *XAddrs,
                      unsigned int MetadataVersion)
{
    RAYTRACE("");
}


/// Global callback function when we receive a Bye WS-Discovery message
/// not used
void wsdd_event_Bye(struct soap *soap,
                    unsigned int InstanceId,
                    const char *SequenceId,
                    unsigned int MessageNumber,
                    const char *MessageID,
                    const char *RelatesTo,
                    const char *EndpointReference,
                    const char *Types,
                    const char *Scopes,
                    const char *MatchBy,
                    const char *XAddrs,
                    unsigned int *MetadataVersion)
{
    RAYTRACE("");
}

/// Global callback function when we receive a Probe WS-Discovery message
/// not used
soap_wsdd_mode wsdd_event_Probe(struct soap *soap,
                                const char *MessageID,
                                const char *ReplyTo,
                                const char *Types,
                                const char *Scopes,
                                const char *MatchBy,
                                struct wsdd__ProbeMatchesType *ProbeMatches)
{
    RAYTRACE("");
    return SOAP_WSDD_ADHOC;
}

/// Global callback function when we receive a ProbeMatches WS-Discovery message
/// We get this when we send a probe message to the multicast
/// We get the onvif node from here and call GotProbeMatch in the discovery service
void wsdd_event_ProbeMatches(struct soap* pSoap,
                             unsigned int /* InstanceId */,
                             const char* /* SequenceId */,
                             unsigned int /* MessageNumber */,
                             const char* /* MessageID */,
                             const char* /* RelatesTo */,
                             struct wsdd__ProbeMatchesType *ProbeMatches)
{
    int NumMatches = ProbeMatches->__sizeProbeMatch;
    CDiscoveryService* pDiscoveryService = reinterpret_cast<CDiscoveryService*>(pSoap->user);

    for (int idx = 0; idx < NumMatches; ++idx)
    {
        wsdd__ProbeMatchType* pProbeMatch = &(ProbeMatches->ProbeMatch[idx]);
        if (pProbeMatch != 0)
        {
    #if 0
            if (pProbeMatch->Types)
            {
                RAYDETAIL("Types=%s", pProbeMatch->Types);
            }
            if (pProbeMatch->XAddrs)
            {
                RAYDETAIL("XAddrs=%s", pProbeMatch->XAddrs);
            }
            if (pProbeMatch->Scopes)
            {
                RAYDETAIL("Scopes->__item=%s, Scopes->MatchBy=%s", pProbeMatch->Scopes->__item, pProbeMatch->Scopes->MatchBy);
            }
    #endif
            wsdd__UriListType XAddrs = pProbeMatch->XAddrs;
            int AddrIdx = 0;
            char* pch = strtok(XAddrs, " ");
            while (pch != 0)
            {
                std::regex ipRegex(c_IpAddressRegex);
                std::smatch match;
                std::string ipStr;
                std::string _pch = std::string(pch);

                if (std::regex_search(_pch, match, ipRegex))
                {
                    ipStr = match[0].str();
                    pDiscoveryService->GotProbeMatch(std::string(pch));
                }

                AddrIdx++;
                pch = strtok(0, " ");
            }
        }
    }
}

/// Global callback function when we receive a Resolve WS-Discovery message
/// not used
soap_wsdd_mode wsdd_event_Resolve(struct soap *soap,
                                  const char *MessageID,
                                  const char *ReplyTo,
                                  const char *EndpointReference,
                                  struct wsdd__ResolveMatchType *match)
{
    RAYTRACE("");
    return SOAP_WSDD_ADHOC;
}


/// Global callback function when we receive a Resolve Matches WS-Discovery message
/// not used
void wsdd_event_ResolveMatches(struct soap *soap,
                               unsigned int InstanceId,
                               const char *SequenceId,
                               unsigned int MessageNumber,
                               const char *MessageID,
                               const char *RelatesTo,
                               struct wsdd__ResolveMatchType *match)
{
    RAYTRACE("");
}

/// Our own version of tcp_connect so we can bind to the right multicast port and IP address
SOAP_SOCKET wsdd_tcp_connect(struct soap* s, const char *endpoint, const char *host, int port)
{
  // real tcp_connect
  struct in_addr inaddr;

  RAYTRACE("");

  inaddr.s_addr = inet_addr(CDiscoveryService::m_LocalIpAddress.c_str());
  s->socket = (orig_tcp_connect)(s, endpoint, host, port);
  setsockopt (s->socket, IPPROTO_IP, IP_MULTICAST_IF, &inaddr, sizeof(inaddr));

  return s->socket;
}

SOAP_SOCKET wsdd_tcp_connect_legacy(struct soap* s, const char *endpoint, const char *host, int port)
{
  // real tcp_connect
  struct in_addr inaddr;

  inaddr.s_addr = inet_addr(CDiscoveryService::m_LocalIpAddressLegacy.c_str());
  s->socket = (orig_tcp_connect_legacy)(s, endpoint, host, port);
  setsockopt (s->socket, IPPROTO_IP, IP_MULTICAST_IF, &inaddr, sizeof(inaddr));

  int flag = CIpInfo::c_markValue;
  if (-1 == setsockopt(s->socket, SOL_SOCKET, SO_MARK, &flag, sizeof(flag))) {
      RAYERROR("setsockopt fail SO_MARK");
  }

  return s->socket;
}

static const struct Namespace wsdd_namespaces[] =
{
    {"SOAP-ENV", "http://www.w3.org/2003/05/soap-envelope", "http://www.w3.org/*/soap-envelope", NULL},
    {"SOAP-ENC", "http://www.w3.org/2003/05/soap-encoding", "http://www.w3.org/*/soap-encoding", NULL},
    {"xsi", "http://www.w3.org/2001/XMLSchema-instance", "http://www.w3.org/*/XMLSchema-instance", NULL},
    {"xsd", "http://www.w3.org/2001/XMLSchema", "http://www.w3.org/*/XMLSchema", NULL},
    {"wsa", "http://schemas.xmlsoap.org/ws/2004/08/addressing", NULL, NULL},
    {"wsdd", "http://schemas.xmlsoap.org/ws/2005/04/discovery", NULL, NULL},
    {"dp0", "http://www.onvif.org/ver10/network/wsdl", NULL, NULL},
    {NULL, NULL, NULL, NULL}
};

CDiscoveryService::CDiscoveryService()
:
m_pSoap(0),
m_pSoapLegacy(0),
m_KeepRunning(true),
m_pIpInfo(new CIpInfo())
{
    if (m_pIpInfo != nullptr)
    {
        CDiscoveryService::m_LocalIpAddress = m_pIpInfo->GetFormattedIPAddress(RaySocketTypes::e_DevWired);
        CDiscoveryService::m_LocalIpAddressLegacy = m_pIpInfo->GetFormattedIPAddress(RaySocketTypes::e_DevWiredLegacy);
    }

    setObjectName("CDiscoveryService");

    _Init(true);
    _Init(false);

    start();
}

CDiscoveryService::~CDiscoveryService()
{
    m_KeepRunning = false;
    close(m_pSoap->socket);
    wait(c_ListenTimeoutS * 1000);
    soap_free(m_pSoap);

    close(m_pSoapLegacy->socket);
    wait(c_ListenTimeoutS * 1000);
    soap_free(m_pSoapLegacy);

    if (m_pIpInfo != nullptr)
    {
        delete m_pIpInfo;
        m_pIpInfo = nullptr;
    }
}

void CDiscoveryService::_DoProbe(bool newSubnet)
{
    // Send a probe request
    struct soap* soapObj = m_pSoap;
    if (!newSubnet)
    {
        soapObj = m_pSoapLegacy;
    }

    const char* id = soap_wsa_rand_uuid(soapObj);
    int rc = soap_wsdd_Probe(	soapObj,
                            SOAP_WSDD_ADHOC,      // ad-hoc mode
                            SOAP_WSDD_TO_TS,      // to a TS
                            "soap.udp://239.255.255.250:3702",    // endpoint address
                            id,                   // message ID
                            "http://schemas.xmlsoap.org/ws/2004/08/addressing/role/anonymous",                 // ReplyTo
                            "dp0:NetworkVideoTransmitter",	// Types
                            "onvif://www.onvif.org",	// Scopes
                            NULL); // MatchBy

    if (rc != SOAP_OK)
    {
        RAYERROR("Failed to send probe:%d\n", rc);
        //assert(rc == 0);
    }
}

void CDiscoveryService::GotProbeMatch(const std::string& OnvifNode)
{
    // Called back from probe match function
    RAYTRACE("OnvifNode=%s", OnvifNode.c_str());
    std::pair<std::set<std::string>::iterator, bool> Pair = m_Nodes.insert(OnvifNode);
    if (Pair.second)
    {
        RAYINFO("New OnvifNode=%s", OnvifNode.c_str());
        Q_EMIT ProbeMatch(OnvifNode);
    }
}

void CDiscoveryService::run()
{
#ifdef TARGET_ATOM
    eCrash_RegisterThread("CDiscoveryService", 0);
#endif

    //check if ip address was populated
    bool ipAddressReady = false;
    if (CDiscoveryService::m_LocalIpAddress.size() > 0 && CDiscoveryService::m_LocalIpAddressLegacy.size() > 0)
    {
        ipAddressReady = true;
        _DoProbe(true);
        _DoProbe(false);
    }

    while (m_KeepRunning)
    {
        if (!ipAddressReady)
        {
            sleep(5);

            if (m_pIpInfo == nullptr)
            {
                m_pIpInfo = new CIpInfo();
            }

            if (m_pIpInfo != nullptr)
            {
#if !defined(PLATFORM_WORKSTATION)
                //try to get again the IP addresses from eth0
                m_pIpInfo->UpdateInterfaceInfo(RaySocketTypes::EthernetDevice_t::e_DevWired);
                m_pIpInfo->UpdateInterfaceInfo(RaySocketTypes::EthernetDevice_t::e_DevWiredLegacy);
#endif

                CDiscoveryService::m_LocalIpAddress = m_pIpInfo->GetFormattedIPAddress(RaySocketTypes::e_DevWired);
                CDiscoveryService::m_LocalIpAddressLegacy = m_pIpInfo->GetFormattedIPAddress(RaySocketTypes::e_DevWiredLegacy);
            }

            if (CDiscoveryService::m_LocalIpAddress.size() > 0 && CDiscoveryService::m_LocalIpAddressLegacy.size() > 0)
            {
                _Init(true);
                _Init(false);

                ipAddressReady = true;

                _DoProbe(true);
                _DoProbe(false);
            }
        }
        else
        {
            _runImpl(true);
            _runImpl(false);
        }
    }
#ifdef TARGET_ATOM
    eCrash_UnregisterThread();
#endif
}

void CDiscoveryService::_runImpl(bool newSubnet)
{
    if (newSubnet)
    {
        // Do a soap listen
        soap_begin(m_pSoap);
        int rc = soap_wsdd_listen(m_pSoap, c_ListenTimeoutS); // listen for messages
        if (rc == SOAP_NO_TAG)
        {
            if (m_pSoap)
            {
                RAYERROR("m_pSoap->buf=%s", m_pSoap->buf);
            }
        }

        soap_destroy(m_pSoap);
        soap_end(m_pSoap);
        if (rc != SOAP_OK)
        {
            if (rc != SOAP_EOF)
            {
                RAYERROR("rc=%d", rc);
                soap_free(m_pSoap);
                m_pSoap = nullptr;
                _Init(newSubnet);
            }
            m_Nodes.clear();
            _DoProbe(newSubnet);
        }
    }
    else
    {
        // Do a soap listen
        soap_begin(m_pSoapLegacy);
        int rc = soap_wsdd_listen(m_pSoapLegacy, c_ListenTimeoutS); // listen for messages
        if (rc == SOAP_NO_TAG)
        {
            if (m_pSoapLegacy)
            {
                RAYERROR("m_pSoap->buf=%s", m_pSoapLegacy->buf);
            }
        }

        soap_destroy(m_pSoapLegacy);
        soap_end(m_pSoapLegacy);
        if (rc != SOAP_OK)
        {
            if (rc != SOAP_EOF)
            {
                RAYERROR("rc=%d", rc);
                soap_free(m_pSoapLegacy);
                m_pSoapLegacy = nullptr;
                _Init(newSubnet);
            }
            m_Nodes.clear();
            _DoProbe(newSubnet);
        }
    }
}

void CDiscoveryService::_Init(bool newSubnet)
{
    int port = 5678;

    if (newSubnet)
    {
        if (m_pSoap != nullptr)
        {
            if (m_pSoap->socket != SOAP_INVALID_SOCKET)
            {
                close(m_pSoap->socket);
                m_pSoap->socket = SOAP_INVALID_SOCKET;
            }
            soap_free(m_pSoap);
            m_pSoap = nullptr;
        }
        m_pSoap = soap_new();
        soap_init1(m_pSoap, SOAP_IO_UDP | SO_BROADCAST);
        soap_set_namespaces(m_pSoap, wsdd_namespaces);
        soap_bind(m_pSoap, NULL, port, 100);
        m_pSoap->version = 1;

        if (CDiscoveryService::m_LocalIpAddress.size() > 0)
        {
            RAYDETAIL("Usig IP %s", m_LocalIpAddress.c_str());
            orig_tcp_connect = m_pSoap->fopen;
            m_pSoap->fopen = &wsdd_tcp_connect;

            ip_mreq multicastRequest;
            memset(&multicastRequest, 0, sizeof(multicastRequest));
            multicastRequest.imr_multiaddr.s_addr = inet_addr("239.255.255.250");
            multicastRequest.imr_interface.s_addr = inet_addr(CDiscoveryService::m_LocalIpAddress.c_str());
            setsockopt(m_pSoap->socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (const char*)&multicastRequest, sizeof(multicastRequest));

            int flag = 1;
            if (-1 == setsockopt(m_pSoap->socket, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag))) {
                RAYERROR("setsockopt fail SO_REUSEADDR");
            }

            flag = 1;

            if (-1 == setsockopt(m_pSoap->socket, SOL_SOCKET, SO_REUSEPORT, &flag, sizeof(flag))) {
                RAYERROR("setsockopt fail SO_REUSEPORT");
            }
        }
        m_pSoap->user = reinterpret_cast<void*>(this);
    }
    else
    {
        if (m_pSoapLegacy != nullptr)
        {
            if (m_pSoapLegacy->socket != SOAP_INVALID_SOCKET)
            {
                close(m_pSoapLegacy->socket);
                m_pSoapLegacy->socket = SOAP_INVALID_SOCKET;
            }
            soap_free(m_pSoapLegacy);
            m_pSoapLegacy = nullptr;
        }

        RAYDEBUG("Init legacy soap");
        m_pSoapLegacy = soap_new();
        soap_init1(m_pSoapLegacy, SOAP_IO_UDP | SO_BROADCAST);
        soap_set_namespaces(m_pSoapLegacy, wsdd_namespaces);
        soap_bind(m_pSoapLegacy, NULL, port, 100);
        m_pSoapLegacy->version = 1;

        if (CDiscoveryService::m_LocalIpAddressLegacy.size() > 0)
        {
            RAYDETAIL("Usig IP %s", m_LocalIpAddressLegacy.c_str());
            orig_tcp_connect_legacy = m_pSoapLegacy->fopen;
            m_pSoapLegacy->fopen = &wsdd_tcp_connect_legacy;

            ip_mreq multicastRequest;
            memset(&multicastRequest, 0, sizeof(multicastRequest));
            multicastRequest.imr_multiaddr.s_addr = inet_addr("239.255.255.250");
            multicastRequest.imr_interface.s_addr = inet_addr(CDiscoveryService::m_LocalIpAddressLegacy.c_str());
            setsockopt(m_pSoapLegacy->socket, IPPROTO_IP, IP_ADD_MEMBERSHIP, (const char*)&multicastRequest, sizeof(multicastRequest));

            int flag = 1;
            if (-1 == setsockopt(m_pSoapLegacy->socket, SOL_SOCKET, SO_REUSEADDR, &flag, sizeof(flag))) {
                RAYERROR("setsockopt fail SO_REUSEADDR");
            }

            flag = 1;

            if (-1 == setsockopt(m_pSoapLegacy->socket, SOL_SOCKET, SO_REUSEPORT, &flag, sizeof(flag))) {
                RAYERROR("setsockopt fail SO_REUSEPORT");
            }


            flag = CIpInfo::c_markValue;

            if (-1 == setsockopt(m_pSoapLegacy->socket, SOL_SOCKET, SO_MARK, &flag, sizeof(flag))) {
                RAYERROR("setsockopt fail SO_MARK");
            }
        }
        m_pSoapLegacy->user = reinterpret_cast<void*>(this);
    }
}
