/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "IPVideoControl.h"
#include "NetworkHelpers.h"
#include "FrameworkCommon/IProcessWatchdog.h"
#include <QFile>
#ifdef __ANDROID__
#include <QtAndroid>
#include <QAndroidJniObject>
#include <QAndroidJniEnvironment>
#endif
static const char* cServerNameObjectPath = "/mnt/tmp/Raymarine.OnvifDaemon";
static const int c_SetTimerPeriodMs = 500;
static const int c_CheckConncetionTimerPeriodMs = 5000;

static const unsigned int c_PollingCurrentCameraInterval = 5000 / c_SetTimerPeriodMs;

CIPVideoControl::CIPVideoControl(IProcessWatchdog* pWatchdog, QObject* pParent):
    QObject(pParent),
    m_pLocalServer(new QLocalServer(this)),
    m_pConnection(nullptr),
    m_DiscoveryService(),
    m_CheckConnectionTimer(),
    m_TimerCounter(0u),
    m_AutoConfigEnabled(false),
    m_pProcessWatchdog(pWatchdog)
{
    RAYINFO("");

    connect(m_pLocalServer, &QLocalServer::newConnection, this, &CIPVideoControl::_OnNewConnection);

    m_pLocalServer->setMaxPendingConnections(1);

    _setupSocketServer();

    m_Timer.setSingleShot(false);
    m_Timer.setInterval(c_SetTimerPeriodMs);
    m_CheckConnectionTimer.setSingleShot(false);
    m_CheckConnectionTimer.setInterval(c_CheckConncetionTimerPeriodMs);

    // Make Qt connections.
    bool connected = false;

    connected = connect(&m_IpCameraStore, &CIpCameraStore::CameraFound,
                        this,             &CIPVideoControl::_OnCameraFound);
    RAYASSERT(connected);
    connected = connect(&m_IpCameraStore, &CIpCameraStore::CameraLost,
                        this,             &CIPVideoControl::_OnCameraLost);
    RAYASSERT(connected);
    connected = connect(&m_IpCameraStore, &CIpCameraStore::CameraFound,
                        this,             &CIPVideoControl::_OnCameraMapChanged, Qt::QueuedConnection);
    RAYASSERT(connected);
    connected = connect(&m_IpCameraStore, &CIpCameraStore::CameraLost,
                        this,             &CIPVideoControl::_OnCameraMapChanged, Qt::QueuedConnection);
    RAYASSERT(connected);
    connected = connect(&m_IpCameraStore, &CIpCameraStore::ParameterChanged,
                         this,            &CIPVideoControl::_OnParameterChanged);
    RAYASSERT(connected);
    connected = connect(&m_DiscoveryService, &CDiscoveryService::ProbeMatch,
                        this,                &CIPVideoControl::_OnProbeMatch, Qt::DirectConnection);
    RAYASSERT(connected);
    connected = connect(&m_CameraProxy,   &CCameraProxy::CameraNameUpdated,
                        &m_IpCameraStore, &CIpCameraStore::OnUpdatedCameraName);
    RAYASSERT(connected);
    connected = connect(&m_Timer, &QTimer::timeout,
                        this,     &CIPVideoControl::_OnCheckForAsyncComms);
    RAYASSERT(connected);
    if (connected)
    {
        m_Timer.start();
    }

    connected = connect(&m_CheckConnectionTimer, &QTimer::timeout,
                        this,     &CIPVideoControl::_OnCheckConnection);
    RAYASSERT(connected);
    if (connected)
    {
        m_CheckConnectionTimer.start();
    }


    RAYINFO("OnvifDaemon started");
}

void CIPVideoControl::_setupSocketServer()
{
    QLocalServer::removeServer(cServerNameObjectPath);
    if (!m_pLocalServer->listen(cServerNameObjectPath))
    {

        RAYWARNING("Error: %d  %s", m_pLocalServer->serverError(), m_pLocalServer->errorString().toStdString().c_str());
        RAYWARNING("Can't start listening on %s", cServerNameObjectPath);
        RAYWARNING("Trying to remove %s other running onvifDaemons might stop working!", cServerNameObjectPath);
        if (QLocalServer::removeServer(cServerNameObjectPath))
        {
            RAYWARNING("%s removed. Trying to listen once again", cServerNameObjectPath);
            if (!m_pLocalServer->listen(cServerNameObjectPath))
            {
                RAYWARNING("Still can't start listening on %s", cServerNameObjectPath);
                RAYWARNING(" Error: %d  %s", m_pLocalServer->serverError(), m_pLocalServer->errorString().toStdString().c_str());

                RAYFATAL("Unable to start OnvifDaemon");
            }
        }
        else
        {
            RAYWARNING("Removing %s failed", cServerNameObjectPath);
            RAYFATAL("Unable to start OnvifDaemon");
        }
    }

    RAYINFO("LocalServer on %s : Exists? %d ", m_pLocalServer->fullServerName().toStdString().c_str(), QFile::exists(cServerNameObjectPath));

}

void CIPVideoControl::_OnNewConnection()
{
    RAYINFO("");

    if (m_pLocalServer->hasPendingConnections())
    {
        if (!m_pConnection)
        {
            m_pConnection = m_pLocalServer->nextPendingConnection();

            connect(m_pConnection, &QLocalSocket::readyRead, this , &CIPVideoControl::_OnClientDataReady);
            connect(m_pConnection, &QLocalSocket::disconnected, this , &CIPVideoControl::_OnClientDisconnect);

            m_pConnection->setReadBufferSize(4);

            RAYINFO("Connection established");
        }
        else
        {
            RAYWARNING("Something tried to connect while the connection is already established");
            QLocalSocket* unneeded = m_pLocalServer->nextPendingConnection();
            _DiconnectSocket(unneeded);
        }
    }
}

void CIPVideoControl::_OnClientDataReady()
{
    m_CommunicationBuffer.append(m_pConnection->readAll());
    readPackedFromBuffer(m_CommunicationBuffer,
        [this](qint8 header, QDataStream& data) { _InterpretPacket(header,data); }
    );
}

void CIPVideoControl::_OnClientDisconnect()
{
    RAYINFO("");
    _DiconnectClient();
}

void CIPVideoControl::_DiconnectSocket(QLocalSocket*& rSocket)
{
    if (rSocket)
    {
        rSocket->close();
        rSocket->deleteLater();
        rSocket = nullptr;
    }
}

void CIPVideoControl::_DiconnectClient()
{
    _DiconnectSocket(m_pConnection);
    m_CommunicationBuffer.clear();
}

void CIPVideoControl::_InterpretPacket(qint8 header, QDataStream& data)
{
    switch (header)
    {
        case ClientRequest_SendAllCameras:
        {
            RAYINFO("Received SendAllCameras request");
            SendAllCameras();
        }
        break;

        case ClientRequest_SetContrast:
        {
            RAYINFO("Received SetContrast request");
            QString macAddress;
            quint8 contrast;
            data >> macAddress >> contrast;
            SetContrast(macAddress, contrast);
        }
        break;

        case ClientRequest_SetBrightness:
        {
            RAYINFO("Received SetBrightness request");
            QString macAddress;
            quint8 brightness;
            data >> macAddress >> brightness;
            SetBrightness(macAddress, brightness);
        }
        break;

        case ClientRequest_SetColour:
        {
            RAYINFO("Received SetColour request");
            QString macAddress;
            quint8 colour;
            data >> macAddress >> colour;
            SetColour(macAddress, colour);
        }
        break;

        case ClientRequest_SetName:
        {
           RAYINFO("Received SetName request");
           QString macAddress;
           QString name;
           data >> macAddress >> name;
           SetName(macAddress, name);
        }
        break;

        case ClientRequest_EnableCameraAutoConfiguration:
        {
            RAYINFO("Received EnableCameraAutoConfiguration request");
            bool enableCameraAutoConfiguration;
            data >> enableCameraAutoConfiguration;
            EnableCameraAutoConfiguration(enableCameraAutoConfiguration);
        }
        break;

        default:
            RAYWARNING("Incorrect header, packet discarded, diconnecting client");
            _DiconnectClient();
    }
}

bool CIPVideoControl::_IsConnectionOpen()
{
    return m_pConnection && m_pConnection->state() == QLocalSocket::ConnectedState;
}

void CIPVideoControl::_OnCameraFound(QVariantMap cameraMap)
{
    if (_IsConnectionOpen())
    {
        RAYINFO("Sending AddCamera response");
        QByteArray packetByteArray = createPacket(ServerResponse_CameraFound, cameraMap);
        int bytesSent = m_pConnection->write(packetByteArray);
        RAYINFO("AddCamera response sent %i/%i", bytesSent, packetByteArray.size());
        RAYASSERT(bytesSent == packetByteArray.size());
    }
}

void CIPVideoControl::_OnCameraLost(QString macAddress)
{
    if (_IsConnectionOpen())
    {
        RAYINFO("Sending DeleteCamera response");
        QByteArray packetByteArray = createPacket(ServerResponse_CameraLost, macAddress);
        int bytesSent = m_pConnection->write(packetByteArray);
        RAYINFO("DeleteCamera response sent %i/%i", bytesSent, packetByteArray.size());
        RAYASSERT(bytesSent == packetByteArray.size());
    }
}

void CIPVideoControl::_OnParameterChanged(QString macAddress, QString parameterName, QVariant value)
{
    if (_IsConnectionOpen())
    {
        RAYINFO("Sending ParameterChanged [%s] response", parameterName.toStdString().c_str());
        QByteArray packetByteArray = createPacket(ServerResponse_ParameterChanged, macAddress, parameterName, value);
        int bytesSent = m_pConnection->write(packetByteArray);
        RAYINFO("ParameterChanged response sent %i/%i", bytesSent, packetByteArray.size());
        RAYASSERT(bytesSent == packetByteArray.size());
    }
}

void CIPVideoControl::_OnCameraMapChanged()
{
    CCameraProxy::NodeList_t NamePollList;
    m_IpCameraStore.GetOnvifNodeList(NamePollList);
    m_CameraProxy.SetNodeList(NamePollList);
    if (NamePollList.find(m_CurrentOnvifNode) == NamePollList.end())
    {
        RAYTRACE("");
        m_CurrentOnvifNode.clear();
    }
}

void CIPVideoControl::EnableCameraAutoConfiguration(bool enable)
{
    RAYTRACE("Enable=%d", enable);
    m_AutoConfigEnabled = enable;
}

void CIPVideoControl::SendAllCameras()
{
    RAYTRACE("");
    m_IpCameraStore.SendAllCameras();
}

void CIPVideoControl::SetActiveCamera(QString macAddress)
{
    RAYTRACE("");
    m_CurrentOnvifNode = m_IpCameraStore.GetOnvifNode(macAddress);
}

void CIPVideoControl::SetContrast(QString macAddress, quint8 contrast)
{
    RAYTRACE("");
    std::string OnvifNode = m_IpCameraStore.GetOnvifNode(macAddress);
    RAYDETAIL("%s Value:%d\n", OnvifNode.c_str(), contrast);
    if (OnvifNode.size() > 0)
    {
        SetPacket setPacket;
        setPacket.Brightness = m_IpCameraStore.GetBrightness(macAddress);
        setPacket.Colour = m_IpCameraStore.GetColour(macAddress);
        setPacket.Contrast = contrast;
        _AddSetPacket(OnvifNode, setPacket);
    }
}

void CIPVideoControl::SetBrightness(QString macAddress, quint8 brightness)
{
    RAYTRACE("");
    std::string OnvifNode = m_IpCameraStore.GetOnvifNode(macAddress);
    RAYDETAIL("%s Value:%d\n", OnvifNode.c_str(), brightness);
    if (OnvifNode.size() > 0)
    {
        SetPacket setPacket;
        setPacket.Brightness = brightness;
        setPacket.Colour = m_IpCameraStore.GetColour(macAddress);
        setPacket.Contrast = m_IpCameraStore.GetContrast(macAddress);
        _AddSetPacket(OnvifNode, setPacket);
    }
}

void CIPVideoControl::SetColour(QString macAddress, quint8 colour)
{
    RAYTRACE("");
    std::string OnvifNode = m_IpCameraStore.GetOnvifNode(macAddress);
    RAYDETAIL("%s Value:%d\n", OnvifNode.c_str(), colour);
    if (!OnvifNode.empty())
    {
        SetPacket setPacket;
        setPacket.Brightness = m_IpCameraStore.GetBrightness(macAddress);
        setPacket.Colour = colour;
        setPacket.Contrast = m_IpCameraStore.GetContrast(macAddress);
        _AddSetPacket(OnvifNode, setPacket);
    }
}

void CIPVideoControl::SetName(QString macAddress, QString name)
{
    RAYINFO("Mac:%s Name:%s\n", macAddress.toStdString().c_str(), name.toStdString().c_str());

    std::string OnvifNode = m_IpCameraStore.GetOnvifNode(macAddress);
    if (!OnvifNode.empty())
    {
        std::string LocalName = name.toStdString();
        RAYDETAIL("CIPVideoControl::SetName: OnvifNode:%s Name:%s\n", OnvifNode.c_str(), LocalName.c_str());
        if (m_CameraProxy.SetName(OnvifNode, LocalName))
        {
            std::string NewName;

            if (m_CameraProxy.GetName(OnvifNode, NewName))
            {
                RAYDETAIL("CIPVideoControl::SetName: OnvifNode:%s New Name:%s\n", OnvifNode.c_str(), NewName.c_str());
                m_IpCameraStore.OnUpdatedCameraName(OnvifNode, NewName);
            }
        }
    }
}

void CIPVideoControl::_OnProbeMatch(const std::string& onvifNode)
{
    RAYINFO("OnvifNode=%s", onvifNode.c_str());
    m_MatchedNodes.Insert(onvifNode);
}

// Adds a new packet of set data into the map. Will replace any previous set for this onvif node
void CIPVideoControl::_AddSetPacket(const std::string& onvifNode, const SetPacket& setPacket)
{
    RAYTRACE("");
    m_SetPacketMap.AddPacket(onvifNode, setPacket);
    m_IpCameraStore.OnUpdatedBrightnessColourContrast(onvifNode, setPacket.Brightness, setPacket.Colour, setPacket.Contrast);
}

void CIPVideoControl::_ProcessMatchedNodes()
{
    RAYTRACE("");

    for (const std::string& onvifNode : m_MatchedNodes.PurgeSet())
    {
        if (m_IpCameraStore.IsNewCamera(onvifNode))
        {
            m_NewNodes.Insert(onvifNode);
        }
        else
        {
            m_IpCameraStore.OnSeenCamera(onvifNode);
            QString macAddress = m_IpCameraStore.GetMacAddress(onvifNode);
            if (m_IpCameraStore.GetLoResRtspUri(macAddress).isEmpty())
            {
                // If we don't have a Lo Res stream then try getting again (someone else may have autoconfigured it)
                m_NewNodes.Insert(onvifNode);
            }
        }
    }
}

void CIPVideoControl::_SendSetPackets()
{
    RAYTRACE("");

    for (const auto& packetPair : m_SetPacketMap.PurgeMap())
    {
        RAYTRACE("");

        const std::string& onvifNode = packetPair.first;
        const SetPacket& setPacket = packetPair.second;

        if (onvifNode == m_CurrentOnvifNode)
        {
            // Defer polling camera if we are updating it
            m_TimerCounter = 0;
        }

        if (!m_CameraProxy.SetBrightnessColourContrast(onvifNode, setPacket.Brightness, setPacket.Colour, setPacket.Contrast))
        {
            RAYERROR("Failed to update BrightnessColourContrast");
        }
    }
}

void CIPVideoControl::_ToggleWatchdog()
{
    m_pProcessWatchdog->Toggle();
}

// Timer slot function.  Checks the map and sets any packets of data out of line
void CIPVideoControl::_OnCheckForAsyncComms()
{
    RAYTRACE("%d", _IsConnectionOpen());
    RAYTRACE("Exists? %d ", QFile::exists(cServerNameObjectPath));

    _ProcessMatchedNodes();

    _SendSetPackets();

    auto firstNodeOptional = m_NewNodes.TakeFirstNode();

    // Check if we have any new nodes that need to be added
    // Only process one node per timer tick to keep the Qt event loop free
    if (firstNodeOptional.first)
    {
        const std::string& OnvifNode = firstNodeOptional.second;

        if (m_IpCameraStore.IsNewCamera(OnvifNode))
        {
            RAYINFO("New OnvifNode detected %s", OnvifNode.c_str());
            m_CameraProxy.AddNode(OnvifNode);
            RAYINFO("Added node to cameraProxy");

            _ToggleWatchdog();

            std::string HiResRtspUri, LoResRtspUri, MacAddress, Name, Manufacturer, Model, FirmwareVersion, SerialNumber, IPAddress;
            if (m_CameraProxy.GetName(OnvifNode, Name) &&
                m_CameraProxy.GetManufacturer(OnvifNode, Manufacturer) &&
                m_CameraProxy.GetModel(OnvifNode, Model) &&
                m_CameraProxy.GetFirmwareVersion(OnvifNode, FirmwareVersion) &&
                m_CameraProxy.GetSerialNumber(OnvifNode, SerialNumber) &&
                m_CameraProxy.GetMacAddress(OnvifNode, MacAddress) &&
                m_CameraProxy.GetRtspUri(OnvifNode, m_AutoConfigEnabled, HiResRtspUri, LoResRtspUri) &&
                m_CameraProxy.GetIPAddress(OnvifNode, IPAddress))
            {
                _ToggleWatchdog();

                // Brightness, Colour & Contrast support is optional
                RAYINFO("Name=%s Manufacturer=%s Model=%s FirmwareVersion=%s SerialNumber=%s MacAddress=%s",
                        Name.c_str(), Manufacturer.c_str(), Model.c_str(), FirmwareVersion.c_str(), SerialNumber.c_str(), MacAddress.c_str());
                int Brightness = -1;
                int Contrast = -1;
                int Colour = -1;

                _ToggleWatchdog();

                RAYINFO("Requesting Brightness Colour Contrast values");
                m_CameraProxy.GetBrightnessColourContrast(OnvifNode, Brightness, Colour, Contrast);
                RAYINFO("Brightness Colour Contrast values received");

                m_IpCameraStore.AddNewCamera(OnvifNode, MacAddress, Name, Manufacturer, Model, FirmwareVersion, SerialNumber, HiResRtspUri, LoResRtspUri, IPAddress, Brightness, Colour, Contrast);
            }
        }
        else
        {
            _ToggleWatchdog();

            std::string HiResRtspUri, LoResRtspUri;
            if (m_CameraProxy.GetRtspUri(OnvifNode, m_AutoConfigEnabled, HiResRtspUri, LoResRtspUri))
            {
                m_IpCameraStore.OnUpdatedRtspUri(OnvifNode, HiResRtspUri, LoResRtspUri);
            }
        }
    }

    // We only poll the Brightness/Colour/Contrast of the current active camera
    if (++m_TimerCounter % c_PollingCurrentCameraInterval == 0)
    {
        if (!m_CurrentOnvifNode.empty())
        {
            int NewBrightness = -1;
            int NewContrast = -1;
            int NewColour = -1;

            _ToggleWatchdog();

            if (m_CameraProxy.GetBrightnessColourContrast(m_CurrentOnvifNode, NewBrightness, NewColour, NewContrast))
            {
                m_IpCameraStore.OnUpdatedBrightnessColourContrast(m_CurrentOnvifNode, NewBrightness, NewColour, NewContrast);
            }
        }
    }
}

void CIPVideoControl::_OnCheckConnection()
{
    if(!QFile::exists(cServerNameObjectPath) || !m_pLocalServer->isListening())
    {
#ifdef __ANDROID__
    QAndroidJniObject main = QtAndroid::androidService();
    if (main.isValid())
    {
        RAYINFO("Attempting to restart onvif");
        main.callMethod<void>("restartOnvifService");

    }
    else
    {
        RAYERROR("Failed to get MainService");
    }

#endif
    }
}


void NewNodeSet::Insert(const std::string& onvifNode)
{
    m_NewNodes.insert(onvifNode);
}

std::pair<bool, std::string> NewNodeSet::TakeFirstNode()
{
    std::pair<bool, std::string> result;

    result.first = !m_NewNodes.empty();
    if (result.first)
    {
        result.second = *m_NewNodes.begin();
        m_NewNodes.erase(m_NewNodes.begin());
    }

    return result;
}

void SetPacketMap::AddPacket(const std::string& onvifNode, const SetPacket& setPacket)
{
    m_SetPacketMap[onvifNode] = setPacket;
}

std::map<std::string, SetPacket> SetPacketMap::PurgeMap()
{
    auto result = m_SetPacketMap;
    m_SetPacketMap.clear();
    return result;
}

void MatchedNodes::Insert(const std::string& onvifNode)
{
    QMutexLocker lock(&m_MatchedNodesMutex);
    m_Nodes.insert(onvifNode);
}

std::set<std::string> MatchedNodes::PurgeSet()
{
    QMutexLocker lock(&m_MatchedNodesMutex);
    auto result = m_Nodes;
    m_Nodes.clear();
    return result;
}
