/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef IPVIDEOCONTROL_H_INCLUDED
#define IPVIDEOCONTROL_H_INCLUDED

#include <QObject>
#include <QString>
#include <QMutex>
#include <QStringList>
#include <QLocalServer>
#include <QLocalSocket>
#include "CameraProxy.h"
#include "IpCameraStore.h"
#include "DiscoveryService.h"

class IProcessWatchdog;

struct SetPacket
{
    int Brightness;
    int Contrast;
    int Colour;
};

class NewNodeSet
{
    std::set<std::string> m_NewNodes;
public:
    void Insert(const std::string& onvifNode);
    std::pair<bool, std::string> TakeFirstNode();
};

class SetPacketMap
{
    std::map<std::string, SetPacket> m_SetPacketMap;
public:
    void AddPacket(const std::string& onvifNode, const SetPacket& setPacket);
    std::map<std::string, SetPacket> PurgeMap();
};

class MatchedNodes
{
    MatchedNodes& operator=(const MatchedNodes&) = delete;
    MatchedNodes(const MatchedNodes&) = delete;
    std::set<std::string> m_Nodes;
    QMutex m_MatchedNodesMutex;
public:
    MatchedNodes() = default;
    void Insert(const std::string& onvifNode);
    std::set<std::string> PurgeSet();
};

class CIPVideoControl: public QObject
{
    Q_OBJECT

    CIPVideoControl& operator=(const CIPVideoControl&) = delete;
    CIPVideoControl(const CIPVideoControl&) = delete;

public:
    CIPVideoControl(IProcessWatchdog* pWatchdog, QObject* pParent = nullptr);

    void EnableCameraAutoConfiguration(bool enable);

    void SendAllCameras();

    void SetActiveCamera(QString macAddress);
    void SetContrast(QString macAddress, quint8 contrast);
    void SetBrightness(QString macAddress, quint8 brightness);
    void SetColour(QString macAddress, quint8 colour);
    void SetName(QString macAddress, QString name);

private:
    Q_SLOT void _OnProbeMatch(const std::string& onvifNode);
    Q_SLOT void _OnCheckForAsyncComms();
    Q_SLOT void _OnCheckConnection();

    Q_SLOT void _OnCameraFound(QVariantMap cameraMap);
    Q_SLOT void _OnCameraLost(QString macAddress);
    Q_SLOT void _OnParameterChanged(QString macAddress, QString parameterName, QVariant value);

    Q_SLOT void _OnNewConnection();
    Q_SLOT void _OnClientDisconnect();
    Q_SLOT void _OnClientDataReady();

    bool _IsConnectionOpen();
    void _InterpretPacket(qint8 header, QDataStream& data);
    void _DiconnectClient();
    void _DiconnectSocket(QLocalSocket*& rSocket);

    void _OnCameraMapChanged();

    void _AddSetPacket(const std::string& onvifNode, const SetPacket& setPacket);
    void _SendSetPackets();
    void _ProcessMatchedNodes();

    void _ToggleWatchdog();

    void _setupSocketServer();

    QLocalServer* m_pLocalServer;           // Our local server
    QLocalSocket* m_pConnection;            // Socket represeting client (only one client)
    QByteArray m_CommunicationBuffer;       // Our buffer for communication

    CCameraProxy m_CameraProxy;             // Camera Proxy object.
    CIpCameraStore m_IpCameraStore;         // Camera Store object.
    CDiscoveryService m_DiscoveryService;   // Discovery service object.

    SetPacketMap m_SetPacketMap;            // Map of set packets for each node
    NewNodeSet m_NewNodes;                  // Set of new nodes to be processed
    MatchedNodes m_MatchedNodes;            //
    QTimer m_Timer;                         // Timer to do perform async camera comms
    QTimer m_CheckConnectionTimer;
    unsigned int m_TimerCounter;            // Async tick timer counter for subdividing timer interval
    std::string m_CurrentOnvifNode;         // Current Onvif node being viewed on MFD
    bool m_AutoConfigEnabled;               // Are we allow to auto configure the camera

    IProcessWatchdog* m_pProcessWatchdog;
};

#endif // IPVIDEOCONTROL_H_INCLUDED
