﻿/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include "IpCameraStore.h"

#include <cstdio>
#include <QMetaType>
#include "FrameworkCommon/EventLogger.h"
#include "FrameworkCommon/RayAssert.h"

static const int c_ReaperTimerPeriodMs = 5000;
static const int c_DeathTimeMs = 15000;

static const QString c_MacAddressParameterKey        = "MacAddress";
static const QString c_NameParameterKey              = "Name";
static const QString c_ManufacturerParameterKey      = "Manufacturer";
static const QString c_ModelParameterKey             = "Model";
static const QString c_FirmwareVersionParameterKey   = "FirmwareVersion";
static const QString c_SerialNumberParameterKey      = "SerialNumber";
static const QString c_HiResRtspUriParameterKey      = "HiResRtspUri";
static const QString c_LoResRtspUriParameterKey      = "LoResRtspUri";
static const QString c_ContrastParameterKey          = "Contrast";
static const QString c_BrightnessParameterKey        = "Brightness";
static const QString c_ColourParameterKey            = "Colour";
static const QString c_TorchParameterKey             = "Torch";
static const QString c_ImageModeParameterKey         = "ImageMode";
static const QString c_CanEditContrastParameterKey   = "CanEditContrastParameterKey";
static const QString c_CanEditBrightnessParameterKey = "CanEditBrightnessParameterKey";
static const QString c_CanEditColourParameterKey     = "CanEditColourParameterKey";
static const QString c_CanEditTorchParameterKey      = "CanEditTorchParameterKey";
static const QString c_CanEditImageModeParameterKey  = "CanEditImageModeParameterKey";
static const QString c_IPAddressKey  = "IPAddress";

CIpCameraStore::CIpCameraStore()
{
    qRegisterMetaType<IpCameraDetails>("IpCameraDetails");
    qRegisterMetaType<std::string>("std::string");

    m_ReaperTimer.setSingleShot(false);
    m_ReaperTimer.setInterval(c_ReaperTimerPeriodMs);


    bool connected = connect(&m_ReaperTimer, &QTimer::timeout,
                             this, &CIpCameraStore::_CheckForCameraLoss);
    RAYASSERT(connected);
    if (connected)
    {
        m_ReaperTimer.start();
    }
}

bool CIpCameraStore::IsMacAdressInStore(QString MacAddress) const
{

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    bool Found = (itr != m_CameraMap.end());


    return Found;
}


bool CIpCameraStore::IsNewCamera(const std::string& OnvifNode) const
{

    CameraDirectory_t::const_iterator itr = m_CameraDirectory.find(OnvifNode);
    bool New = (itr == m_CameraDirectory.end());

    return New;
}

void CIpCameraStore::AddNewCamera(
        const std::string& OnvifNode,
        const std::string& MacAddress,
        const std::string& Name,
        const std::string& Manufacturer,
        const std::string& Model,
        const std::string& FirmwareVersion,
        const std::string& SerialNumber,
        const std::string& HiResRtspUri,
        const std::string& LoResRtspUri,
        const std::string& IPAddress,
        int Brightness,
        int Colour,
        int Contrast)
{
    bool ListChanged = false;

    // Do we already have this camera?
    CameraDirectory_t::const_iterator itr = m_CameraDirectory.find(OnvifNode);
    if (itr == m_CameraDirectory.end())
    {
        // Adding new entry
        RAYINFO("CameraFound adding new entry: %s", OnvifNode.c_str());
        m_CameraDirectory[OnvifNode] = MacAddress.c_str();

        IpCameraDetails NewDetails;
        NewDetails.Name = Name;
        NewDetails.Manufacturer = Manufacturer;
        NewDetails.Model = Model;
        NewDetails.FirmwareVersion = FirmwareVersion;
        NewDetails.SerialNumber = SerialNumber;
        NewDetails.HiResRtspUri = HiResRtspUri;
        NewDetails.LoResRtspUri = LoResRtspUri;
        NewDetails.MacAddress = MacAddress;
        NewDetails.LastSeenTime.start();
        NewDetails.OnvifNode = OnvifNode;
        NewDetails.Brightness = Brightness;
        NewDetails.Colour = Colour;
        NewDetails.Contrast = Contrast;
        NewDetails.IPAddress = IPAddress;
        m_CameraMap[MacAddress.c_str()] = NewDetails;

        QVariantMap Entry;
        Entry.insert(c_MacAddressParameterKey, QVariant(QString(MacAddress.c_str())));
        Entry.insert(c_NameParameterKey, QVariant(QString(Name.c_str())));
        Entry.insert(c_ManufacturerParameterKey, QVariant(QString(Manufacturer.c_str())));
        Entry.insert(c_ModelParameterKey, QVariant(QString(Model.c_str())));
        Entry.insert(c_FirmwareVersionParameterKey, QVariant(QString(FirmwareVersion.c_str())));
        Entry.insert(c_SerialNumberParameterKey, QVariant(QString(SerialNumber.c_str())));
        Entry.insert(c_HiResRtspUriParameterKey, QVariant(QString(HiResRtspUri.c_str())));
        Entry.insert(c_LoResRtspUriParameterKey, QVariant(QString(LoResRtspUri.c_str())));
        Entry.insert(c_ContrastParameterKey, QVariant(Contrast));
        Entry.insert(c_BrightnessParameterKey, QVariant(Brightness));
        Entry.insert(c_ColourParameterKey, QVariant(Colour));
        Entry.insert(c_TorchParameterKey, QVariant(false));
        Entry.insert(c_ImageModeParameterKey, QVariant(0));
        Entry.insert(c_CanEditContrastParameterKey, QVariant(Contrast != -1));
        Entry.insert(c_CanEditBrightnessParameterKey, QVariant(Brightness != -1));
        Entry.insert(c_CanEditColourParameterKey, QVariant(Colour != -1));
        Entry.insert(c_CanEditTorchParameterKey, QVariant(false));
        Entry.insert(c_CanEditImageModeParameterKey, QVariant(false));
        Entry.insert(c_IPAddressKey, QVariant(QString(IPAddress.c_str())));

        Q_EMIT CameraFound(Entry);
    }
}

void CIpCameraStore::OnSeenCamera(const std::string& OnvifNode)
{

    CameraDirectory_t::const_iterator itr = m_CameraDirectory.find(OnvifNode);
    RAYASSERT(itr != m_CameraDirectory.end());

    if (itr != m_CameraDirectory.end())
    {
        const QString& MacAddress = itr->second;
        IpCameraDetails& NewDetails = m_CameraMap[MacAddress];
        NewDetails.LastSeenTime.start();
    }



}

QString CIpCameraStore::GetMacAddress(const std::string& OnvifNode) const
{
    QString MacAddress;


    CameraDirectory_t::const_iterator itr = m_CameraDirectory.find(OnvifNode);
    if (itr != m_CameraDirectory.end())
    {
        MacAddress = itr->second;
    }



    return MacAddress;
}

int CIpCameraStore::GetBrightness(const QString& MacAddress) const
{
    int Brightness = 0;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Brightness = Details.Brightness;
    }

    return Brightness;
}


int CIpCameraStore::GetColour(const QString& MacAddress) const
{
    int Colour = 0;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Colour = Details.Colour;
    }

    return Colour;
}


int CIpCameraStore::GetContrast(const QString& MacAddress) const
{
    int Contrast = 0;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Contrast = Details.Contrast;
    }

    return Contrast;

}


QString CIpCameraStore::GetName(const QString& MacAddress) const
{
    QString Name;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Name = Details.Name.c_str();
    }

    return Name;
}

QString CIpCameraStore::GetManufacturer(const QString& MacAddress) const
{
    QString Name;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Name = Details.Manufacturer.c_str();
    }

    return Name;
}

QString CIpCameraStore::GetModel(const QString& MacAddress) const
{
    QString Name;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Name = Details.Model.c_str();
    }

    return Name;
}

QString CIpCameraStore::GetFirmwareVersion(const QString& MacAddress) const
{
    QString Name;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Name = Details.FirmwareVersion.c_str();
    }

    return Name;
}

QString CIpCameraStore::GetSerialNumber(const QString& MacAddress) const
{
    QString Name;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        Name = Details.SerialNumber.c_str();
    }

    return Name;
}


std::string CIpCameraStore::GetOnvifNode(const QString& MacAddress) const
{
    std::string OnvifNode;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        OnvifNode = Details.OnvifNode;
    }

    return OnvifNode;

}


QString CIpCameraStore::GetHiResRtspUri(const QString& MacAddress) const
{
    QString RtspUri;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        RtspUri = Details.HiResRtspUri.c_str();
    }

    return RtspUri;

}

QString CIpCameraStore::GetLoResRtspUri(const QString& MacAddress) const
{
    QString RtspUri;

    CameraMap_t::const_iterator itr = m_CameraMap.find(MacAddress);
    if (itr != m_CameraMap.end())
    {
        const IpCameraDetails& Details = itr->second;
        RtspUri = Details.LoResRtspUri.c_str();
    }

    return RtspUri;

}

void CIpCameraStore::GetOnvifNodeList (std::set<std::string>& rOnvifNodes) const
{
    rOnvifNodes.clear();

    CameraDirectory_t::const_iterator itr = m_CameraDirectory.begin();
    for (; itr != m_CameraDirectory.end(); ++itr)
    {
        rOnvifNodes.insert(itr->first);
    }

}

template <typename Map, typename F>
void map_erase_if(Map& m, F pred)
{
    typename Map::iterator i = m.begin();
    while ((i = std::find_if(i, m.end(), pred)) != m.end())
    {
        m.erase(i++);
    }
}

/// Slot function called by Reaper timer.  Removes cameras we haven't seen for a while
void CIpCameraStore::_CheckForCameraLoss()
{
    QStringList lostCameras;

    // Go through the cameras and kill any which have dropped off the network

    map_erase_if(m_CameraMap,
        [&lostCameras, this](const std::pair<QString, IpCameraDetails>& camera)
        {
            if (camera.second.LastSeenTime.elapsed() > c_DeathTimeMs)
            {
                lostCameras.push_back(camera.first);
                RAYINFO("Camera %s has been lost", camera.second.OnvifNode.c_str());
                m_CameraDirectory.erase(camera.second.OnvifNode);
                return true;
            }
            else
            {
                return false;
            }
        }
    );


    for (auto& name : lostCameras)
    {
        Q_EMIT CameraLost(name);
    }
}

QStringList CIpCameraStore::GetCameraList() const
{
    QStringList Output;


    CameraMap_t::const_iterator itr = m_CameraMap.begin();
    for (; itr != m_CameraMap.end(); ++itr)
    {
        const IpCameraDetails& Details = itr->second;
        Output.append(QString(Details.MacAddress.c_str()));
        RAYASSERT(itr->first.toStdString() == Details.MacAddress);
        RAYDETAIL("%s : MAC=%s, Name=%s", Details.OnvifNode.c_str(), Details.MacAddress.c_str(), Details.Name.c_str());
    }

    return Output;
}

void CIpCameraStore::SendAllCameras()
{
    QList<QVariantMap> Entries;
    // This is a variant map (keyed on MAC address) of variant maps (keyed on parameter name)


    CameraMap_t::const_iterator itr = m_CameraMap.begin();
    for (; itr != m_CameraMap.end(); ++itr)
    {
        const IpCameraDetails& Details = itr->second;
        RAYASSERT(itr->first.toStdString() == Details.MacAddress);

        // Make an entry for the variant map
        QVariantMap Entry;
        Entry.insert(c_MacAddressParameterKey, QVariant(QString(Details.MacAddress.c_str())));
        Entry.insert(c_NameParameterKey, QVariant(QString(Details.Name.c_str())));
        Entry.insert(c_ManufacturerParameterKey, QVariant(QString(Details.Manufacturer.c_str())));
        Entry.insert(c_ModelParameterKey, QVariant(QString(Details.Model.c_str())));
        Entry.insert(c_FirmwareVersionParameterKey, QVariant(QString(Details.FirmwareVersion.c_str())));
        Entry.insert(c_SerialNumberParameterKey, QVariant(QString(Details.SerialNumber.c_str())));
        Entry.insert(c_HiResRtspUriParameterKey, QVariant(QString(Details.HiResRtspUri.c_str())));
        Entry.insert(c_LoResRtspUriParameterKey, QVariant(QString(Details.LoResRtspUri.c_str())));
        Entry.insert(c_ContrastParameterKey, QVariant(Details.Contrast));
        Entry.insert(c_BrightnessParameterKey, QVariant(Details.Brightness));
        Entry.insert(c_ColourParameterKey, QVariant(Details.Colour));
        Entry.insert(c_TorchParameterKey, QVariant(false));
        Entry.insert(c_ImageModeParameterKey, QVariant(0));
        Entry.insert(c_CanEditContrastParameterKey, QVariant(true));
        Entry.insert(c_CanEditBrightnessParameterKey, QVariant(true));
        Entry.insert(c_CanEditColourParameterKey, QVariant(true));
        Entry.insert(c_CanEditTorchParameterKey, QVariant(false));
        Entry.insert(c_CanEditImageModeParameterKey, QVariant(false));
        Entry.insert(c_IPAddressKey, QVariant(QString(Details.IPAddress.c_str())));

        Entries << Entry;

        RAYDETAIL("%s : MAC=%s, Name=%s", Details.OnvifNode.c_str(), Details.MacAddress.c_str(), Details.Name.c_str());
}


    QList<QVariantMap>::const_iterator eitr = Entries.begin();
    for (; eitr != Entries.end(); ++ eitr)
    {
        Q_EMIT CameraFound(*eitr);
    }

}

void CIpCameraStore::OnNewCameraDetails(const IpCameraDetails& NewDetails)
{
    RAYDETAIL("%s : Name=%s", NewDetails.OnvifNode.c_str(), NewDetails.Name.c_str());

    bool BrightnessChange = false;
    bool ContrastChange = false;
    bool ColourChange = false;
    bool NameChange = false;
    bool ManufacturerChanged = false;
    bool ModelChanged = false;
    bool FirmwareVersionChanged = false;
    bool SerialNumberChanged = false;
    bool HiResUriChange = false;
    bool LoResUriChange = false;
    QString MacAddress;

    const std::string& OnvifNode = NewDetails.OnvifNode;
    if (!IsNewCamera(OnvifNode))
    {
        MacAddress = m_CameraDirectory[OnvifNode];
        IpCameraDetails& Details = m_CameraMap[MacAddress];
        RAYASSERT(MacAddress.toStdString() == Details.MacAddress);

        // Compare the bits we know about
        ManufacturerChanged = (NewDetails.Manufacturer != Details.Manufacturer);
        ModelChanged = (NewDetails.Model != Details.Manufacturer);
        FirmwareVersionChanged = (NewDetails.FirmwareVersion != Details.FirmwareVersion);
        SerialNumberChanged = (NewDetails.SerialNumber != Details.SerialNumber);
        BrightnessChange = (NewDetails.Brightness != Details.Brightness);
        ContrastChange = (NewDetails.Contrast != Details.Contrast);
        ColourChange = (NewDetails.Colour != Details.Colour);
        NameChange = (NewDetails.Name != Details.Name);
        HiResUriChange = (NewDetails.HiResRtspUri != Details.HiResRtspUri);
        LoResUriChange = (NewDetails.LoResRtspUri != Details.LoResRtspUri);

        Details.Brightness = NewDetails.Brightness;
        Details.Contrast = NewDetails.Contrast;
        Details.Colour = NewDetails.Colour;
        Details.Name = NewDetails.Name;
        Details.Manufacturer = NewDetails.Manufacturer;
        Details.Model = NewDetails.Model;
        Details.FirmwareVersion = NewDetails.FirmwareVersion;
        Details.SerialNumber = NewDetails.SerialNumber;
        Details.HiResRtspUri = NewDetails.HiResRtspUri;
        Details.LoResRtspUri = NewDetails.LoResRtspUri;
    }


    if (BrightnessChange)
    {
        Q_EMIT ParameterChanged(MacAddress, c_BrightnessParameterKey, QVariant(NewDetails.Brightness));
    }
    if (ContrastChange)
    {
        Q_EMIT ParameterChanged(MacAddress, c_ContrastParameterKey,  QVariant(NewDetails.Contrast));
    }
    if (ColourChange)
    {
        Q_EMIT ParameterChanged(MacAddress, c_ColourParameterKey, QVariant(NewDetails.Colour));
    }

    if (NameChange)
    {
        RAYINFO("Camera name changed : %s -> %s\n",
                MacAddress.toStdString().c_str(), NewDetails.Name.c_str());
        QString Name = NewDetails.Name.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_NameParameterKey, QVariant(Name));
    }

    if (ManufacturerChanged)
    {
    	QString Manufacturer = NewDetails.Manufacturer.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_ManufacturerParameterKey, QVariant(Manufacturer));
    }

    if (ModelChanged)
    {
    	QString Model = NewDetails.Model.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_ModelParameterKey, QVariant(Model));
    }

    if (FirmwareVersionChanged)
    {
    	QString FirmwareVersion = NewDetails.FirmwareVersion.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_FirmwareVersionParameterKey, QVariant(FirmwareVersion));
    }

    if (SerialNumberChanged)
    {
    	QString SerialNumber = NewDetails.SerialNumber.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_SerialNumberParameterKey, QVariant(SerialNumber));
    }

    if (HiResUriChange)
    {
        QString Uri = NewDetails.HiResRtspUri.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_HiResRtspUriParameterKey, QVariant(Uri));
    }

    if (LoResUriChange)
    {
        QString Uri = NewDetails.LoResRtspUri.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_LoResRtspUriParameterKey, QVariant(Uri));
    }
}


void CIpCameraStore::OnUpdatedCameraName(const std::string& OnvifNode, const std::string& Name)
{
    bool NameChange = false;
    QString MacAddress;

    if (!IsNewCamera(OnvifNode))
    {
        MacAddress = m_CameraDirectory[OnvifNode];
        IpCameraDetails& Details = m_CameraMap[MacAddress];

        if (Name != Details.Name)
        {
            Details.Name = Name;
            NameChange = true;
        }
    }


    if (NameChange)
    {
        QString EmitName = Name.c_str();
        RAYINFO("Camera name Changed %s -> %s\n", MacAddress.toStdString().c_str(), Name.c_str());
        Q_EMIT ParameterChanged(MacAddress, c_NameParameterKey, QVariant(EmitName));
    }
}



void CIpCameraStore::OnUpdatedBrightnessColourContrast(const std::string& OnvifNode, int Brightness, int Colour, int Contrast)
{
    bool BrightnessChange = false;
    bool ContrastChange = false;
    bool ColourChange = false;
    QString MacAddress;

    if (!IsNewCamera(OnvifNode))
    {
        MacAddress = m_CameraDirectory[OnvifNode];
        IpCameraDetails& Details = m_CameraMap[MacAddress];
        // Compare the bits we know about
        BrightnessChange = (Brightness != Details.Brightness);
        ContrastChange = (Contrast != Details.Contrast);
        ColourChange = (Colour != Details.Colour);

        if (BrightnessChange || ContrastChange || ColourChange)
        {
            Details.Brightness = Brightness;
            Details.Contrast = Contrast;
            Details.Colour = Colour;
        }
    }


    if (BrightnessChange)
    {
        RAYDETAIL("%s New Brightness:%d", OnvifNode.c_str(), Brightness);
        Q_EMIT ParameterChanged(MacAddress, c_BrightnessParameterKey, QVariant(Brightness));
    }
    if (ContrastChange)
    {
        RAYDETAIL("%s New Contrast:%d", OnvifNode.c_str(), Contrast);
        Q_EMIT ParameterChanged(MacAddress, c_ContrastParameterKey, QVariant(Contrast));
    }
    if (ColourChange)
    {
        RAYDETAIL("%s New Colour:%d", OnvifNode.c_str(), Colour);
        Q_EMIT ParameterChanged(MacAddress, c_ColourParameterKey, QVariant(Colour));
    }
}

void CIpCameraStore::OnUpdatedRtspUri(const std::string& OnvifNode, const std::string& HiResRtspUri,
                                      const std::string& LoResRtspUri)
{
    bool HiResUriChange = false;
    bool LoResUriChange = false;
    QString MacAddress;

    if (!IsNewCamera(OnvifNode))
    {
        MacAddress = m_CameraDirectory[OnvifNode];
        IpCameraDetails& Details = m_CameraMap[MacAddress];
        RAYASSERT(MacAddress.toStdString() == Details.MacAddress);

        // Compare the bits we know about
        HiResUriChange = (HiResRtspUri != Details.HiResRtspUri);
        LoResUriChange = (LoResRtspUri != Details.LoResRtspUri);

        if (HiResUriChange)
        {
            Details.HiResRtspUri = HiResRtspUri;
        }
        if (LoResUriChange)
        {
            Details.LoResRtspUri = LoResRtspUri;
        }
    }


    if (HiResUriChange)
    {
        QString Uri = HiResRtspUri.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_HiResRtspUriParameterKey, QVariant(Uri));
    }

    if (LoResUriChange)
    {
        QString Uri = LoResRtspUri.c_str();
        Q_EMIT ParameterChanged(MacAddress, c_LoResRtspUriParameterKey, QVariant(Uri));
    }

}
