/*
Copyright (c) 2016, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#ifndef IPCAMERASTORE_H_INCLUDED
#define IPCAMERASTORE_H_INCLUDED

#include <QObject>
#include <QString>
#include <QTime>
#include <QTimer>
#include <map>
#include <QStringList>
#include <set>
#include <QVariant>
#include <QElapsedTimer>

/// Structure holding a camera's details
struct IpCameraDetails
{
    IpCameraDetails()
    {
        LastSeenTime.start();
    }

    std::string OnvifNode;
    std::string HiResRtspUri;
    std::string LoResRtspUri;
    std::string MacAddress;
    std::string Manufacturer;
    std::string Model;
    std::string FirmwareVersion;
    std::string SerialNumber;
    std::string IPAddress;

    int Brightness = -1;
    int Colour = -1;
    int Contrast = -1;
    std::string Name;
    QElapsedTimer LastSeenTime;
};

/// Store object.  Holds details about the cameras we have.  Deals with timing
/// out cameras that have dropped off the network
class CIpCameraStore: public QObject
{
    Q_OBJECT

public:
    /// Standard constructor
    CIpCameraStore();

    /// Is this Onvif Node already known?
    /// \param[in] OnvifNode node to query
    /// \return bool true if we don't have this camera in the store
    bool IsNewCamera(const std::string& OnvifNode) const;

    /// Is this MAC already known?
    /// \param[in] MacAddress address to query
    /// \return bool true if we don't have this camera in the store
    bool IsMacAdressInStore(QString MacAddress) const;

    /// Function called when we've got a new camera.  Creates a new entry in the store
    /// \param[in] OnvifNode new node
    /// \param[in] MacAddress its MAC
    /// \param[in] Name its name
    /// \param[in] Manufacturer
    /// \param[in] Model
    /// \param[in] FirmwareVersion
    /// \param[in] SerialNumber
    /// \param[in] HiResRtspUri its High Resolution RTSP URI
    /// \param[in] LoResRtspUri its Low Resolution RTSP URI
    /// \param[in] Brightness its Brightness
    /// \param[in] Colour its Colour
    /// \param[in] Contrast its Contrast
    void AddNewCamera(const std::string& OnvifNode,
                      const std::string& MacAddress,
                      const std::string& Name,
                      const std::string& Manufacturer,
                      const std::string& Model,
                      const std::string& FirmwareVersion,
                      const std::string& SerialNumber,
                      const std::string& HiResRtspUri,
                      const std::string& LoResRtspUri,
                      const std::string& IPAddress,
                      int Brightness,
                      int Colour,
                      int Contrast);

    /// Notification function for when we've seen an already seen camera
    /// We use this to update the "last seen time"
    void OnSeenCamera(const std::string& OnvifNode);

    /// Populates the list with the OnvifNodes we currently know about
    /// \param[out] rOnvifNodes output list
    void GetOnvifNodeList(std::set<std::string>& rOnvifNodes) const;

    /// Getters
    QString GetMacAddress(const std::string& OnvifNode) const;

    int GetBrightness(const QString& MacAddress) const;
    int GetColour(const QString& MacAddress) const;
    int GetContrast(const QString& MacAddress) const;
    QString GetName(const QString& MacAddress) const;
    QString GetManufacturer(const QString& MacAddress) const;
    QString GetModel(const QString& Model) const;
    QString GetFirmwareVersion(const QString& FirmwareVersion) const;
    QString GetSerialNumber(const QString& SerialNumber) const;
    std::string GetOnvifNode(const QString& MacAddress) const;
    QString GetHiResRtspUri(const QString& MacAddress) const;
    QString GetLoResRtspUri(const QString& MacAddress) const;

    /// Getter for list of MAC addresses
    /// \return QStringList list of MAC Addresses
    QStringList GetCameraList() const;

    /// Getter for map of camera data
    void SendAllCameras();

    /// Signals for values changing.  Keyed on MAC address
    Q_SIGNAL void CameraFound(QVariantMap CameraMap);
    Q_SIGNAL void CameraLost(QString MacAddress);
    Q_SIGNAL void ParameterChanged(QString MacAddress, QString ParameterName, QVariant Value);

    /// Slot for when we've polled for up to date camera details.
    /// Can emit changed signals
    Q_SLOT void OnNewCameraDetails(const IpCameraDetails& NewDetails);

    /// Slot for polling camera names
    Q_SLOT void OnUpdatedCameraName(const std::string& OnvifNode, const std::string& Name);

    /// Slot for polling camera brightness, etc.
    Q_SLOT void OnUpdatedBrightnessColourContrast(const std::string& OnvifNode, int Brightness, int Colour, int Contrast);

    /// Slot for polling RTSP URIs
    Q_SLOT void OnUpdatedRtspUri(const std::string& OnvifNode, const std::string& HiResRtspUri,
                                 const std::string& LoResRtspUri);

private:
    Q_SLOT void _CheckForCameraLoss();
    typedef std::map<QString, IpCameraDetails> CameraMap_t; ///< MAC address to details
    typedef std::map<std::string, QString> CameraDirectory_t; ///< OnvifNode to MAC Address

    CameraDirectory_t m_CameraDirectory; ///< Map of OnvifNode to MAC Address
    CameraMap_t m_CameraMap;             ///< Map of MAC address to details structs
    QTimer m_ReaperTimer;                ///< Timer to trigger a death check
};

#endif // IPCAMERASTORE_H_INCLUDED
