/*
Copyright (c) 2013, Raymarine UK Limited
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of Raymarine UK Limited nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

// Originally developed from eCrash published in Linux Jornal by David Frascone
// http://www.linuxjournal.com/article/8724

// eCrash will allow you to capture stack traces in the
// event of a crash, and write those traces to disk, stdout,
// or any other file handle.

#include "FrameworkCommon/OMAPCrash.h"
#ifdef ENABLE_MEMORY_EVENT_LOGGER
#include "FrameworkCommon/MemoryEventLogger.h"
#endif
#include <sstream>
#include <iomanip>

#include <unistd.h>
#include <dirent.h>
#include <stdarg.h>
#include <string.h>
#include <syscall.h>
#include <fcntl.h>
#include <sys/sysinfo.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/klog.h>
#include <execinfo.h>
#include <pthread.h>
#include <errno.h>
#include <assert.h>
#include <sys/mman.h>
#include <limits.h>
#include <dlfcn.h>
#include <sys/resource.h>
#include <cxxabi.h>
#include <set>
#include <ucontext.h>
#include <dlfcn.h>
#include <stdlib.h>
#include <time.h>
#include <typeinfo>

// A flag which can be set prior to calling RAYABORT() to inhibit the generation
// of a screen dump image.
bool g_InhibitScreenDumpOnCrash = false;

// glibc abort message buffer
extern "C" char* __abort_msg;

//------------------------------------------------------------------------------
#define NIY()   printf("%s: Not Implemented Yet!\n", __FUNCTION__)

static eCrashParameters gbl_params;
static volatile int gbl_fd=-1;

static volatile bool gbl_backtraceActive = false;      // this flag is set when a backtrace is started, and is used to
// stop the thread this from being adjusted.

static volatile pid_t gbl_reportingThread = 0;    // this is the thread that has generated the first signal.
static char*  gbl_DemangleString = 0;
static char*  gbl_KLogBuffer = 0;
static int gbl_KLogBufferSize = 0;
static int gbl_InitialCrashSigNo = 0;

static bool   gbl_alreadyAborting = false;
static char   gbl_abortString[MAX_LINE_LEN] = {0};
static char   gbl_assertStringCrashLog[MAX_LINE_LEN] = {0};
static char   gbl_assertStringSyslog1[MAX_LINE_LEN] = {0};
static char   gbl_assertStringSyslog2[MAX_LINE_LEN] = {0};

//
static const unsigned int MAXIMUM_CODE_DUMP_SIZE_IN_BLOCKS = 100*1024*1024;
static size_t gbl_DemanglingStringSize = 1024*5;
/*
 * Private structures for our thread list
 */
struct ThreadListNode
{
    char *threadName;
    pthread_t thread;
    int backtraceSignal;
    struct sigaction prevAction;
    ThreadListNode *Next;
    pid_t LWP;

    volatile bool BacktraceDone;
    bool BacktraceOutput;
    int iBacktraceEntries;
    void** BacktraceAddrs;
};

static pthread_mutex_t ThreadListMutex = PTHREAD_MUTEX_INITIALIZER;
static ThreadListNode *ThreadList = NULL;

// Crash log sequence - This enum defines the ordered steps in the crash log
enum CrashLogSequence_t
{
    e_CrashLogNotStarted,
    e_CrashLogCollectThreads,
    e_CrashLogHaltAllThreads,
    e_CrashLogHeader,
    e_CrashLogBacktraceCurrentThread,
    e_CrashLogBacktraceAllThreads,
    e_CrashLogUnitInfo,
    e_CrashLogSTDOUT,
    e_CrashLogSysLog,
    e_CrashLogProcessLog,
    e_CrashLogMemMap,
    e_CrashLogFooter,
    e_CrashLogScreenDump
};
static CrashLogSequence_t gbl_CurrentCrashLogSequence = e_CrashLogNotStarted;

// Globals used for timing purposes.
static unsigned long long gbl_InitialStartTime;
static unsigned long long gbl_LastTimeRead;

#ifdef ENABLE_MEMORY_EVENT_LOGGER
static const unsigned int c_SanityCounterLimit = 20000;
#endif

/*********************************************************************
 *********************************************************************
 **     P  R  I  V  A  T  E      F  U  N  C  T  I  O  N  S
 *********************************************************************
 ********************************************************************/

static unsigned int GetTimeMillisecs(void)
{
    // Get Current time.
    struct timespec tp;
    clock_gettime(CLOCK_MONOTONIC, &tp);

    // Convert to milliseconds.
    unsigned long long TimeNow = (tp.tv_sec * 1000) + (tp.tv_nsec / 1000000);

    // Test for 32 bit wrap.
    static bool WrapHappened = false;

    if (TimeNow < gbl_LastTimeRead)
    {
        gbl_InitialStartTime = (gbl_LastTimeRead - gbl_InitialStartTime + 1);
        WrapHappened = true;
    }

    // Store this as the last read time.
    gbl_LastTimeRead = TimeNow;

    unsigned int ReturnTime = 0;

    if (!WrapHappened)
    {
        ReturnTime = (unsigned int)(TimeNow - gbl_InitialStartTime);
    }
    else
    {
        ReturnTime = (unsigned int)(TimeNow + gbl_InitialStartTime);
    }

    return ReturnTime;
}

static int  outputPrintf(const char *format, ...);
static void bt_handler(int signo, siginfo_t* pInfo, void* pVoid);

/*!
 * Insert a node into our threadList
 *
 * @param name   Text string indicating our thread
 * @param thread Our Thread Id
 * @param signo  Signal to create backtrace with
 * @param previousAction Our old handler for signo
 * @param isOssThread if true we get the OSS task, otherwise set field to null
 *
 * @returns pointer to new node no success, or null for failure
 */
static ThreadListNode* addThreadToList(const char *name, pthread_t thread,pid_t lwp,int signo)
{
    ThreadListNode* node = new ThreadListNode;

    if (node == 0)
    {
        return 0;
    }

    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));
    sa.sa_flags  = SA_SIGINFO;
    sa.sa_sigaction = bt_handler;

    sigaction(signo, &sa, &node->prevAction);

    DPRINTF(ECRASH_DEBUG_VERBOSE,
            "Adding thread LWP:%u (%s)\n", (unsigned int)lwp, name);
    node->threadName = strdup(name);
    node->thread = thread;
    node->backtraceSignal = signo;
    node->LWP = lwp;
    node->BacktraceDone = false;
    node->BacktraceOutput = false;
    node->iBacktraceEntries = 0;
    node->BacktraceAddrs = new void*[gbl_params.maxStackDepth + 5];

    /* And, add it to the list */
    pthread_mutex_lock(&ThreadListMutex);
    node->Next = ThreadList;
    ThreadList = node;
    pthread_mutex_unlock(&ThreadListMutex);

    return node;

} // addThreadToList

/*!
 * Remove a node from our threadList
 *
 * @param thread Our Thread Id
 *
 * @returns zero on success
 */
static int removeThreadFromList(pid_t tid)
{

    ThreadListNode *Probe, *Prev=NULL;
    ThreadListNode *Removed = NULL;

    DPRINTF(ECRASH_DEBUG_VERBOSE,
            "Removing LWP %u from list . . .\n", (unsigned int)tid);
    pthread_mutex_lock(&ThreadListMutex);
    for (Probe=ThreadList;Probe != NULL; Probe = Probe->Next)
    {
        if (Probe->LWP == tid)
        {
            // We found it!  Unlink it and move on!
            Removed = Probe;
            if (Prev == NULL) // head of list
            {
                ThreadList = Probe->Next;
            }
            else
            {
                // Prev != null, so we need to link around ourselves.
                Prev->Next = Probe->Next;
            }
            Removed->Next = NULL;
            break;
        }

        Prev = Probe;
    }
    pthread_mutex_unlock(&ThreadListMutex);

    // Now, if something is in Removed, free it, and return success
    if (Removed)
    {

        DPRINTF(ECRASH_DEBUG_VERBOSE,
                "   Found %s -- removing\n", Removed->threadName);
        // Reset the signal handler
        //sigaction(Removed->backtraceSignal, &Removed->prevAction, NULL);

        // And free the allocated memory
        delete [] Removed->BacktraceAddrs;
        free (Removed->threadName);
        delete Removed;

        return 0;
    }
    else
    {
        DPRINTF(ECRASH_DEBUG_VERBOSE,
                "   Not Found\n");
        return -1; // Not Found
    }
} // removeThreadFromList

/*!
 * Output text to a fd, looping to avoid being interrupted.
 *
 * @param str   String to output
 * @param bytes String length
 * @param fd    File descriptor to write to
 *
 * @returns bytes written, or error on failure.
 */
static int blockingWrite(const char *str, int bytes, int fd)
{
    int offset=0;
    int bytesWritten;
    int totalWritten = 0;

    while (bytes > 0)
    {
        bytesWritten = write(fd, &str[offset], bytes);
        if (bytesWritten < 1) break;
        totalWritten += bytesWritten;
        bytes -= bytesWritten;
    }

    return totalWritten;

} // blockingWrite

/*!
 * Print out an array of bytes to all of our output destinations.
 *
 * Return failure if we fail to output to any of them.
 *
 * @param string   String to be output
 *
 * @returns bytes written, or error on failure.
 */
static int outputBytes(const char *Data, int ByteCount)
{
    int return_value=0;

    // We're a happy camper -- start printing
    if (gbl_params.filename)
    {
        // append to our file -- hopefully it's been opened
        if (gbl_fd != -1)
        {
            if (blockingWrite(Data, ByteCount, gbl_fd))
            {
                return_value=-2;
            }
        }
    }

    // Write to our file pointer
    if (gbl_params.filep != NULL)
    {
        if (fwrite(Data, ByteCount, 1, gbl_params.filep) != 1)
        {
            return_value=-3;
        }
        fflush(gbl_params.filep);
    }

    // Write to our fd
    if (gbl_params.fd != -1)
    {
        if (blockingWrite(Data, ByteCount, gbl_params.fd))
        {
            return_value=-4;
        }
    }

    return return_value;
} // outputBytes

/*!
 * Print out a line of output to all our destinations
 *
 * One by one, output a line of text to all of our output destinations.
 *
 * Return failure if we fail to output to any of them.
 *
 * @param format   Normal printf style vararg format
 *
 * @returns bytes written, or error on failure.
 */
static int outputPrintf(const char *format, ...)
{
    // Our output line of text
    static char outputLine[MAX_LINE_LEN];
    int bytesInLine;
    va_list ap;
    int return_value=0;

    va_start(ap, format);

    bytesInLine = vsnprintf(outputLine, MAX_LINE_LEN-1, format, ap);
    if (bytesInLine > -1 && bytesInLine < (MAX_LINE_LEN-1))
    {
        // We're a happy camper -- start printing
        return_value = outputBytes(outputLine, bytesInLine);
    }
    else
    {
        // We overran our string.
        return_value=-1;
    }

    return return_value;
} // outputPrintf

/*!
 * Initialize our output (open files, etc)
 *
 * This file initializes all output streams, since we're about
 * to have output.
 *
 */
static void outputInit( void )
{
    if (gbl_params.filename)
    {
        gbl_fd = open(gbl_params.filename, O_WRONLY|O_CREAT|O_TRUNC,
                      S_IREAD|S_IWRITE|S_IRGRP|S_IROTH); // 0644
        if (gbl_fd < 0)
        {
            gbl_fd = -1;
        }
    }
} // outputInit

/*!
 * Reopen our output (open files, etc)
 *
 * This file initializes all output streams, since we're about
 * to have output.
 *
 */
static void outputReOpen( void )
{
    if (gbl_params.filename)
    {
        gbl_fd = open(gbl_params.filename, O_APPEND|O_WRONLY,
                      S_IREAD|S_IWRITE|S_IRGRP|S_IROTH); // 0644
        if (gbl_fd < 0)
        {
            gbl_fd = -1;
        }
    }
} // outputInit

/*!
 * Finalize our output (close files, etc)
 *
 * This file closes all output streams.
 *
 */
static void outputFini( void )
{
    if (gbl_fd > -1)
        close(gbl_fd);

    if (gbl_params.filep != NULL)
        fclose(gbl_params.filep);

    if (gbl_params.fd > -1)
        close(gbl_params.fd);

    // Just in case someone tries to call outputPrintf after outputFini
    gbl_fd = gbl_params.fd = -1;
    gbl_params.filep = NULL;
    sync();

} // outputFini


static void *lookupClosestSymbol(eCrashSymbolTable *table,
                                 void *address)
{
    int addr;
    eCrashSymbol *last=NULL;

    // For now, use a linear lookup.
    DPRINTF(ECRASH_DEBUG_VERBOSE,
            "Looking for %p in %d symbols\n", address, table->numSymbols);
    for (addr=0; addr < table->numSymbols; addr++)
    {
        DPRINTF(ECRASH_DEBUG_VERBOSE,
                "  Examining [%d] %p\n", addr,
                table->symbols[addr].address);
        if (table->symbols[addr].address > address)
        {
            break;
        }
        last = &table->symbols[addr];
    }

    // last will either be NULL, or the last address less than the
    // one we're looking for.
    DPRINTF(ECRASH_DEBUG_VERBOSE,
            "Returning %s (%p)\n", last?last->function:"(nil)",
            last?last->address:0);
    return last;

} // lookupClosestSymbol

/*!
 * Convert a signal number to a string.
 *
 */

static const char* SignalToString(int signo)
{
    const char *s = "Unknown";
    switch (signo)
    {
    case    SIGSEGV:    s = "Segmentation fault";           break;
    case    SIGFPE:     s = "Floating Point Exception";     break;
    case    SIGILL:     s = "Illegal Instruction";          break;
    case    SIGBUS:     s = "Bus error";                    break;
    case    SIGSYS:     s = "Invalid System Call";          break;
    case    SIGABRT:    s = "Abort";                        break;

    }
    return s;
}

static void spinForever() __THROW __attribute__ ((__noreturn__));
static void spinForever() __THROW
{
    while (true)
    {
        usleep(1000000); // sleep for 1 second.
    }
}

/*!
 * Demangle the output from backtrace_symbols
 */
static void demangleSymbol(const char *symbol)
{
    char   *pcDemangledName;
    size_t iDemangledSize;
    int status = 0;

/* __cxa_demangle will call realloc on the buffer if it is too small. We
 * therefore check the returned pointer and update the location of the global
 * buffer in the event that it is different. */
    pcDemangledName = abi::__cxa_demangle
       (symbol, gbl_DemangleString, &iDemangledSize, &status);
    if ( pcDemangledName != NULL && pcDemangledName != gbl_DemangleString )
    {
       gbl_DemangleString       = pcDemangledName;
       gbl_DemanglingStringSize = iDemangledSize;
    }

    if (status != 0)
    {
        // If demangling failed just copy the original symbol
        strncpy(gbl_DemangleString, symbol, gbl_DemanglingStringSize-1);
        gbl_DemangleString[gbl_DemanglingStringSize-1] = 0;
    }
}

/*!
 * Output our current stack's backtrace
 */
static void outputBacktrace(ThreadListNode *pNode)
{
    DPRINTF(ECRASH_DEBUG_VERBOSE, "outputBacktrace %s started\n", pNode->threadName);

    outputPrintf("*  Backtrace of \"%s\" (pthread:0x%08X, LWP:%d)\n",
                         pNode->threadName,
                         (unsigned int)pNode->thread,
                         (int)pNode->LWP);


    // Remove recursive call stack from top of backtrace. This happens on some targets (e.g. ARM) with the clone() function at the start of a thread
    int backtraceCount = pNode->iBacktraceEntries;
    for (int i = pNode->iBacktraceEntries -1; i > 0; --i)
    {
        if (pNode->BacktraceAddrs[i] == pNode->BacktraceAddrs[i - 1])
        {
            backtraceCount = i;
        }
        else
        {
            break;
        }
    }

    for (int i = 0; i < backtraceCount; ++i)
    {
        if (gbl_params.symbolTable)
        {
            eCrashSymbol *symbol;

            symbol = (eCrashSymbol *) lookupClosestSymbol(gbl_params.symbolTable,
                                                          const_cast<void*>(pNode->BacktraceAddrs[i]));

            if (symbol)
            {
                outputPrintf("*      Frame %02d: %s+%u\n",
                             i, symbol->function,
                             (char*) pNode->BacktraceAddrs[i] - (char*) symbol->address);
            }
            else
            {
                outputPrintf("*      Frame %02d: %p\n", i,
                             pNode->BacktraceAddrs[i]);
            }
        }
        else
        {
            gbl_DemangleString[0] = 0;
            Dl_info       dlInfo;

            dladdr(pNode->BacktraceAddrs[i], &dlInfo);
            if (dlInfo.dli_saddr != 0)
            {
                char Offset[16];
                // Demangle the symbol using a global buffer so the c++ lib doesn't allocate one with malloc
                demangleSymbol(dlInfo.dli_sname);
                sprintf(Offset, "+0x%lx", reinterpret_cast<unsigned long>(pNode->BacktraceAddrs[i]) - reinterpret_cast<unsigned long>(dlInfo.dli_saddr));
            }

            outputPrintf
               ("*      Frame %02d: %s(%s) ",i, dlInfo.dli_fname, gbl_DemangleString);
            outputPrintf
               ("[%p]\n",pNode->BacktraceAddrs[i]);
        } // symbolTable
    }

    pNode->BacktraceOutput = true;

    DPRINTF(ECRASH_DEBUG_VERBOSE, "outputBacktrace %s finished\n", pNode->threadName);

} // outputGlobalBacktrace

/*!
 * Handle signals (bt signals)
 *
 * This function shoudl be called to generate a crashdump into our
 * global area.  Once the dump has been completed, this function will
 * return after tickling a global.  Since mutexes are not async
 * signal safe, the main thread, after signaling us to generate our
 * own backtrace, will sleep for a few seconds waiting for us to complete.
 * To ensure that this
 *
 * @param signum Signal received.
 */
static void
bt_handler(int /*signo*/,
           siginfo_t* /*pInfo*/,
           void* pVoid)
{
    /* Give thread co-ordinating crash log a chance to finish signalling all threads before we start backtracing */
    usleep(10000);

    /* Get the thread id for what we are about to backtrace.*/
    pid_t id = (pid_t)syscall(__NR_gettid);

    ThreadListNode *probe;
    for (probe=ThreadList; probe; probe=probe->Next)
    {
        if (probe->LWP == id) break;
    }

    DPRINTF(ECRASH_DEBUG_VERBOSE, "Backtracing LWP:%u, %s started\n", id, probe->threadName);
    // Call the backtrace function(s) for the specific hardware platform
    probe->iBacktraceEntries = backtrace(const_cast<void**>(probe->BacktraceAddrs),
                                         gbl_params.maxStackDepth);

    probe->BacktraceDone = true;

    DPRINTF(ECRASH_DEBUG_VERBOSE, "Backtracing %s finished\n", probe->threadName);

    // If this is not the thread being used to report the errors,
    // then spin it forever, so that it cannot do any harm and cause
    // any more signal, or corrupt the mutex list.
    if (id != gbl_reportingThread)
    {
        spinForever();
    }

}

static void
outputBacktraceThreads(void)
{
    // When we're backtracing, don't worry about the mutex . . hopefully
    // we're in a safe place.

    unsigned int StartTime = GetTimeMillisecs();
    int LastOutstandingThreads = 0;

    while ((GetTimeMillisecs() - StartTime) < gbl_params.threadWaitTime)
    {
        int OutstandingThreads = 0;

        __sync_synchronize();
        for (ThreadListNode* probe=ThreadList; probe != 0; probe=probe->Next)
        {
            if (probe->BacktraceOutput == false)
            {
                if (probe->BacktraceDone)
                {
                    outputBacktrace(probe);
                    outputPrintf("*\n");
                }
                else
                {
                    ++OutstandingThreads;
                }
            }
        }

        if (LastOutstandingThreads != OutstandingThreads)
        {
            StartTime = GetTimeMillisecs();
            LastOutstandingThreads = OutstandingThreads;
        }
        if (OutstandingThreads > 0)
        {
            // Sleep and let the threads backtrace
            usleep(500000);
        }
        else
        {
            break;
        }
    }


    for (ThreadListNode* probe=ThreadList; probe != 0; probe=probe->Next)
    {
        if (probe->BacktraceDone == false)
        {
            outputPrintf("*  Error: unable to get backtrace of \"%s\" (0x%08X) [LWP=%d]\n",
                         probe->threadName, (unsigned int)probe->thread, (int)probe->LWP);
        }
    }

} // outputBacktraceThreads

/*!
 * get the thread name for given thread.
 *
 */
static const char* getThreadName(pid_t thread)
{
    ThreadListNode *Probe;
    const char* threadName = NULL;

    for (Probe=ThreadList;Probe != NULL && threadName == NULL; Probe = Probe->Next)
    {
        if (Probe->LWP == thread)
        {
            threadName = Probe->threadName;
        }
    }

    if (threadName == NULL)
    {
        threadName = "unknown";
    }

    return threadName;
}

/*!
 * get the thread name for the current (offending) thread.
 *
 */
static const char* getCurrentThreadName()
{
    return getThreadName((pid_t)syscall(__NR_gettid));
}

/*!
 * DoesFileExist
 *
 * Determine if the given file exists.
 */
static bool DoesFileExist(const char* filename)
{
    bool result = false;
    struct stat fileStatus;

    int r = stat(filename, &fileStatus);
    if (r == 0 && S_ISREG(fileStatus.st_mode))
    {
        result = true;
    }
    return result;
}

/*!
 * IsCoreDumpRequired
 *
 * Determine is a core dump is to be created.
 */
static bool IsCoreDumpRequired()
{
    bool result = false;

    if (gbl_params.doCoreDumpIndicatorFile)
    {
        result = DoesFileExist(gbl_params.doCoreDumpIndicatorFile);
    }
    return result;
}

/*!
 * IsCoreDumpRequired
 *
 * Determine is a core dump is to be created.
 */
static void SetCoreDumpSize(unsigned int coreDumpSizeInBlocks)
{
    struct rlimit limits;

    limits.rlim_cur = coreDumpSizeInBlocks;
    limits.rlim_max = coreDumpSizeInBlocks;

    int r = setrlimit(RLIMIT_CORE, &limits);

    if (r != 0)
    {
        fprintf(stderr, "failed to set core dump limit\n");
    }
}

/*!
 * DoAbort
 *
 * Having decided that a crash has occurred the only way out is to abort.
 * abort() is used rather than exit(), as exit() will attempt to do a
 * tidy c++ exit and hence destroy any static objects. In the land of RL90
 * (C,E,G Series), that will probably (actually definately) cause another crash.
 *
 * Whether abort() generates a core dump is determined by the
 * ulimit -c value. See eCrash.h 'doCoreDumpIndicatorFile' for description
 * of how to configure eCrash to generate a coreDump.
 */
static void DoAbort()
{
    // Should we delay here before aborting?
    // NOTE: Depends if we have environment variable DELAY_BEFORE_ABORT set!
    if (gbl_params.delayBeforeFinalAbort)
    {
        fprintf(stderr, "\n\n*************** DELAYING BEFORE ABORT ***************\n\n");
        fflush(stderr);

        struct timespec tsreq, tsrem;
        tsreq.tv_sec = 99999;
        tsreq.tv_nsec = 0;
        nanosleep(&tsreq, &tsrem);
    }

    if (IsCoreDumpRequired())
    {
        SetCoreDumpSize(MAXIMUM_CODE_DUMP_SIZE_IN_BLOCKS);
    }
    else
    {
        // 0 means no core dump.
        SetCoreDumpSize(0);
    }

    kill(getpid(), SIGKILL);

    // Replaced for now
    /*
    // We now do a SIGABRT not SIGKILL because a SIGKILL won't be handled by DDS libraries
    // and their shared memory could have consistency problems.
    
    // Uninstall the SIGABRT handler so when we do a SIGABRT we don't get called back
    struct sigaction sa;
    memset(&sa, 0, sizeof(sa));

    struct sigaction previousSa;
    sigaction(SIGABRT, &sa, &previousSa);

    // Now trigger the SIGABRT signal
    kill(getpid(),SIGABRT);
    */
}

/*!
 * dumpWIndow
 *
 * Grab the contents of the main window and write them to a file.
 */
static void dumpWindow()
{
#ifndef RELEASE
    // TODO - Implement new Qt version
#endif
}

/*!
 * output a list of the registered threads
 */
#ifdef DEBUG_OUTPUT_HELPER

static void outputThreadList()
{
    ThreadListNode *probe;

    outputPrintf("* Thread list:\n");

    for (probe=ThreadList; probe; probe=probe->Next)
    {
        outputPrintf("* Thread: %s (0x%08X) [LWP=%d]\n", probe->threadName, (unsigned int)probe->thread, (int)probe->LWP);
    }
    outputPrintf("* End of thread list\n");
}

#endif // DEBUG_OUTPUT_HELPER

/*!
 * output current date and time
 */
static void outputDateAndTime()
{
    char timeAndDate[256] = {0};
    time_t t = time(0);
    struct tm localtimeResult;
    if (strftime(timeAndDate, sizeof(timeAndDate) - 1, "%a %d %b %Y %H:%M:%S", localtime_r(&t, &localtimeResult)) > 0)
    {
        outputPrintf("* Time...................: %s\n", timeAndDate);
    }
}

static char * formatTime(unsigned int ms)
{
    static char buffer[100];

    const unsigned int msInSecond = 1000;
    const unsigned int msInMinute = msInSecond * 60;
    const unsigned int msInHour   = msInMinute * 60;
    const unsigned int msInDay    = msInHour * 24;

    snprintf(buffer, sizeof(buffer),
             "%d day%s %02d:%02d:%02d.%03d",
             ms / msInDay,
             (ms / msInDay) == 1 ? "" : "s",
             (ms % msInDay) / msInHour,
             (ms % msInHour) / msInMinute,
             (ms % msInMinute) / msInSecond,
             (ms % msInSecond));

    return buffer;
}

static void outputApplicationRunningTime()
{
    // GetTimeMillisecs -- returns the number milliseconds since the application
    // was started.
    //
    unsigned int applicationRunningTime = GetTimeMillisecs();

    outputPrintf("* Running time...........: %s = %ums\n", formatTime(applicationRunningTime), applicationRunningTime);
}

//
// running popen from a signal handler may be a bit mad, as popen is
// not one of the defined system reentrant methods. But as uptime gives us
// some useful information it is run last so that if it should fail
// it will not matter.
//
static void outputUpTime()
{
    struct sysinfo SysInfo;

    if (sysinfo(&SysInfo) == 0)
    {
        outputPrintf("* Uptime.................: %s\n", formatTime((unsigned int)SysInfo.uptime * 1000));
        outputPrintf("* Load average...........: %.2f, %.2f, %.2f\n", ((float)SysInfo.loads[0] / 65536.0), ((float)SysInfo.loads[1] / 65536.0), ((float)SysInfo.loads[2] / 65536.0));
        outputPrintf("* Mem usage..............: totalram=%dk, freeram=%dk, bufferram=%dk, procs=%d\n", (SysInfo.totalram  / 1024) * SysInfo.mem_unit,
                                                                                                  (SysInfo.freeram   / 1024) * SysInfo.mem_unit,
                                                                                                  (SysInfo.bufferram / 1024) * SysInfo.mem_unit,
                                                                                                  SysInfo.procs);
    }
    else
    {
        outputPrintf("* Failed to call sysinfo\n");
    }
}

/*!
 * output time related informatiion.
 */
static void outputTimeInformation()
{
    outputDateAndTime();
    outputApplicationRunningTime();
    outputUpTime();
}

/*!
 * output the old contenst of STDOUT
 */
static void outputSTDOUT()
{
    outputPrintf("*----------------------------------------------------------------------------\n");
    outputPrintf("* STDOUT buffer (probably incomplete!)\n");
    unsigned int len = stdout->_IO_buf_end - stdout->_IO_buf_base;
    outputBytes(stdout->_IO_buf_base, len);
    outputPrintf("\n* End of STDOUT\n");
}

static unsigned int uiStoreUnwindOffset = 0;
static int *piStoreUnwind;

/* Store unwind information at a location defined by uiStoreUnwindOffset and
 * piStoreUnwind. A pointer to this function is given to the RayUnwind_Context
 * funtion in libgcc_s which then calls it to store the unwind information. */
void StoreUnwind(int pc)
{
   if (uiStoreUnwindOffset < gbl_params.maxStackDepth)
   {
      piStoreUnwind[uiStoreUnwindOffset] = pc;
      uiStoreUnwindOffset++;
   }
}

static void WriteToSyslog(int facility_priority, const char* pFormatString, ... )
{
    va_list ArgumentList;
    va_start(ArgumentList, pFormatString);

    if (gbl_params.applicationName != NULL)
    {
        const char* pName = strrchr(gbl_params.applicationName, '/');
        if (pName)
        {
            pName++;
        }
        else
        {
            pName = gbl_params.applicationName;
        }
        openlog(pName, LOG_PID, LOG_USER);
    }
    else
    {
        openlog("Unknown", LOG_PID, LOG_USER);
    }
    vsyslog(facility_priority, pFormatString, ArgumentList);
    closelog();

    va_end(ArgumentList);
}

/*!
 * Handle signals (crash signals)
 *
 * This function will catch all crash signals, and will output the
 * crash dump.
 *
 * It will physically write (and sync) the current thread's information
 * before it attempts to send signals to other threads.
 *
 * @param signum Signal received.
 * @param pInfo  Singal Info structure, values are signo dependant
 * @param pVoid  ?
 */
static void crash_handler(int signo, siginfo_t * pInfo, void * pVoid)
{
    ThreadListNode* pThisThreadNode = 0;
    pid_t thisThread = (pid_t)syscall(__NR_gettid);

    DPRINTF(ECRASH_DEBUG_VERY_VERBOSE,"Crash LWP:%u with signal %s occurred, gbl_backtraceActive=%d\n", thisThread, SignalToString(signo), gbl_backtraceActive);
    WriteToSyslog(LOG_INFO, "Crash with signal %s occurred, gbl_backtraceActive=%d", SignalToString(signo), gbl_backtraceActive);

    if (gbl_backtraceActive == false)
    {
        // Stop threads from adjusting the list whilst dumping it.
        gbl_backtraceActive = true;
        gbl_reportingThread = thisThread;
        gbl_InitialCrashSigNo = signo;

    }
    else
    {
        //
        // If we get another signal from any thread other
        // then the thread report the error, then spin
        // forever. The reporting thread we let continue
        // in the hope if will finish as best it can.
        //
        if (thisThread != gbl_reportingThread)
        {
            DPRINTF(ECRASH_DEBUG_ERROR,"Non reporting LWP %u crashed with signal %s, spinning forever...", thisThread, SignalToString(signo));
            WriteToSyslog(LOG_ERR, "Non reporting LWP %u crashed with signal %s, spinning forever...", thisThread, SignalToString(signo));
            spinForever();
        }
        else if ((gbl_CurrentCrashLogSequence >= e_CrashLogHeader) && (gbl_CurrentCrashLogSequence <= e_CrashLogFooter))
        {
            WriteToSyslog(LOG_WARNING, "Crash Logger Crashed with signal %s at sequence %d, attempting to continue", SignalToString(signo), gbl_CurrentCrashLogSequence);
            outputPrintf("*  Error: Crash Logger Crashed with signal %s, attempting to continue\n",  SignalToString(signo));
        }
    }

    if (gbl_assertStringSyslog1[0])
    {
        WriteToSyslog(LOG_INFO, "%s", gbl_assertStringSyslog1);
        WriteToSyslog(LOG_INFO, "%s", gbl_assertStringSyslog2);
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogCollectThreads)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogCollectThreads;
        DIR* dir = opendir("/proc/self/task");
        if (dir)
        {
            struct dirent* dirent;
            while ((dirent = readdir(dir)) != 0)
            {
                if (isdigit(dirent->d_name[0]))
                {
                    pid_t tid = atoi(dirent->d_name);
                    if (tid != 0)
                    {
                        ThreadListNode *probe;
                        for (probe=ThreadList; probe; probe=probe->Next)
                        {
                            if (probe->LWP == tid) break;
                        }
                        if (probe == 0)
                        {
                            char buffer[128];
                            sprintf(buffer, "/proc/self/task/%s/stat", dirent->d_name);
                            FILE* f = fopen(buffer, "r");
                            if (f)
                            {
                                if (fscanf(f, "%*d (%[^)]", buffer) == 0)
                                {
                                    sprintf(buffer, "NonOssThread%s", dirent->d_name);
                                }
                                fclose(f);
                            }
                            addThreadToList(buffer, 0, tid, gbl_params.defaultBacktraceSignal);
                        }
                    }
                }
            }
            closedir(dir);
        }
    }

    DPRINTF(ECRASH_DEBUG_VERY_VERBOSE,"Done collecting threads\n");

    if (gbl_CurrentCrashLogSequence < e_CrashLogHaltAllThreads)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogHaltAllThreads;

        WriteToSyslog(LOG_INFO, "Crash logger backtracing all threads");
        // Make all our threads backtrace immediately so we get an accurate view of the current state in the log
        for (ThreadListNode *probe=ThreadList; probe != 0; probe=probe->Next)
        {
            if (gbl_reportingThread == probe->LWP)
            {
                pThisThreadNode = probe;
                continue;
            }

            int iKill = syscall(SYS_tgkill, (int)getpid(), (int)probe->LWP, probe->backtraceSignal);
            if (iKill != 0)
            {
                const char* errString = "";
                if (errno == EINVAL) errString = " (Not a valid signal number)";
                if (errno == ESRCH) errString = " (Not a running thread)";
                WriteToSyslog(LOG_ERR, "Error: %s tgkill returned %d%s", probe->threadName, iKill, errString);
            }
        }

        // If the offending thread isn't known to OSS (e.g. a DDS thread),
        // we need to add an entry for it in the task list
        if (pThisThreadNode == 0)
        {
            const unsigned int c_MaxNameLength = 32;
            char name[c_MaxNameLength];
            pid_t lwp = (pid_t)syscall(__NR_gettid);
            snprintf(name, c_MaxNameLength, "NonOssThread%d", lwp);

            // Add to the thread list
            // The new node is returned from this function
            pThisThreadNode = addThreadToList(name, pthread_self(), lwp, gbl_params.defaultBacktraceSignal);

        }
    }

    DPRINTF(ECRASH_DEBUG_VERY_VERBOSE,"Done halting all threads\n");

    if (gbl_CurrentCrashLogSequence < e_CrashLogHeader)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogHeader;

        if (gbl_params.pFunctionTraceOutputFunction != 0)
        {
            // Dummy call with null parameter to tell function tracer to stop tracing
            gbl_params.pFunctionTraceOutputFunction(0);
        }


        outputInit();

        outputPrintf("*********************************************************\n");
        outputPrintf("*                   Crash Handler\n");
        outputPrintf("*********************************************************\n");
        outputPrintf("*\n");


        bool AbortMsgFound = false;
        if (gbl_InitialCrashSigNo == SIGABRT)
        {
            if (gbl_abortString[0])
            {
                outputPrintf("*  %s\n", gbl_abortString);
                AbortMsgFound = true;
            }
            else if (gbl_assertStringCrashLog[0])
            {
                outputPrintf("*  %s\n", gbl_assertStringCrashLog);
                AbortMsgFound = true;
            }
            else if (__abort_msg)
            {
                if (__abort_msg[0])
                {
                    outputPrintf("*  %s\n", __abort_msg);
                    AbortMsgFound = true;
                }
            }
        }
        
        if (AbortMsgFound == false)
        {
            outputPrintf("*  Got a crash! signo=%d - %s\n", signo, SignalToString(gbl_InitialCrashSigNo));
            outputPrintf("*      Errno: %d\n", pInfo->si_errno);
            outputPrintf("*      Code: %d\n", pInfo->si_code);
            outputPrintf("*      Pid: %d\n", pInfo->si_pid);
            outputPrintf("*      Uid: %d\n", pInfo->si_uid);
            outputPrintf("*      Status: %d\n", pInfo->si_status);
            outputPrintf("*      Utime: %d\n", pInfo->si_utime);
            outputPrintf("*      Stime: %d\n", pInfo->si_stime);
            outputPrintf("*      Value: %d\n", pInfo->si_value);
            outputPrintf("*      Int: %d\n", pInfo->si_int);
            outputPrintf("*      ptr: %d\n", pInfo->si_ptr);
            outputPrintf("*      Addr: %p\n", pInfo->si_addr);
            outputPrintf("*      Band: %d\n", pInfo->si_band);
            outputPrintf("*      Fd: %d\n", pInfo->si_fd);
        }
        outputPrintf("*\n");

        bool const ApplicationNameIsValid = ((gbl_params.applicationName != NULL) && (strlen(gbl_params.applicationName) > 0));
        outputPrintf("*  Application     : %s\n",  (ApplicationNameIsValid) ? gbl_params.applicationName : "unknown");
        bool const SoftwareVersionIsValid = ((gbl_params.softwareVersionString != NULL) && (strlen(gbl_params.softwareVersionString) > 0));
        outputPrintf("*  Software version: %s\n",  (SoftwareVersionIsValid) ? gbl_params.softwareVersionString : "unknown");
        outputPrintf("*\n");

        if (gbl_params.filename)
        {
            WriteToSyslog(LOG_ERR, "Crash with signal %s occurred. Generating Log File %s", SignalToString(gbl_InitialCrashSigNo), gbl_params.filename);
        }
        else
        {
            WriteToSyslog(LOG_ERR, "Crash with signal %s occurred", SignalToString(gbl_InitialCrashSigNo));
        }
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogBacktraceCurrentThread)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogBacktraceCurrentThread;
        WriteToSyslog(LOG_INFO, "Crash logger backtracing current thread");

        const char* offendingThreadName = getCurrentThreadName();
        outputPrintf("*  Offending Thread's Backtrace: %s [LWP=%d]\n", offendingThreadName ? offendingThreadName : " ", (int)syscall(__NR_gettid));
        outputPrintf("*\n");

        if (pThisThreadNode)
        {
            bt_handler(signo, pInfo, pVoid);
            outputBacktrace(pThisThreadNode);
        }
        else
        {
            outputPrintf("*  Error: Unable to determine current thread node\n");
        }
        outputPrintf("*\n");
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogBacktraceAllThreads)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogBacktraceAllThreads;
        if (gbl_params.dumpAllThreads)
        {
            if (gbl_params.filename)
            {
                WriteToSyslog(LOG_INFO, "Crash logger writing all backtraces to %s", gbl_params.filename);
            }
            outputBacktraceThreads();
        }
        outputPrintf("*\n");
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogUnitInfo)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogUnitInfo;
        if (gbl_params.filename)
        {
            WriteToSyslog(LOG_INFO, "Crash logger writing unit info to %s", gbl_params.filename);
        }
        //
        // extended information, this is place at the end of log as it was added late on the project
        // and it was very important that printing the extended information did not break the
        // output of the really important stuff. Additional, see comments in outputUpTime.
        //
        //
        outputPrintf("*----------------------------------------------------------------------------\n");
        outputPrintf("*  Additional Information\n");
        outputPrintf("*\n");

        if (gbl_params.pFunctionTraceOutputFunction != 0)
        {
            gbl_params.pFunctionTraceOutputFunction(outputPrintf);
        }

        outputTimeInformation();
        outputPrintf("*\n");
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogSTDOUT)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogSTDOUT;
        if (gbl_params.filename)
        {
            WriteToSyslog(LOG_INFO, "Crash logger writing stdout to %s", gbl_params.filename);
        }
        outputSTDOUT();
        outputPrintf("*\n");
    }

#ifdef ENABLE_MEMORY_EVENT_LOGGER
    {
        outputPrintf("===========================\n");
        outputPrintf("Log buffer\n");
        outputPrintf("===========================\n");
        unsigned int SanityCounter = 0;
        MemoryEventLogger::InitMemoryBufferRead();
        char *str = MemoryEventLogger::GetNextMemoryBufferEntry();
        while (str && SanityCounter < c_SanityCounterLimit)
        {
            outputPrintf("%s\n", str);
            ++SanityCounter;
            str = MemoryEventLogger::GetNextMemoryBufferEntry();
        }
        if (SanityCounter >= c_SanityCounterLimit)
        {
            outputPrintf("** Log limit reached -- these are NOT the final entries (syslog will have more) **");
        }
        outputPrintf("===========================\n");
    }
#endif

    // Close the file here because we're about to
    outputFini();

    if (gbl_CurrentCrashLogSequence < e_CrashLogSysLog)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogSysLog;
        if (gbl_params.filename)
        {
            WriteToSyslog(LOG_INFO, "Crash logger writing syslog to %s", gbl_params.filename);

            const int c_MaxCommand = 256;
            char Command[c_MaxCommand];
            snprintf(Command, c_MaxCommand, "echo \\* Start of Syslog >> %s", gbl_params.filename);
            int result = system(Command);
            (void)result; // Prevent cppcheck performance issue about unused result.
            snprintf(Command, c_MaxCommand, "tail -n1000 /raymarine/Data/Log/syslog/current >> %s", gbl_params.filename);
            result = system(Command);
            (void)result; // Prevent cppcheck performance issue about unused result.
            snprintf(Command, c_MaxCommand, "echo \\* End of Syslog >> %s", gbl_params.filename);
            result = system(Command);
            (void)result; // Prevent cppcheck performance issue about unused result.
            snprintf(Command, c_MaxCommand, "echo \\* ------------------------------------------ >> %s", gbl_params.filename);
            result = system(Command);
            (void)result; // Prevent cppcheck performance issue about unused result.
        }
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogProcessLog)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogProcessLog;
        if (gbl_params.filename)
        {
            WriteToSyslog(LOG_INFO, "Crash logger writing process log to %s", gbl_params.filename);
            const int c_MaxCommand = 256;
            char Command[c_MaxCommand];
            snprintf(Command, c_MaxCommand, "echo \\* Start of Process Log >> %s", gbl_params.filename);
            int result = system(Command);
            (void)result; // Prevent cppcheck performance issue about unused result.
            snprintf(Command, c_MaxCommand, "top -bn1 >> %s", gbl_params.filename);
            result = system(Command);
            (void)result; // Prevent cppcheck performance issue about unused result.
            snprintf(Command, c_MaxCommand, "echo \\* End of Process Log >> %s", gbl_params.filename);
            result = system(Command);
            (void)result; // Prevent cppcheck performance issue about unused result.
            snprintf(Command, c_MaxCommand, "echo \\* ------------------------------------------ >> %s", gbl_params.filename);
            result = system(Command);
        (void)result; // Prevent cppcheck performance issue about unused result.
        }
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogMemMap)
    {
        if (gbl_params.filename)
        {
            WriteToSyslog(LOG_INFO, "Crash logger writing memory map to %s", gbl_params.filename);
        }
        gbl_CurrentCrashLogSequence = e_CrashLogMemMap;
        const int c_MaxCommand = 256;
        char Command[c_MaxCommand];
        snprintf(Command, c_MaxCommand, "echo \\* Start of Memory Map >> %s", gbl_params.filename);
        int result = system(Command);
        (void)result; // Prevent cppcheck performance issue about unused result.
        snprintf(Command, c_MaxCommand, "cat /proc/%d/maps >> %s", getpid(), gbl_params.filename);
        result = system(Command);
        (void)result; // Prevent cppcheck performance issue about unused result.
        snprintf(Command, c_MaxCommand, "echo \\* End of Memory Map >> %s", gbl_params.filename);
        result = system(Command);
        (void)result; // Prevent cppcheck performance issue about unused result.
        snprintf(Command, c_MaxCommand, "echo \\* ------------------------------------------ >> %s", gbl_params.filename);
        result = system(Command);
        (void)result; // Prevent cppcheck performance issue about unused result.
    }

    outputReOpen();

    if (gbl_CurrentCrashLogSequence < e_CrashLogFooter)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogFooter;
        outputPrintf("*\n");
        outputPrintf("*********************************************************\n");
        outputPrintf("*               eCrash Crash Handler\n");
        outputPrintf("*********************************************************\n");

        outputFini();

        if (gbl_params.filename)
        {
            WriteToSyslog(LOG_ERR, "Finished generating Log File %s", gbl_params.filename);
        }
    }

    if (gbl_CurrentCrashLogSequence < e_CrashLogScreenDump)
    {
        gbl_CurrentCrashLogSequence = e_CrashLogScreenDump;
        if (gbl_params.windowDumpFile && g_InhibitScreenDumpOnCrash == FALSE)
        {
            dumpWindow();
        }
    }

    DoAbort();

} // crash_handler


/*!
 * Validate a passed-in symbol table
 *
 * For now, just print it out (if verbose), and make sure it's
 * sorted and none of the pointers are zero.
 */
static int ValidateSymbolTable( void )
{
    int i;
    int rc=0;
    unsigned long lastAddress =0;

    // Get out of here if the table is empty
    if (!gbl_params.symbolTable) return 0;

    // Dump it in verbose mode
    DPRINTF(ECRASH_DEBUG_VERBOSE,
            "Symbol Table Provided with %d symbols\n",
            gbl_params.symbolTable->numSymbols);
    for (i=0; i < gbl_params.symbolTable->numSymbols; i++)
    {
        // Dump it in verbose mode
        DPRINTF(ECRASH_DEBUG_VERBOSE,
                "%-30s %p\n",
                gbl_params.symbolTable->symbols[i].function,
                gbl_params.symbolTable->symbols[i].address);
        if (lastAddress >
            (unsigned long)gbl_params.symbolTable->symbols[i].address)
        {
            DPRINTF(ECRASH_DEBUG_ERROR,
                    "Error: symbol table is not sorted (last=%p, current=%p)\n",
                    (void *)lastAddress,
                    gbl_params.symbolTable->symbols[i].address);
            rc = -1;
        }

    } // for

    return rc;

} // ValidateSymbolTable

/*********************************************************************
 *********************************************************************
 **      P  U  B  L  I  C      F  U  N  C  T  I  O  N  S
 *********************************************************************
 ********************************************************************/

#ifndef NDEBUG
// Replace the libstdc++ function that is called to throw exceptions. This allows
// us to get decent backtraces from the point that they are thrown.
extern "C" void __cxa_throw(void* Exception, void* TypeInfo, void (*Dest)(void*))
{
    const char* ExceptionName = reinterpret_cast<const std::type_info*>(TypeInfo)->name();

//    printf("Exception: '%s'\n", ExceptionName);

    // Qpid and Navionics libs throw their own exceptions in normal use, so
    // we unfortunately have to ignore anything that isn't a std::exception.
    if (ExceptionName != 0 && ExceptionName[0] == 'S' && ExceptionName[1] == 't')  // 'St' is mangled shorthand for '::std::'
    {
        // The demangling doesn't look thread safe, so keep it for actual abort scenarios.
        gbl_DemangleString[0] = 0;
        demangleSymbol(ExceptionName);
        eCrash_Abort("An exception (%s) was thrown, see stack trace for details", gbl_DemangleString);
    }
    else
    {
        typedef void (*ThrowFuncPtr)(void*,void*,void(*)(void*));
        static const ThrowFuncPtr RealThrow = reinterpret_cast<ThrowFuncPtr>(dlsym(RTLD_NEXT, "__cxa_throw"));
        if (RealThrow)
        {
            RealThrow(Exception, TypeInfo, Dest);
        }
        else
        {
            // Couldn't find the real __cxa_throw function in a later shared library?
            eCrash_Abort("An exception was thrown, but we couldn't rethrow it! See %s:%d", __FILE__, __LINE__);
        }
    }
    for (;;);
}

extern "C" void __cxa_pure_virtual(void)
{
    eCrash_Abort("A non-overridden pure virtual function was called (probably from a base-class destructor).");
}
#endif

/*!
 * Initialize eCrash.
 *
 * This function must be called before calling any other eCrash
 * functions.  It sets up the global behavior of the system, and
 * registers the calling thread for crash dumps.
 *
 * @param params Our input parameters.  The passed in structure will be copied.
 *
 * @return Zero on success.
 */
int eCrash_Init(eCrashParameters *params)
{
    gbl_InitialStartTime = GetTimeMillisecs();

    //pUnitInfo = CUnitInfo::instance();

    int ret = 0;

#ifdef DO_SIGNALS_RIGHT

    sigset_t blocked;
    struct sigaction act;
    sigemptyset(&blocked);
    act.sa_sigaction = crash_handler;
    act.sa_mask = blocked;
    act.sa_flags = SA_SIGINFO;

#endif // DO_SIGNALS_RIGHT

    if (params != NULL)
    {
        // Make ourselves a global copy of params.
        gbl_params = *params;
        if (gbl_params.filename)
            gbl_params.filename = strdup(params->filename);

        if (gbl_params.doCoreDumpIndicatorFile)
            gbl_params.doCoreDumpIndicatorFile = strdup(params->doCoreDumpIndicatorFile);

        if (gbl_params.windowDumpFile)
            gbl_params.windowDumpFile = strdup(params->windowDumpFile);

        // Set our defaults, if they weren't specified
        if (gbl_params.maxStackDepth == 0)
            gbl_params.maxStackDepth = ECRASH_DEFAULT_STACK_DEPTH;

        if (gbl_params.defaultBacktraceSignal == 0)
            gbl_params.defaultBacktraceSignal = ECRASH_DEFAULT_BACKTRACE_SIGNAL;

        if (gbl_params.threadWaitTime == 0)
            gbl_params.threadWaitTime = ECRASH_DEFAULT_THREAD_WAIT_TIME_MS;

        if (gbl_params.debugLevel == 0)
            gbl_params.debugLevel = ECRASH_DEBUG_DEFAULT;

        DPRINTF(ECRASH_DEBUG_VERY_VERBOSE,"Init Starting params = %p\n", params);

        if (gbl_params.softwareVersionString)
            gbl_params.softwareVersionString = strdup(params->softwareVersionString);

        // Copy our symbol table
        if (gbl_params.symbolTable)
        {
            DPRINTF(ECRASH_DEBUG_VERBOSE,
                    "symbolTable @ %p -- %d symbols\n", gbl_params.symbolTable,
                    gbl_params.symbolTable->numSymbols);
            // Make a copy of our symbol table
            gbl_params.symbolTable = (eCrashSymbolTable*) malloc(sizeof(eCrashSymbolTable));
            memcpy(gbl_params.symbolTable, params->symbolTable,
                   sizeof(eCrashSymbolTable));

            // Now allocate / copy the actual table.
            gbl_params.symbolTable->symbols = (eCrashSymbol*) malloc(sizeof(eCrashSymbol) *
                                                                     gbl_params.symbolTable->numSymbols);
            memcpy(gbl_params.symbolTable->symbols,
                   params->symbolTable->symbols,
                   sizeof(eCrashSymbol) * gbl_params.symbolTable->numSymbols);

            ValidateSymbolTable();
        }

        struct sigaction sa;
        memset(&sa, 0, sizeof(sa));
        sa.sa_sigaction = crash_handler;
        sa.sa_flags = SA_SIGINFO;

        // And, finally, register for our signals
        for (int sigIndex=0; gbl_params.signals[sigIndex] != 0; ++sigIndex)
        {
            DPRINTF(ECRASH_DEBUG_VERY_VERBOSE,
                    "   Catching signal[%d] %d\n", sigIndex,
                    gbl_params.signals[sigIndex]);

            sigaction(gbl_params.signals[sigIndex], &sa, NULL);
        }
    }
    else
    {
        DPRINTF(ECRASH_DEBUG_ERROR, "   Error:  Null Params!\n");
        ret = -1;
    }

    gbl_DemangleString = static_cast<char*>(malloc(gbl_DemanglingStringSize));

    // Read Kernel Log ring buffer size
    gbl_KLogBufferSize = klogctl(10, NULL, 0);
    gbl_KLogBuffer = new char[gbl_KLogBufferSize];

    DPRINTF(ECRASH_DEBUG_VERY_VERBOSE, "Init Complete ret=%d\n", ret);

    return ret;
} /* eCrash_Init */

/*!
 * UnInitialize eCrash.
 *
 * This function may be called to de-activate eCrash, release the
 * signal handlers, and free any memory allocated by eCrash.
 *
 * @return Zero on success.
 */
int eCrash_Uninit( void )
{
    NIY();

    free(gbl_DemangleString);
    gbl_DemangleString = 0;

    delete [] gbl_KLogBuffer;
    gbl_KLogBuffer = 0;

    // Decrement the Unit Info singleton reference count.
    //pUnitInfo.reset();

    return 0;
} /* eCrash_Uninit */

/*!
 * Register a thread for backtracing on crash.
 *
 * This function must be called by any thread wanting it's stack
 * dumped in the event of a crash.  The thread my specify what
 * signal should be used, or the default, SIGUSR1 will be used.
 *
 * @param signo Signal to use to generate dump (default: SIGUSR1)
 *
 * @return Zero on success.
 */
int eCrash_RegisterThread(const char *name, int signo)
{
    int result = 1;

    //
    // If there is a crash being handled at this moment
    // then don't register this thread, as it will fiddle with
    // the list of thread.
    //
    if (!gbl_backtraceActive)
    {
        // Register for our signal
        if (signo == 0)
        {
            signo = gbl_params.defaultBacktraceSignal;
        }

        // result is zero for success
        // cppcheck-suppress leakNoVarFunctionCall
        if (addThreadToList(name, pthread_self(), (pid_t)syscall(__NR_gettid), signo) != 0)
        {
            result = 0;
        }
    }

    return result;

} /* eCrash_RegisterThread */

/*!
 * Un-register a thread for stack dumps.
 *
 * This function may be called to un-register any previously
 * registered thread.
 *
 * @return Zero on success.
 */
int eCrash_UnregisterThread( void )
{
    int result = 1;

    //
    // Don't remove thread from list if a backtrace is currently active
    // as there is no sensible method of semaphoring the thread list.
    //
    if (!gbl_backtraceActive)
    {
        result = removeThreadFromList((pid_t)syscall(__NR_gettid));
    }
    return result;

} /* eCrash_UnregisterThread */


/*!
 * Abort method.
 */
void eCrash_Abort(const char *format, ...) __THROW
{
    //
    // Ignore all apart from the first abort.
    //
    if (!gbl_alreadyAborting)
    {
        gbl_alreadyAborting = true;

        va_list ap;
        va_start(ap, format);

        strcpy(gbl_abortString, "*  ");
        vsnprintf(gbl_abortString + strlen(gbl_abortString), sizeof(gbl_abortString) - strlen(gbl_abortString) - 1, format, ap);

        //
        // Abort will cause a SIGABT which will kick off the complete logging process
        //
        abort();
    }
    spinForever();
}

/*!
 * Abort method.
 */
void eCrash_Abort(const char* err_str, int err_no, const char *file, int line) __THROW
{

    //
    // Ignore all apart from the first abort.
    //
    if (!gbl_alreadyAborting)
    {
        gbl_alreadyAborting = true;

        snprintf(gbl_abortString, sizeof(gbl_abortString) - 1, "Abort: %s (%d)\n*         %s:%d", err_str, err_no, file, line);

        //
        // Abort will cause a SIGABT which will kick off the complete logging process
        //
        abort();
    }
    spinForever();
}

/*!
 * Assert method.
 */
void eCrash_assert(const char *cond, const char* file, int line) __THROW
{
    //
    // Ignore all apart from the first abort.
    //
    if (!gbl_alreadyAborting)
    {
        gbl_alreadyAborting = true;

        snprintf(gbl_assertStringCrashLog, sizeof(gbl_assertStringCrashLog) - 1, "Assertion '%s' failed.\n*  %s:%d", cond, file, line);
        snprintf(gbl_assertStringSyslog1, sizeof(gbl_assertStringSyslog1) - 1, "Assertion '%s' failed", cond);
        snprintf(gbl_assertStringSyslog2, sizeof(gbl_assertStringSyslog2) - 1, "At file & line: %s:%d", file, line);


#ifndef NDEBUG
        // Note that __assert_fail() will result in abort() being called.
        __assert_fail(cond, file, line, (const char *) 0);
#endif
    }
    spinForever();
}
